#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdio>
#include <ext/hash_map> // for string hash code
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/fs.h>

#include "pCache.h"  // for confMap
#include "BlockAccessor.h"
#include "Util.h"     // for make directory

// BlockAccessor abstracts the disk access
BlockAccessor::BlockAccessor() {
    std::stringstream root_dir_string(confMap[std::string("ROOT_DIR")]);
    root_dir_string >> root_dir;
    std::stringstream dev_name_string(confMap[std::string("DEV_NAME")]);
    dev_name_string >> dev_name;

    // update UI
    stats->root_dir = root_dir;
    stats->dev_name = dev_name;
}

bool BlockAccessor::time_stamp = false;     // initialize it
// This function is called *after* all disk accesses if the time_stamp
// flag is still false. So, this function is call only once.
// Typically, caller functions calls sync() before mark_time_stamp() to
// instruct disk buffer flusing the data. This should triger the first
// log entry in block trace file.
void BlockAccessor::mark_time_stamp() {
    stats->lprintf(Stats::INFO, "DL %lld", Util::get_timestamp());
    time_stamp = true;
}

BlockAccessor::~BlockAccessor() {

}

// *******************************************************************************
// 1) Flat directory approach, each page cluster would be written into a separate
// file 
FlatAccessor::FlatAccessor() :
    BlockAccessor()
{

}

// read blocks, store the meta data (_oid) in file space use simple file name
void FlatAccessor::blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks, char* _data) {
	std::string file_name = get_name(_oid, _offset);
    FILE *fp = fopen(file_name.c_str(), "r"); 
    if (fp != NULL) {
        if (0 >= fread(_data, blockPool->blockSize, _no_blocks, fp)) {
            std::cout << "Unable to read from file: " << file_name << std::endl;
        }
        fclose(fp);
    }
    else {
        std::cout << "Unable to open file for reading: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// write blocks, store the meta data (_oid) in file space use simple file name
void FlatAccessor::blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                         long long _no_blocks, char* _data) {
	std::string file_name = get_name(_oid, _offset);
    FILE *fp = fopen(file_name.c_str(), "w"); 
    if (fp != NULL) {	
        if (0 >= fwrite(_data, blockPool->blockSize, _no_blocks, fp)) {
            std::cout << "Unable to write to file: " << file_name << std::endl;
        }
        fclose(fp);
    }
    else {
        std::cout << "Unable to open file for writing: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// remove blocks, remove a file or a few associated blocks
void FlatAccessor::blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                      long long _no_blocks) {
	std::string file_name = get_name(_oid, _offset);
    if (0 != remove(file_name.c_str())) {
        std::cout << "Unable to remove file: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// determine the file name using a hash function
// the returned name is in absolute path format
// caller should free the string in heap
std::string FlatAccessor::get_name(std::string _oid, long long _offset){
	std::stringstream buf;
	buf << root_dir << "/" << _oid << "." << _offset << ".cache";

	return buf.str();
}

// *******************************************************************************
// 2) SquidAccessor simulates 2 level squid approach: 16 first level, 256 second
// level (in each first level directory).
SquidAccessor::SquidAccessor() :
    BlockAccessor()
{

}

// read blocks, store the meta data (_oid) in file space use simple file name
void SquidAccessor::blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks, char* _data) {
	std::string file_name = get_filename(_oid, _offset);
    FILE *fp = fopen(file_name.c_str(), "r"); 
    if (fp != NULL) {
        if (0 >= fread(_data, blockPool->blockSize, _no_blocks, fp)) {
            std::cout << "Unable to read from file: " << file_name << std::endl;
        }
        fclose(fp);
    }
    else {
        std::cout << "Unable to open file for reading: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// write blocks, store the meta data (_oid) in file space use simple file name
void SquidAccessor::blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                         long long _no_blocks, char* _data) {
	std::string file_name = get_filename(_oid, _offset);
    Util::mkdirs(file_name.c_str(), 777);  // make sure the directory exists
    FILE *fp = fopen(file_name.c_str(), "w"); 
    if (fp != NULL) {	
        if (0 >= fwrite(_data, blockPool->blockSize, _no_blocks, fp)) {
            std::cout << "Unable to write to file: " << file_name << std::endl;
        }
        fclose(fp);
    }
    else {
        std::cout << "Unable to open file for writing: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// remove blocks, remove a file or a few associated blocks
void SquidAccessor::blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                      long long _no_blocks) {
	std::string file_name = get_filename(_oid, _offset);
    if (0 != remove(file_name.c_str())) {
        std::cout << "Unable to remove file: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// determine the file name using a hash function
// the returned name is in absolute path format
// caller should free the string in heap
std::string SquidAccessor::get_1_dir_name(std::string _oid, long long _offset){
	std::stringstream buf;
	buf << _oid << "." << _offset; 
    std::string full_name = buf.str();
    // get a char between A and P <--- same as 0 to F used in squid
    char _num = 65 + (__gnu_cxx::__stl_hash_string(full_name.c_str()) % 16);
	return std::string(1, _num); // repeat once
}

std::string SquidAccessor::get_2_dir_name(std::string _oid, long long _offset){
	std::stringstream buf;
	buf << _oid << "." << _offset; 
    std::string full_name = buf.str();
    size_t second_hash_code = (size_t) (__gnu_cxx::__stl_hash_string(full_name.c_str()) / 16.0f);
    // get a char between A and P <--- same as 0 to F used in squid
    char _num1 = 65 + (second_hash_code % 16);
    char _num2 = 65 + (((int)(second_hash_code / 16.0f)) % 16);

	return std::string(1, _num1) + std::string(1, _num2); // repeat once
}

std::string SquidAccessor::get_filename(std::string _oid, long long _offset){
	std::stringstream buf;
	buf << root_dir << "/" << 
           get_1_dir_name(_oid, _offset) << "/" <<
           get_2_dir_name(_oid, _offset) << "/" << 
           _oid << "." << _offset << ".cache";
	return buf.str();
}

// *******************************************************************************
// 3) OneLevelAccessor creates a directory for every object (note, for BitTorrent
// a directory for every piece). It is analogous to saving web objects from the
// same web site inside one directory.
OneLevelAccessor::OneLevelAccessor() :
    BlockAccessor()
{

}

// read blocks, store the meta data (_oid) in file space use simple file name
void OneLevelAccessor::blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks, char* _data) {
	std::string file_name = get_filename(_oid, _offset);
    FILE *fp = fopen(file_name.c_str(), "r"); 
    if (fp != NULL) {
        if (0 >= fread(_data, blockPool->blockSize, _no_blocks, fp)) {
            std::cout << "Unable to read from file: " << file_name << std::endl;
        }
        fclose(fp);
    }
    else {
        std::cout << "Unable to open file for reading: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// write blocks, store the meta data (_oid) in file space use simple file name
void OneLevelAccessor::blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                         long long _no_blocks, char* _data) {
	std::string file_name = get_filename(_oid, _offset);
    Util::mkdirs(file_name.c_str(), 777);  // make sure the directory exists
    FILE *fp = fopen(file_name.c_str(), "w"); 
    if (fp != NULL) {	
        if (0 >= fwrite(_data, blockPool->blockSize, _no_blocks, fp)) {
            std::cout << "Unable to write to file: " << file_name << std::endl;
        }
        fclose(fp);
    }
    else {
        std::cout << "Unable to open file for writing: " << file_name << std::endl;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// remove blocks, remove a file or a few associated blocks
void OneLevelAccessor::blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                      long long _no_blocks) {
	std::string file_name = get_filename(_oid, _offset);
    if (0 != remove(file_name.c_str())) {
        std::cout << "Unable to remove file: " << file_name << std::endl;
    }
}

std::string OneLevelAccessor::get_filename(std::string _oid, long long _offset){
	std::stringstream buf;
	buf << root_dir << "/" << 
           _oid << "/" << 
           _oid << "." << _offset << ".cache";
	return buf.str();
}

// *******************************************************************************
// 4) PCacheFs - PCacheFs uses only one file to save all data. Meta-data are
// stored in-memory. Disk accesses are instructed using block numbers with
// llseek calls. 
PCacheFsAccessor::PCacheFsAccessor() :
    BlockAccessor()
{
	std::stringstream filename;
	filename << root_dir << "/" << "cache.cache";
    std::cout << "Opening (single) file: " << filename.str() << " for disk cache." << std::endl;
    fd = open(filename.str().c_str(), O_RDWR | O_CREAT | O_LARGEFILE);  // indeed, this is a large file
    if (fd < 0) {
        perror("Warning: Unable to open (single) file for disk cache: ");
    }
}

// not called. put it here first
PCacheFsAccessor::~PCacheFsAccessor() {
    close(fd);
}

// read blocks, store the meta data (_oid) in file space use simple file name
void PCacheFsAccessor::blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks, char* _data) {
    // move the file descriptor
    if (-1 == lseek64(fd, (long long) _start_block_num * (long long) blockPool->blockSize, SEEK_SET)) {
        perror("Warning: Unable to seek for reading file: ");
        return;
    }

    // note: length may be too large for read to handle...although unlikely
    if (-1 == read(fd, _data, (size_t) (blockPool->blockSize * _no_blocks))) {
        perror("Warning: Unable to read from cache file: " );
        return;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// write blocks, store the meta data (_oid) in file space use simple file name
void PCacheFsAccessor::blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                         long long _no_blocks, char* _data) {
    // move the file descriptor
    if (-1 == lseek64(fd, (long long) _start_block_num * (long long) blockPool->blockSize, SEEK_SET)) {
        perror("Warning: Unable to seek for writing file: ");
        return;
    }

    // note: length may be too large for write to handle...although unlikely
    if (-1 == write(fd, _data, blockPool->blockSize * _no_blocks)) {
        perror("Warning: Unable to write to cache file: " );
        return;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// remove blocks, remove a file or a few associated blocks
void PCacheFsAccessor::blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                      long long _no_blocks) {
    // need not do anything...nice nice.
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// *******************************************************************************
// 5) PCacheRaw - PCacheRaw uses raw partition to save all data. Meta-data are
// stored in-memory. Disk accesses are instructed using block numbers with
// llseek calls. 
PCacheRawAccessor::PCacheRawAccessor() :
    BlockAccessor()
{
    // Bear, 08/01/02, we no longer use direct flag in raw partition io
    //fd = open(dev_name.c_str(),  O_DIRECT | O_RDWR);
    fd = open(dev_name.c_str(),  O_RDWR);
    if (fd < 0) {
        perror("Warning: Unable to open raw partition for disk cache: ");
    }

    // quick check on block size
    int soft_block_size;
    if (0 != ioctl(fd, BLKBSZGET, &soft_block_size)) {
        perror("Unable to get soft block size of the raw parition: ");
    }
    else {
        printf("soft_block_size is %d, memory buffer should be aligned.\n", soft_block_size);
    }
}

// not called. put it here first
PCacheRawAccessor::~PCacheRawAccessor() {
    close(fd);
}

// read blocks, store the meta data (_oid) in file space use simple file name
void PCacheRawAccessor::blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks, char* _data) {
    // move the file descriptor
    if (-1 == lseek64(fd, (long long) _start_block_num * (long long) blockPool->blockSize, SEEK_SET)) {
        perror("Warning: Unable to seek for reading file: ");
        return;
    }

    // note: length may be too large for read to handle...although unlikely
    if (-1 == read(fd, _data, (size_t) (blockPool->blockSize * _no_blocks))) {
        perror("Warning: Unable to read from cache file: " );
        return;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// write blocks, store the meta data (_oid) in file space use simple file name
void PCacheRawAccessor::blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                         long long _no_blocks, char* _data) {
    // move the file descriptor
    if (-1 == lseek64(fd, (long long) _start_block_num * (long long) blockPool->blockSize, SEEK_SET)) {
        perror("Warning: Unable to seek for writing file: ");
        return;
    }

    // note: length may be too large for write to handle...although unlikely
    if (-1 == write(fd, _data, blockPool->blockSize * _no_blocks)) {
        perror("Warning: Unable to write to cache file: " );
        return;
    }
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// remove blocks, remove a file or a few associated blocks
void PCacheRawAccessor::blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                      long long _no_blocks) {
    // need not do anything...nice nice.
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// *******************************************************************************
// 99) DummyAccessor do not do anything, and is only used in simulations where content
//     is not important.
DummyAccessor::DummyAccessor() :
    BlockAccessor()
{
}

// not called. put it here first
DummyAccessor::~DummyAccessor() {
}

// read blocks, store the meta data (_oid) in file space use simple file name
void DummyAccessor::blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks, char* _data) {
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// write blocks, store the meta data (_oid) in file space use simple file name
void DummyAccessor::blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                         long long _no_blocks, char* _data) {
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}

// remove blocks, remove a file or a few associated blocks
void DummyAccessor::blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                      long long _no_blocks) {
    // check whether the first time stamp is required. we do not do this in the function
    // to reduce the number of function calls
    if (false == time_stamp) {
        sync();
        mark_time_stamp();
    }
}


