#ifndef __BLOCKACCESSOR_H
#define __BLOCKACCESSOR_H

#include <string>
#include <list>

// abstrats the disk access
// supports disk read and write 
class BlockAccessor {
    protected:
        // the following two variables are read from confMap
        std::string root_dir; // the base directory for file-system based accessors
        std::string dev_name;  // the device name for raw disk based accessors
        static bool time_stamp;  // have we written the timestamp of first disk access
                                 // to disk.log?
        static void mark_time_stamp();  // writes a time stamp to disk.log to indicate
                                         // the time of first disk access
    public:
        BlockAccessor();
        virtual ~BlockAccessor();
        // read blocks, concrete subclass may only use a subset of arguments
        virtual void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data) = 0;
        // write blocks, concrete subclass may only use a subset of arguments
        virtual void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data) = 0;
        // delete a file or a few blocks assigned to that file
        virtual void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks) = 0;
};

class FlatAccessor : public BlockAccessor {
    private:

    public:
        FlatAccessor();
        void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data);
        void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data);
        void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks);
        std::string get_name(std::string _oid, long long _offset);
};

// SquidAccessor simulate 2 level squid approach: 16 first level, 256 second
// level (in each first level directory).
class SquidAccessor : public BlockAccessor {
    private:

    public:
        SquidAccessor();
        void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data);
        void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data);
        void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks);
        // the following three function returns first, second level directory names
        // and then filename
        std::string get_1_dir_name(std::string _oid, long long _offset);
        std::string get_2_dir_name(std::string _oid, long long _offset);
        std::string get_filename(std::string _oid, long long _offset);
};

// OneLevelAccessor creates a directory for every object (note, for BitTorrent
// a directory for every piece). It is analogous to saving web objects from the
// same web site inside one directory.
class OneLevelAccessor : public BlockAccessor {
    private:

    public:
        OneLevelAccessor();
        void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data);
        void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data);
        void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks);
        std::string get_filename(std::string _oid, long long _offset);
};

// PCacheFs uses only one file to save all data. Meta-data are stored
// in-memory. Disk accesses are instructed using block numbers with llseek
// calls. 
class PCacheFsAccessor : public BlockAccessor {
    private:
        int fd;     // the single file descriptor 

    public:
        PCacheFsAccessor();
        ~PCacheFsAccessor();
        void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data);
        void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data);
        void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks);
};

// PCacheRaw uses raw partition save all data. Meta-data are stored
// in-memory. Disk accesses are instructed using block numbers with llseek
// calls. 
class PCacheRawAccessor : public BlockAccessor {
    private:
        int fd;     // the single file descriptor 

    public:
        PCacheRawAccessor();
        ~PCacheRawAccessor();
        void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data);
        void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data);
        void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks);
};

// DummyAccessor do not do anything, and is only used in simulations where content
// is not important.
class DummyAccessor : public BlockAccessor {
    private:
    public:
        DummyAccessor();
        ~DummyAccessor();
        void blk_read(std::string _oid, long long _offset, long long _start_block_num, 
                          long long _no_blocks, char* _data);
        void blk_write(std::string _oid, long long _offset, long long _start_block_num, 
                           long long _no_blocks, char* _data);
        void blk_remove(std::string _oid, long long _offset, long long _start_block_num, 
                        long long _no_blocks);
};


#endif // __BLOCKACCESSOR_H
