#include <iostream>
#include <sstream>
#include <string>
#include <algorithm>
#include <cmath>
#include "pCache.h"  // for confMap
#include "BlockPool.h"

// represent a chunck of contigous blocks
BlockCluster::BlockCluster(long long _start_num, long long _no_blocks, bool _busyBit) :
    start_num(_start_num), no_blocks(_no_blocks), busyBit(_busyBit), pc(NULL),
    first_pc(NULL), next_pc(NULL)
{

}

std::string BlockCluster::toString(bool no_pc) {
    std::string ret("");
    ret.append("Start_num: ").append(Util::stringify(start_num));
    ret.append(" No_blocks: ").append(Util::stringify(no_blocks));
    ret.append(" Busy_bit: ").append(Util::stringify(busyBit));
    if (no_pc == false) {
        ret.append(" Pc: ").append(pc == NULL ? "null" : pc->toString());
        ret.append(" First_pc: ").append(first_pc == NULL ? "null" : first_pc->toString());
        ret.append(" Next_pc: ").append(next_pc == NULL ? "null" : next_pc->toString());
    }
    return ret;
}

// we use left_offset to sort them
bool BlockCluster::operator< (const BlockCluster &other) const{
    return start_num < other.start_num;
}

// BlockPool maintains a list of disk blocks
BlockPool::BlockPool() :
    fit_method(1), last_startnum(0), next_startnum(0), next_freenum(0), free_block_counter(0), accesorType(1)
{
    // for ascii to integer conversion
    std::stringstream numBlocksStr(confMap[std::string("BLOCK_NUM")]);
    numBlocksStr >> numBlocks;
    std::stringstream blockSizeStr(confMap[std::string("BLOCK_SIZE")]);
    blockSizeStr >> blockSize;
    // std::cout << "block num/size = " << numBlocks << "/" << blockSize << std::endl;
    std::stringstream accesorTypeStr(confMap[std::string("ACCESOR_TYPE")]);
    accesorTypeStr >> accesorType;
    std::stringstream fitMethodStr(confMap[std::string("FIT_METHOD")]);
    fitMethodStr >> fit_method;
    // set the total number of blocks
    free_block_counter = numBlocks;

    // update the stats module
    stats->disk_block_size = blockSize;
    stats->disk_block_number = numBlocks;

    // create the block accesor
    switch (accesorType) {
        case 1:
            accessor = new FlatAccessor();
            stats->access_type = "Flat Accessor";
            break;
        case 2:
            accessor = new SquidAccessor();
            stats->access_type = "Squid Accessor";
            break;
        case 3:
            accessor = new OneLevelAccessor();
            stats->access_type = "One-level Accessor";
            break;
        case 4:
            accessor = new PCacheFsAccessor();
            stats->access_type = "pCache/Fs Accessor";
            break;
        case 5:
            accessor = new PCacheRawAccessor();
            stats->access_type = "pCache/Raw Accessor";
            break;
        case 6:
            // we reuse the raw accessor for COSS
            accessor = new PCacheRawAccessor();  
            stats->access_type = "COSS Accessor";
            fit_method = 4;   // force to use COSS fit
            break;
        case 99:
            accessor = new DummyAccessor();
            stats->access_type = "Dummy Accessor";
            break;
    }
    
    switch(fit_method) {
        case 1:     // first fit
            stats->fit_method = "First Fit";
            break;
        case 2:     // next fit
            stats->fit_method = "Next Fit";
            break;
        case 3:     // clustering fit
            stats->fit_method = "Object-aware Fit";
            break;
        case 4:     // COSS fit
            stats->fit_method = "COSS Fit";
            break;
    }
 
    // add the first block cluster
    free_blocks.insert(new BlockCluster(0, numBlocks, false));  // all blocks, not occupied
}

// TODO: who will call this to close the accesor? 
BlockPool::~BlockPool() {
    if (accessor != NULL) delete accessor;
}

// This class helps to find the first fit empty block
class block_first_fit {
    private:
        // how many contigous blocks is required
        long long req_no_block;   

    public:
        // intiialize the value of req_blocks
        block_first_fit(long long _req_no_block) :
            req_no_block(_req_no_block) {
        }

    // return true if found
    bool operator() (BlockCluster* _block) {
        return (_block->no_blocks >= req_no_block && _block->busyBit == false); 
    }
};

// used by COSS, to find a busy block cluster that is closest to 
// (should be equal to) a particular block number
class block_first_coss_fit {
    private:
        // how many contigous blocks is required
        long long req_no_block;    // how many we need
        long long req_block_num;   // next block number

    public:
        // intiialize the value of req_blocks
        block_first_coss_fit(long long _req_no_block, long long _req_block_num) :
            req_no_block(_req_no_block), req_block_num(_req_block_num) {
        }

    // return true if found
    bool operator() (BlockCluster* _block) {
        // first case, if the free block is a super set of it
        // second case, if the free block's startnum is >= the next block number
        //              and the number of blocks there is enough
        // No, strict COSS should not have the second case...
        return ( _block->no_blocks + _block->start_num >= req_no_block + req_block_num &&
                 _block->start_num <= req_block_num);
        // sec case, (_block->start_num >= req_block_num && _block->no_blocks >= req_no_block));
    }
};

// allocate disk blocks for _length (bytes)
BlockCluster* BlockPool::allocate(long long _length, std::string _oid, PageCluster *_owner) {
    // how many blocks do we need?
    long long _no_blocks = (long long)ceil((float) _length / (float) blockSize);
    // marker for large enough free block cluster
    std::set<BlockCluster*>::iterator _old_pos; 
    // select the search ranges based on fit_method
    // all searches are divided into two steps (parts)
    std::set<BlockCluster*>::iterator _first_start;
    std::set<BlockCluster*>::iterator _first_end;
    std::set<BlockCluster*>::iterator _second_start;
    std::set<BlockCluster*>::iterator _second_end;

    // should we skip the block allocation algorithm?
    if (accesorType == 1 || accesorType == 2 || accesorType == 3) {
        // if contigous blocks are not important, we should let file systems
        // handle fragementations...
        free_block_counter -= _no_blocks;
        return new BlockCluster(0, _no_blocks, 1);
    }

    // handle the COSS as a special case here
    // The idea is, keep a block number counter for the next target victim. 
    // Whenever there is an allocation request, do next fit on free list first, if
    // any, return as next-fit method
    // Otherwise, find the victim using the next block number, and call abort explicitly
    // to free that page cluster, then update the next block number.
    // After that, try to do next fit again, if failed, explicitly free the other page cluster
    if (accesorType == 6) {
        while (true) {
#if 0
            std::cout << "Requesting no. blocks: " << _no_blocks <<
                         " COSS next free number: " << next_freenum << 
                         " last start number: " << last_startnum << 
                         " next start number: " << next_startnum << std::endl;
#endif

            // next fit - note that, if there is any empty blocks less than free_blocks, 
            //            we can not use it because of the cyclic nature!
            _old_pos = find_if(free_blocks.begin(), free_blocks.end(), block_first_coss_fit(_no_blocks, next_startnum));
            if (_old_pos != free_blocks.end()) { // we have found a large enough block
#if 0
                std::cout << "Allocating " << _no_blocks << " from block: " << (*_old_pos)->toString(true).c_str() << std::endl;
#endif
                goto final;     // about to return
            }
            // otherwise, need to free something
            if (next_freenum == 0) {
                std::cout << "COSS wrap around next start number: " << 
                             Util::stringify(next_startnum) << std::endl;
                assert(next_startnum > 0);
                next_startnum = 0;  // wrap around the next start number
            }
            std::set<BlockCluster*>::iterator _victim; 
            BlockCluster *bc = new BlockCluster(next_freenum, 0, false);
            _victim = busy_blocks.lower_bound(bc);    // find the one that is not less than next_freenum!
            delete bc;
            if (_victim == busy_blocks.end()) {
                // wrap around.
                std::cout << "COSS wrap around next free number: " << 
                             Util::stringify(next_freenum) << std::endl;
                assert(next_freenum > 0);
                next_freenum = 0; 
                continue;
            }
            // free it
            next_freenum = (*_victim)->start_num + (*_victim)->no_blocks;
#if 0
            std::cout << "Freeing block cluster: " << (*_victim)->toString(true).c_str() << std::endl;
#endif
            mCache->abort((*_victim)->pc);  // it looks fine to me to use abort(...) here!
                                            // note, victim here is the block cluster, and pc is its owner
            // then try to find first fit again!
        } 
    }

    // find the first one that is not busy
    bool found; 
    BlockCluster *bc;
    do {
		switch(fit_method) {
			case 1:     // first fit
				_first_start = free_blocks.begin();
				_first_end = free_blocks.end();
				_second_start = free_blocks.begin();
				_second_end = free_blocks.begin();
				break;
			case 2:     // next fit
		        bc = new BlockCluster(last_startnum, 0, false);
				_first_start = free_blocks.lower_bound(bc);  
		        delete bc;
				_first_end = free_blocks.end();
				_second_start = free_blocks.begin();
				_second_end = _first_start;
				break;
			case 3:     // clustering fit
				Item* _item = cacheManager->getItem(_oid);
                // find the smallest start_num
                long long _smallest_start_num = 0;
                if (_item != NULL) {
                    // set the default base on some hash on obj id 
                    // this is to equally distribute the first block all over the disk
                    // The following idea doesn't seem to work
                    _smallest_start_num = __gnu_cxx::__stl_hash_string(_item->objId.c_str()) % numBlocks; 
                    // go thru the list and find anything smaller that the one..
				    std::set<PageCluster*, DereferenceLess>::iterator _anypc;	
                    for (_anypc = _item->pcList.begin(); _anypc != _item->pcList.end(); ++_anypc) {
					    if ((*_anypc)->blocks != NULL) {
					        if ((*_anypc)->blocks->start_num < _smallest_start_num) {
                                _smallest_start_num = (*_anypc)->blocks->start_num;
                            }
                        }
                    }
		            bc = new BlockCluster(_smallest_start_num, 0, false);
					_first_start = free_blocks.lower_bound(bc);  
                    delete bc;
                }
                else {
                    // this is a saftu fallback, should not happen at all
					_first_start = free_blocks.begin();   // to first-fit
				}
				_first_end = free_blocks.end();
				_second_start = free_blocks.begin();
				_second_end = _first_start;
				break;
		}

        found = false;
        // first step
        _old_pos = find_if(_first_start, _first_end, block_first_fit(_no_blocks));
        if (_old_pos == _first_end) {
            _old_pos = find_if(_second_start, _second_end, block_first_fit(_no_blocks));
            if (_old_pos == _second_end) {  // we failed both, free something
                // evict objects from disk to make room 
                mCache->free_init();
                PageCluster* victim_pc; 
                do {
                    victim_pc = mCache->getFreeCandidate();
                } while (victim_pc->isSwappingOut() == true);
                if (victim_pc) {
                    mCache->freeCurrent();
                }
            }
            else {
                found = true;
            }
        }
        else {
            found = true;
        }
    }
    while (found != true);  // don't stop until find one

final:   // used by the COSS filesystem
    // get a hold of old block
    BlockCluster *_old_block = *_old_pos;   
    if (_old_block->no_blocks == _no_blocks) {
        // perfect match
        _old_block->busyBit = true;  // mark the bit, and we are done 
        // before returning, we detach this block from free block list
        free_blocks.erase(_old_pos);
        last_startnum = _old_block->start_num;
        next_startnum = _old_block->start_num + _old_block->no_blocks;
        _old_block->pc = _owner;
        if (accesorType == 6) {
            busy_blocks.insert(_old_block);   // added to the busy list (used by COSS)
        }
        sanity();
        return _old_block;
    }
    else {
        if (accesorType != 6) { 
            // split the block into two parts 
            BlockCluster *_new_block = new BlockCluster(_old_block->start_num,  
                                                        _no_blocks, true); // create a new busy block
            // remove and re-insert the old block
            free_blocks.erase(_old_pos);
            _old_block->start_num += _no_blocks;    // update the old block
            _old_block->no_blocks -= _no_blocks;  
            free_blocks.insert(_old_block);
            last_startnum = _new_block->start_num;
            next_startnum = _new_block->start_num + _new_block->no_blocks;
            _new_block->pc = _owner;
            return _new_block;
        }
        else { // for coss, we have to cut this block on next_freenum, 
               // which means we may create two free blocks and one busy block
            long long tmp_upper_bound = _old_block->start_num + _old_block->no_blocks;
            if (_old_block->start_num < next_startnum) {
                // create a block from _old_block->start_num to next_startnum 
                free_blocks.erase(_old_pos);
                _old_block->no_blocks = next_startnum- _old_block->start_num;
                free_blocks.insert(_old_block);
            }
            else {
                // just drop the old free blocks
                free_blocks.erase(_old_pos);
            }
            // second free part of the block
            // is this a valid one?
            if (next_startnum + _no_blocks < numBlocks && 
                tmp_upper_bound - (next_startnum + _no_blocks) > 0) {

                BlockCluster *_new_free_blocks = new BlockCluster(next_startnum + _no_blocks, 
                                                                  tmp_upper_bound - (next_startnum + _no_blocks), 
                                                                  false);  
                free_blocks.insert(_new_free_blocks);
            }

            // create a new busy block
            BlockCluster *_new_block = new BlockCluster(next_startnum, _no_blocks, true);
            last_startnum = _new_block->start_num;
            next_startnum = _new_block->start_num + _new_block->no_blocks;
            _new_block->pc = _owner;
            busy_blocks.insert(_new_block);   // added to the busy list (used by COSS)
            sanity();
            return _new_block;
        }    
    }
}

// return disk blocks, we first find the position of the block, merge left, and
// then merge right
void BlockPool::deallocate(BlockCluster* _block) {
    // busyBit should be 1
    assert(1 == _block->busyBit);

    // should we skip the block allocation algorithm?
    if (accesorType == 1 || accesorType == 2 || accesorType == 3) {
        // if contigous blocks are not important, we should let file systems
        // handle fragementations...
        free_block_counter += _block->no_blocks;
        delete _block;
        return;
    }
    if (accesorType == 6) {
        busy_blocks.erase(_block);   // remove from the busy list (used by COSS)
    }
    // this block becomes free, 
    _block->busyBit = false;
    _block->pc = NULL;
    if (_block->first_pc != NULL) {  // break the chain...
        PageCluster *_tmp_pc = _block->first_pc;
        while(_tmp_pc != NULL && _tmp_pc->blocks != NULL) {
            _tmp_pc->blocks->first_pc = NULL; 
            PageCluster *_tmp_pc2 = _tmp_pc->blocks->next_pc;
            _tmp_pc->blocks->next_pc = NULL;
            _tmp_pc = _tmp_pc2;
        }
    }
    _block->first_pc = NULL;
    _block->next_pc = NULL;
    // we should re-insert it into the free set
    std::pair<std::set<BlockCluster*>::iterator, bool> result;
    result = free_blocks.insert(_block);
    std::set<BlockCluster*>::iterator _pos = result.first;
    // Next, we compact the block list... 
    // Note: _pos points to the newly returned block
    // merge toward left
    std::set<BlockCluster*>::iterator _left_pos = _pos;
    bool _reinsert = false;
    while (_left_pos != free_blocks.begin()) {
        long long _prev_left_start_num = (*_left_pos)->start_num; 
        --_left_pos;  // move to left
        assert((*_left_pos)->busyBit == false);  // if it's in use, then abort
        if ((*_left_pos)->start_num + (*_left_pos)->no_blocks == _prev_left_start_num) {
            // can merge with it
            _block->start_num = (*_left_pos)->start_num;
            _block->no_blocks += (*_left_pos)->no_blocks;
            BlockCluster *_old_block = (*_left_pos);   // get a hold of the block that is going to be freed
            free_blocks.erase(_left_pos);  // remove the merged block, and return pos of next element
            _left_pos = _pos;              // reset the pos
            delete _old_block;                  // free it
            _reinsert = true;
        }
    }
    if (_reinsert == true) {
        // re-insert the _block, since we have changed the start_num
        free_blocks.erase(_block);
        result = free_blocks.insert(_block);
        _pos = result.first;
        _reinsert = false;
        
    }
    // merge toward right
    std::set<BlockCluster*>::iterator _right_pos = _pos;
    while (true) {  // break when _right_pos reaches end
        long long _prev_right_start_num_plus_no_blocks = (*_right_pos)->start_num + (*_right_pos)->no_blocks; 
        ++_right_pos;  // move to right
        if (_right_pos == free_blocks.end()) break;
        assert((*_right_pos)->busyBit == false);
        if ((*_right_pos)->start_num == _prev_right_start_num_plus_no_blocks) {
            // can merge with it
            _block->no_blocks += (*_right_pos)->no_blocks;
            BlockCluster *_old_block = (*_right_pos);   // get a hold of the block that is going to be freed
            free_blocks.erase(_right_pos); // remove the merged block, and return pos of next element
            _right_pos = _pos;              // reset the pos
            delete _old_block;                   // free it
       }
    }
    // sanity check!    
    sanity();
}

long long BlockPool::file_size(long long _length) {
    long long _no_blocks = (long long)ceil((float) _length / (float) blockSize);
    return ((long long)_no_blocks) * ((long long) blockSize);
}

std::string BlockPool::toString() {
    std::string ret("");
    ret.append("Num_blocks: ").append(Util::stringify(numBlocks));
    ret.append(" Block_size: ").append(Util::stringify(blockSize)); 
    if (accesorType == 1 || accesorType == 2 || accesorType == 3) {
        ret.append(" Free_block_num:").append(Util::stringify(free_block_counter));
    }
    else {  // pCacahe and COSS methods
        ret.append(" Free_blocks: | ");
        std::set<BlockCluster*>::iterator _pos;
        for (_pos = free_blocks.begin(); _pos != free_blocks.end(); ++_pos) {
            ret.append((*_pos)->toString()).append(" | ");
        }
        ret.append(" Busy_blocks: | ");
        for (_pos = busy_blocks.begin(); _pos != busy_blocks.end(); ++_pos) {
            ret.append((*_pos)->toString()).append(" | ");
        }
    }
    ret.append(Util::CR_str);
    return ret;
}

// sanity check routine
void BlockPool::sanity() {
    std::set<BlockCluster*>::iterator _pos;
    std::set<BlockCluster*>::iterator _next_pos;
    for (_pos = free_blocks.begin(); _pos != free_blocks.end(); ++_pos) {
        _next_pos = _pos;
        if ( ++_next_pos != free_blocks.end()) {
            assert((*_pos)->pc == NULL);
            assert((*_pos)->first_pc == NULL);
            assert((*_pos)->next_pc == NULL);
            if ((*_pos)->start_num + (*_pos)->no_blocks > (*_next_pos)->start_num) {
                std::cout << "Sanity check failed (1), current block: " << (*_pos)->toString() << " next block: " << (*_next_pos)->toString() << std::endl;
                std::cout << "Dumping disk blocks: " << toString() << std::endl;
                assert(0);
            }

            // cross check the busy blocks part
		    BlockCluster *bc = new BlockCluster((*_pos)->start_num, 0, false);
            std::set<BlockCluster*>::iterator _right = busy_blocks.lower_bound(bc);   // the one on its right
		    delete bc;
            if (_right != busy_blocks.end()) {
                // if they are not completely disjoint, we failed
                if (!((*_pos)->start_num >= (*_right)->start_num + (*_right)->no_blocks ||
                    (*_right)->start_num >= (*_pos)->start_num + (*_pos)->no_blocks)) {
                    std::cout << "Sanity check failed (3), current free block: " << (*_pos)->toString() << " conflict busy block: " << (*_right)->toString() << std::endl;
                    std::cout << "Dumping disk blocks: " << toString() << std::endl;
                    assert(0);
                }
                --_right;
                if (_right != busy_blocks.end()) {
                // if they are not completely disjoint, we failed
                    if (!((*_pos)->start_num >= (*_right)->start_num + (*_right)->no_blocks ||
                        (*_right)->start_num >= (*_pos)->start_num + (*_pos)->no_blocks)) {
                        std::cout << "Sanity check failed (4), current free block: " << (*_pos)->toString() << " conflict busy block: " << (*_right)->toString() << std::endl;
                        std::cout << "Dumping disk blocks: " << toString() << std::endl;
                        assert(0);
                    }
                }
            }
        }
    }
    for (_pos = busy_blocks.begin(); _pos != busy_blocks.end(); ++_pos) {
        _next_pos = _pos;
        if ( ++_next_pos != busy_blocks.end()) {
            if ((*_pos)->start_num + (*_pos)->no_blocks > (*_next_pos)->start_num) {
                std::cout << "Sanity check failed (2), current block: " << (*_pos)->toString() << " next block: " << (*_next_pos)->toString() << std::endl;
                std::cout << "Dumping disk blocks: " << toString() << std::endl;
                assert(0);
            }
        }
    }
}

// The following function is called to get the PMF of free block cluster size
// THe unit is in no. block
// This function is typically called  once before exiting
//
// This function only works when disk access method is one of the pCache ones....
std::map<long long, long long> BlockPool::get_bc_pmf() {
    std::map<long long, long long> ret;
    if (accesorType == 1 || accesorType == 2 || accesorType == 3) {
        // abort....
        return ret;
    }
    std::set<BlockCluster*>::iterator _pos;
    for (_pos = free_blocks.begin(); _pos != free_blocks.end(); ++_pos) {
        ret[(*_pos)->no_blocks]++;
    }
    return ret;
}

