#ifndef __BLOCKPOOL_H
#define __BLOCKPOOL_H

#include <set>
#include "Util.h"
#include "BlockAccessor.h"

class PageCluster;

// represent a chunck of contigous blocks
class BlockCluster {
    private:

    public:
        long long start_num;    // starting block number
        long long no_blocks;    // lenth of the contiguous block (in No. blocks)
        bool busyBit;           // used by any page cluster?
        PageCluster *pc;        // the page cluster who owns us, used by COSS
        PageCluster *first_pc;  // points to the first page cluster in a chain, 
        PageCluster *next_pc;   // points to the next page cluster in a chain, 
                                // which consists of all page clusters that were 
                                // written to the disk in one (buffered) operation 

        BlockCluster(long long _start_num, long long _no_blocks, bool _busyBit);
        std::string toString(bool no_pc = false);
        // we define less operator for sorted set
        bool operator< (const BlockCluster &other) const;
};

// a list of disk blocks to support first-fit allocation
class BlockPool {
    private:
	    std::set<BlockCluster*, DereferenceLess> free_blocks;  // list of all free block clusters
        // busy blocks is only used by COSS file system, but not others
	    std::set<BlockCluster*, DereferenceLess> busy_blocks;  // list of all busy block clusters
        int fit_method;    // 1 <-- first fit, 2 <-- next fit, 3 <-- clustering fit
        long long last_startnum;   // last start number
        long long next_startnum;   // next start number
        long long next_freenum;    // next block number to be freed (used by COSS)
        void sanity();     // sanity check routine, system exits if failed..
        long long free_block_counter;   // free_block_counter is used for non-pCache access methods! 

    public:
        int accesorType;   // type of block accesor
        BlockPool();
        ~BlockPool();

        // XXX super blocks are not implemented yet
        long long numBlocks;   // total number of blocks
        long long blockSize;   // size (in bytes) of each disk block        

        BlockAccessor* accessor;  // how to write to and read from the disk

        BlockCluster* allocate(long long _length, std::string _oid, PageCluster *_owner);  // allocate disk blocks for _length (bytes)
        void deallocate(BlockCluster* _block);    // return disk blocks
        long long file_size(long long _length);  // compute the actual file size 
                                                 // (round-it up using block size)
        std::string toString();
        // The following function is called to get the PMF of free block cluster size
        // THe unit is in no. block
        // This function is typically called  once before exiting
        std::map<long long, long long> get_bc_pmf();
};

#endif // __BLOCKPOOL_H
