#include "Cache.h"
#include <iostream>
#include "CacheManager.h"
#include "pCache.h"

//	Default constructor
//	ALL replacement policies should call this constructor first
Cache::Cache() :
    size(0)
{
	//	Read the cache capaticy from config file.
	//	Total cache capacity is equal to block_size times the number of blocks on disk
	capacity = (long long)(atoi(confMap[std::string("BLOCK_SIZE")].c_str())) * (long long)(atoi(confMap[std::string("BLOCK_NUM")].c_str()));
    float h_watermark = atof(confMap[std::string("MAX_FILLED_SIZE")].c_str());
    float l_watermark = atof(confMap[std::string("MIN_FILLED_SIZE")].c_str());
    // for COSS we set extreme high h_watermark
    if (blockPool->accesorType == 6) {
        h_watermark = 1.0f;     // choose a > 1.0 value to completely avoid swapout
        l_watermark = 1.0f;
    }
    max_filled_size = (long long)(h_watermark * capacity);
    min_filled_size = (long long)(l_watermark * capacity);
    // update UI
    stats->high_watermark = h_watermark;
    stats->low_watermark = l_watermark;

    std::cout << "Initializing cache module with capacity: " << Util::stringify(capacity) << 
                 " high watermark: " << Util::stringify(max_filled_size) << 
                 " and low watermark: " << Util::stringify(min_filled_size) << std::endl;

	pthread_mutex_init(&qmut, NULL);
}

Cache::~Cache()
{
	pthread_mutex_destroy(&qmut);
}

CacheResponse* Cache::query(Segment* seg)
{
	//	update the total traffic log
	stats->totalTraff += seg->rRange - seg->lRange;

	//stats->lprintf(Stats::INFO, "Cache: Query Received for %s", seg->toString().c_str());

	//	MUST be initialize this way
	long long lRange = seg->lRange, rRange = seg->rRange, availableSize = 0;

	PageCluster* newPC = NULL;

	std::list<PageCluster*>::iterator itr, removeItr;
	CacheResponse* mResponse = new CacheResponse();
    mResponse->requestedSegment = new std::list<PageCluster*>();

	//	remove the Item from the cache queue (if exist)
	//	this function should be overide in sub classes
	pthread_mutex_lock(&qmut);

	mResponse->availableSegment = cacheManager->lookup(seg->objId, seg);

	//stats->lprintf(Stats::INFO, "Cache: No. of available PC within the requested Seg: %d", mResponse->availableSegment->size());

	//	if any PageCluster is available on the cache
	if (!mResponse->availableSegment->empty()) {
		itr = mResponse->availableSegment->begin();

		if ((*itr)->getLeft() <= seg->lRange) {
			lRange = (*itr)->getRight();

			removeItr = itr;
			itr++;

		    //	if this PC is being modified by other thread
			//	means its data is not available for download
			if ((*removeItr)->isNew()) {
				mResponse->availableSegment->erase(removeItr);
				stats->memTraff -= std::min(seg->rRange, (*removeItr)->getRight()) - 
                                   std::max(seg->lRange, (*removeItr)->getLeft());  // undo the memory hit record
                (*removeItr)->decRefCnt();   // undo the ref cnt increment.
            }
			else
				availableSize += std::min((*removeItr)->getRight(), seg->rRange) - std::max((*removeItr)->getLeft(), seg->lRange);

		}
	}

	Segment *tmpSeg;
	while (lRange < seg->rRange) {
		if (mResponse->availableSegment->empty() || itr == mResponse->availableSegment->end()) {
			rRange = seg->rRange;
		} else {
			rRange = (*itr)->getLeft();
		}
		
		//	if we have two consequent page cluster without any space between them...
		if (rRange <= lRange) {
			lRange = (*itr)->getRight();
			removeItr = itr;
			itr++;

		    //	if this PC is being modified by other thread
			//	means its data is not available for download
			if ((*removeItr)->isNew()) {
				mResponse->availableSegment->erase(removeItr);
				stats->memTraff -= std::min(seg->rRange, (*removeItr)->getRight()) - 
                                   std::max(seg->lRange, (*removeItr)->getLeft());  // undo the memory hit record
                (*removeItr)->decRefCnt();   // undo the ref cnt increment.
            }
			else
				availableSize += std::min((*removeItr)->getRight(), seg->rRange) - std::max((*removeItr)->getLeft(), seg->lRange);
			continue;
		}

        // admission policy starts here
        long long admitted_length = admit(seg);
        if (admitted_length > 0) {
            //	if the queue is full, then we first need to free some space in the queue
            if ((size + blockPool->file_size(std::min(rRange - lRange, admitted_length))) > max_filled_size)
                free(false);
    
            //	create and insert the new PageCluster into the objTable
            tmpSeg = new Segment(seg->objId, lRange, std::min(lRange + admitted_length , rRange));
            newPC = cacheManager->insert(seg->objId, tmpSeg);
            delete tmpSeg;
    
            //	if the new page cluster is created successfully
            if (newPC != NULL) {
                mResponse->requestedSegment->push_back(newPC);
                // item->size is updated in CacheManager using Cache->addSize()
            } else {
                std::cout << "newPC = NULL" << std::endl;
            }
        }
        // admission policy ends here

		if (mResponse->availableSegment->empty() || itr == mResponse->availableSegment->end())
			lRange = rRange;
		else {
			lRange = (*itr)->getRight();

			removeItr = itr;
			itr++;

		    //	if this PC is being modified by other thread
			//	means its data is not available for download
			if ((*removeItr)->isNew()) {
				mResponse->availableSegment->erase(removeItr);
				stats->memTraff -= std::min(seg->rRange, (*removeItr)->getRight()) - 
                                   std::max(seg->lRange, (*removeItr)->getLeft());  // undo the memory hit record
                (*removeItr)->decRefCnt();   // undo the ref cnt increment.
            }
			else
				availableSize += std::min((*removeItr)->getRight(), seg->rRange) - std::max((*removeItr)->getLeft(), seg->lRange);
		}
	}

    // update frequency
    if (policy() == P2P) { 
        for (itr = mResponse->availableSegment->begin(); itr != mResponse->availableSegment->end(); itr++) {
            updateFreq(*itr, 
                       availableSize, 
                       std::min((*itr)->getRight(), seg->rRange) - std::max((*itr)->getLeft(), seg->lRange)
                       );
            (*itr)->refresh();     // update the LRU queue for page memory
        }
        for (itr = mResponse->requestedSegment->begin(); itr != mResponse->requestedSegment->end(); itr++) {
            updateFreq(*itr, 
                       availableSize, 
                       std::min((*itr)->getRight(), seg->rRange) - std::max((*itr)->getLeft(), seg->lRange)
                      );
            (*itr)->refresh();     // update the LRU queue for page memory
        }    
    }

    // old page clusters
	for (itr = mResponse->availableSegment->begin(); itr != mResponse->availableSegment->end(); itr++) {
        if (item_based() == true) {
            // pop a page cluster from the queue
            removeItem(*itr);
            // push a page cluster to top of the queue
            insertItem(*itr, 
                       availableSize, 
                       std::min((*itr)->getRight(), seg->rRange) - std::max((*itr)->getLeft(), seg->lRange)
                       );
            break;
        }
        else {
            // Non item based algorithm (currently only LRU) does not support safe-removal
            // That is, it does not do a (linear) lookup to make sure the victim object does exist.
            // This linear lookup consumes too much CPU power, especially when queue gets long
            // pop a page cluster from the queue
            removeItem(*itr);
            // push a page cluster to top of the queue
            insertItem(*itr, 
                       availableSize, 
                       std::min((*itr)->getRight(), seg->rRange) - std::max((*itr)->getLeft(), seg->lRange)
                       );
        }
    }
    // new page clusters
	for (itr = mResponse->requestedSegment->begin(); itr != mResponse->requestedSegment->end(); itr++) {
        if (item_based() == true) {
            // pop a page cluster from the queue
            removeItem(*itr);
            // push a page cluster to top of the queue
            insertItem(*itr, 
                       availableSize, 
                       std::min((*itr)->getRight(), seg->rRange) - std::max((*itr)->getLeft(), seg->lRange)
                      );
            break;  // for item-based item, no need to do more than once
        }
        else {
            // Non item based algorithm (currently only LRU) does not support safe-removal
            // That is, it does not do a (linear) lookup to make sure the victim object does exist.
            // This linear lookup consumes too much CPU power, especially when queue gets long
            // push a page cluster to top of the queue
            insertItem(*itr, 
                       availableSize, 
                       std::min((*itr)->getRight(), seg->rRange) - std::max((*itr)->getLeft(), seg->lRange)
                       );
        }
    }

	//	update the byte hit rate log
	for (itr = mResponse->availableSegment->begin();
		itr != mResponse->availableSegment->end(); itr++) {
			stats->availTraff += std::min(seg->rRange, (*itr)->getRight()) -
						std::max(seg->lRange, (*itr)->getLeft());
	}

	//stats->lprintf(Stats::INFO, "Cache: response: %s", mResponse->toString().c_str());
	pthread_mutex_unlock(&qmut);

	return mResponse;
}

// by default, we admit all incoming segments. 
// Current GDS and P2P algorithms have a different admission policy
// They override this function
long long Cache::admit(Segment* _seg){
    return _seg->rRange - _seg->lRange;
}

// This function is used to find the first page cluser that is 
// not inused (with zero ref. cnt.)
bool victim_pc(PageCluster* _pc) {
    return (_pc->refCnt) <= 0;
}

//	Remove a PageCluster from the candidate Item in the queue
void Cache::free(bool force)
{
	//stats->lprintf(Stats::INFO, "Cache: Reach the high-watermark, Start evacuation...");

	if (size > max_filled_size || force == true) {
        //	lock the queue
        pthread_mutex_lock(&qmut);
		_free();
		//	unlock the queue
        pthread_mutex_unlock(&qmut);
    }
}

void Cache::abort(PageCluster* _pc)
{
    stats->lprintf(Stats::INFO, "Cache::abort(...) is called: %s\n", _pc->toString().c_str());

    _pc->decRefCnt();
	pthread_mutex_lock(&qmut);
	abortPC(_pc);
	pthread_mutex_unlock(&qmut);
}

//	Return TRUE if the cache is full
bool Cache::isFull()
{
	if (size >= capacity * max_filled_size)
		return TRUE;

	return FALSE;
}

//	update the cache utilized size
//	_length can be either positive or negetive
void Cache::addSize(long long _length)
{
	size += _length;
}

//	query the cache size
long long Cache::getSize()
{
	return size;
}

CacheResponse::CacheResponse()
{
}

CacheResponse::~CacheResponse() {
#ifdef ONLY_TO_SINGLE_THREAD_SIMU
    // sanity check - only applicable to single-thread simu
	std::list<PageCluster*>::iterator availableIter;
	for (availableIter = availableSegment->begin();
			availableIter != availableSegment->end(); availableIter++) {
        if ((*availableIter)->refCnt > 0) {
            std::cout << "Warning, refcnt > 0 (1): " << (*availableIter)->toString() << std::endl;
        }
    }
	std::list<PageCluster*>::iterator requestedIter;
	for (requestedIter = requestedSegment->begin();
			requestedIter != requestedSegment->end(); requestedIter++) {
        if ((*requestedIter)->refCnt > 0) {
            std::cout << "Warning, refcnt > 0 (2): " << (*requestedIter)->toString() << std::endl;
        }
    }
#endif // ONLY_TO_SINGLE_THREAD_SIMU

    // free the two lists, but not page clusters on them!
    if (availableSegment != NULL) delete availableSegment;
    if (requestedSegment != NULL) delete requestedSegment; 
}

long long Cache::getQueueSize(){return 0LL;}

// by default do nothing...
void Cache::updateFreq(PageCluster* _pc, long long serverSize, long long pcServedSize){
}

