#ifndef __CACHE_H
#define __CACHE_H

#include <deque>
#include <list>
#include <string>
#include <algorithm>
#include <pthread.h>
#include <cassert>

#include "Util.h"
#include "Segment.h"
#include "Item.h"
#include "Util.h"

extern bool victim_pc(PageCluster* _pc);	// This function is used to find the first page cluser that is 
											// not inused (with zero ref. cnt.)

class CacheResponse
{
public:
	CacheResponse();
    ~CacheResponse();

	std::list<PageCluster*>*	availableSegment;
	std::list<PageCluster*>*	requestedSegment;

    std::string toString() {
        std::string ret("Cache response: available=(");
        std::list<PageCluster*>::iterator avail_iter;
        for (avail_iter = availableSegment->begin(); avail_iter != availableSegment->end(); ++avail_iter) {
            ret.append("<").append(Util::stringify((*avail_iter)->getLeft())).append(", ").
                           append(Util::stringify((*avail_iter)->getRight())).append("> | ");
        }
        ret.append(")");
        ret.append("; requested=(");
        std::list<PageCluster*>::iterator req_iter;
        for (req_iter = requestedSegment->begin(); req_iter != requestedSegment->end(); ++req_iter) {
            ret.append("<").append(Util::stringify((*req_iter)->getLeft())).append(", ").
                           append(Util::stringify((*req_iter)->getRight())).append("> | ");
        }
        ret.append(")");
        return ret;
    }
};

class Cache
{
public:	
	Cache();									//	default constructor
	virtual ~Cache();
    bool isFull();
	void addSize(long long _length);			//	update the cache size
	long long getSize();						//	get the cache size

	virtual CacheResponse* query(Segment*);		//	receiving a request for an item
												//	receiving a request for a file range
												//	(the request response should be positive)
	void free(bool force);						//	remove an item or a PageCluster (depends on the policy)
												//	from the cache to make room for new items
	void abort(PageCluster* _pc);				//	abort a PageCluster download session

// free_init() and friend functions should not get the lock, actually, the callers should have
// already got the lock, otherwise we are facing serious race conditions.
	virtual void free_init() = 0;					//	initialize the candidate iterators
	virtual PageCluster* getFreeCandidate() = 0;	//	find a candidate PageCluster to be removed
	virtual void freeCurrent() = 0;					//	remove the current PageCluster which candidate
													//	iterators point to, from the cache
	virtual long long admit(Segment* _seg);         // called by query(), returns how many bytes we should admit 

	//	all these methods should be override in subclasses
	virtual void removeItem(PageCluster *_pc) = 0;		//	remove an Item knowing its objId from the queue
	virtual void insertItem(PageCluster* _pc, 
                            long long serverSize, 
                            long long pcServedSize) = 0; //	insert an Item into the queue, we also need to
														            //	pass the function the available portion of Item
	virtual void updateFreq(PageCluster* _pc, 
                            long long serverSize, 
                            long long pcServedSize);    //	for P2P to update gamma values
	
	virtual void abortPC(PageCluster* _pc) = 0;			//	abort a PageCluster download
	virtual void _free() = 0;							//	free the cache queue till reach the low watermark
    virtual bool item_based() = 0;  // true, if the replacement queue is based on objects (not page clusters)
                                // this is set in constructor, and for optimization purpose
    virtual int policy() = 0;   // set by the constructor of derived classes
    // supported replace policy
    const static int LRU = 1;
    const static int LFU = 2;
    const static int LSB = 3;
    const static int P2P = 4;
    const static int GDS = 4;
    const static int COSS = 5;
protected:
	pthread_mutex_t qmut;						//	Lock the Cache queue
	long long capacity;							//	Maximum size of the cache
	long long size;								//	Size of filled part of the cache
    long long max_filled_size;                  //  high watermark to trigger cache replacement policy
    long long min_filled_size;                  //  low water mark to stop cache replacement policy
	virtual long long getQueueSize();

};

#endif	//	__CACHE_H
