#include <iostream>
#include <algorithm>
#include <cmath>
#include "CacheManager.h"
#include "pCache.h"
#include "Util.h"
#include "Conn.h"
#include "pthread.h"

CacheManager::CacheManager(Cache* _mCache):
    mCache(_mCache), merge(false), split_copyover(false), max_merge(2000000)
{
    // Note: enable recursive, for insert->something in page cluster->abort
    pthread_mutexattr_t   attr;  
    pthread_mutexattr_init(&attr);  
    pthread_mutexattr_settype(&attr,PTHREAD_MUTEX_RECURSIVE);   
    pthread_mutex_init(&cmut, &attr);
    pthread_mutexattr_destroy(&attr);  
    if (0 == confMap[std::string("PC_MERGE")].compare("1")) {
        merge = true;
        stats->merge = true;  // update UI
    }
    else {
        stats->merge = false;  // update UI
    }
    if (0 == confMap[std::string("SPLIT_COPYOVER")].compare("1")) {
        split_copyover = true;
        stats->split_copyover = true;  // update UI
    }
    else {
        stats->split_copyover = false;  // update UI
    }
    std::stringstream max_merge_string(confMap[std::string("PC_MERGE_SIZE")]);
    max_merge_string >> max_merge;
}

CacheManager::~CacheManager(){
    pthread_mutex_destroy(&cmut);

}

// search for cached page cluster caller owns the returned list, and shall
// call delete list after it has done with it.
std::list<PageCluster*>* CacheManager::lookup(std::string _oid, Segment* _seg){
	pthread_mutex_lock(&cmut);
    cleanup();

    // There are 3 cases,
    // use pc_l, pc_r to represent the range of a page cluster
    // use s_l, s_r to denote the range of segment
    //
    // case 1: s_l < s_r <= pc_l < pc_r  return emptyset
    // case 2: pc_l < pc_r <= s_l < s_r  return emptyset
    // case 3 (else): [ max(s_l, pc_l), min(s_r, pc_r) )
    long long s_l = _seg->lRange;
    long long s_r = _seg->rRange;
    std::list<PageCluster*>* ret = new std::list<PageCluster*>();

restart:
    // find the object first
    __gnu_cxx::hash_map<std::string, Item*>::iterator _map = objTable.find(_oid);
    if (_map != objTable.end()) {  // found something
        Item* _item = _map->second;
        // check the page cluster list in _item,
        // this may not be OO, but save us some time on implementation
        // first, find the starting page cluster, we should look into
        // We use a dummy (fake) page cluster here, because pcList is a set of pcs
        // We might elimiate this overhead (complication) by changing pcList into a map in the future
        std::set<PageCluster*, DereferenceLess>::iterator _pc; 
        PageCluster *_dummy = new PageCluster(_oid, s_l, s_l); // note that pcs are sorted using offset.
        if (_dummy != NULL) {
            _pc = _item->pcList.upper_bound(_dummy); // points to the next
            if (_pc != _item->pcList.begin()) {
                --_pc;  // points to the first one we should look into
            }
            delete _dummy;
        }        
        else {
            _pc = _item->pcList.begin();  // fall back
        }
        bool has_seen_pcs = false;      // early quit mechanism
        for (; _pc != _item->pcList.end(); ++_pc) {
            long long pc_l = (*_pc)->getLeft();
            long long pc_r = (*_pc)->getRight();
			// only need to consider case 3...
            // if (pc_l >= s_l && pc_r <= s_r) { 
            if (!(s_r <= pc_l || pc_r <= s_l)) {   // if not totally disjoint
                // we have to increase this pc's reference count early, otherwise
                // the just allocated buffer pages may be recycled, becase of zero
                // refCnt
                (*_pc)->incRefCnt();
				if ((*_pc)->isNotInMem()) {
                    PageCluster *_tmp_pc = (*_pc);
                    if (false == _tmp_pc->chained_swapin()) {   //	copy the page from disk into the memory
                                                               //	because we are about to read the page cluster
                        // page cluster no longer exists delete it, this is bad, because we did change the
                        // data structure, so the iterator does not work anymore.
                        delete (_tmp_pc);
                        ret = new std::list<PageCluster*>();  // TODO: this leads to memory leak, but this only
                                                              //       happens to COSS
                        goto restart;  // restart the whole lookup
                    }                                                               
                }
				else {
					stats->memTraff += std::min(s_r, pc_r) - std::max(s_l, pc_l);		//	update the memory hit log
                }
                // one possibility is: we do not have enough pages
                // in that case, (*_pc)->data.empty() is true.
                // When that happens, we remove this pc from the available list
                if (false == (*_pc)->isNotInMem()) {
                    ret->push_back(*_pc);  // do get data
                    // for P2P, we have to update the freq first before refresh, we do that in P2P class
                    if (mCache->policy() != Cache::P2P) { 
                        (*_pc)->refresh();     // update the LRU queue for page memory
                    }
                }
                else { // unlikely
                    std::cout << "Unable to allocate pages for page cluster: " << 
                                 (*_pc)->toString() << std::endl;
                    (*_pc)->decRefCnt();  // undo the ref inc...
                }
                has_seen_pcs = true;
            }
            else {
                if (has_seen_pcs == true) break;  // we can abort because pcList is sorted by offsets
            }
        }
	}
	pthread_mutex_unlock(&cmut);

    return ret;
}

// return a newly created page cluster the returned object is owned by cache
// manager, and will be removed in delete() method Note that, if oid is not in
// the objectTable, a new object will be created as a side-effect
// IMPORTANT NOTICE: We assume the _seg argument is disjoint with any in-memory
//                   current page clusters. While calling lookup() in insert()
//                   can be used as sanity check, it leads to complex excetion
//                   handling procedure. So, we better have done the check
//                   before calling insert().
PageCluster* CacheManager::insert(std::string _oid, Segment* _seg) {
    pthread_mutex_lock(&cmut);
    cleanup();
	__gnu_cxx::hash_map<std::string, Item*>::iterator _map = objTable.find(_oid);
    if (_map != objTable.end()) {  // found something
        Item* _item = _map->second;

        // check the page cluster list in _item,
        // this may not be OO, but save us some time on implementation
        std::set<PageCluster*, DereferenceLess>::iterator _pc;
        for (_pc = _item->pcList.begin(); _pc != _item->pcList.end(); ++_pc) {
            long long pc_l = (*_pc)->getLeft();
            long long pc_r = (*_pc)->getRight();

			// only need to consider case 3...
            // if (pc_l >= s_l && pc_r <= s_r) { 
			if (!(_seg->rRange <= pc_l || pc_r <= _seg->lRange)) {   // if not totally disjoint
				std::cout << "Overlaped Page Cluster insertion!" << std::endl;
		        pthread_mutex_unlock(&cmut);

				return NULL;
			}
		}
	}

	// first, allocate the page cluster...
	assert(_seg->rRange - _seg->lRange > 0);
    PageCluster *_ret = new PageCluster(_oid, _seg->lRange, _seg->rRange);

	//	no page cluster allocated (e.g. memory is full) -- a special (rare) case
	if (_ret->pages.empty()) {
		delete(_ret);
		pthread_mutex_unlock(&cmut);
		return NULL;
	}

    // this points to the subject table entry
    Item *_item;
    // do we already have this object 
	_map = objTable.find(_oid);
    if (_map == objTable.end()) {  // we don't have this object in table
        _item = new Item(_oid, 0); // create an item with empty pcList
        objTable.insert(make_pair(_oid, _item));    // insert it into obj table

		//	update the lifetime log
		stats->itemLifetime.insert(make_pair(_oid, Util::get_tracetime()));
    }
    else {
        _item = _map->second; // add the new pagecluster to this item
    }
    _item->addPC(_ret);         // add the new pc to sorted set
    _item->size += _ret->length;// accumulate the cached size
    _ret->item = _item;         // save the pointer 
	mCache->addSize(blockPool->file_size(_ret->length));	//	update the cache size
//	mCache->addSize(_ret->length);	//	update the cache size
    pthread_mutex_unlock(&cmut);

    return _ret;
}

// Similar, an objet will be removed from the table if it has no page cluster
// after the delection. 
// *_pc will be free in this function, callers should not refer to it anymore
// IMPORTANT NOTICE: We assume the _pc is not inused (ref. cnt. == 0)
//                   While this can be checked here, it will complicate
//                   the exception handling process, so we better do that
//                   before calling this function.
void CacheManager::remove(std::string _oid, PageCluster* _pc){
	pthread_mutex_lock(&cmut);
    cleanup();

    // do we have this object?
    __gnu_cxx::hash_map<std::string, Item*>::iterator _map = objTable.find(_oid);
    if (_map != objTable.end()) {    // we do have this object, inspect its pcList
        Item* _item = _map->second;  // remove the new pagecluster from this item
        _item->pcList.erase(_pc);    // remove the new pc from sorted set
        _item->size -= _pc->length; // deduct the cached size

        // if the item has zero size or empty pcList (should occur at the same time though)
        // we remove the entry from objTable
        if (_item->pcList.empty()) {
            // remove item
            objTable.erase(_map);  // remove it from the object table
            // It turns out that Cache object would call us to remove the pc from object
            // table, so, we should *NOT* delete item nor pc here...It is done in
            // Cache object.
            // delete _item; // Note: We delete _item here, however, the same item may be
                          //       used by other data structure. 
                          //       Cleanup operations for those structure can be done
                          //       here, or before calling remove method
			//	update the lifetime log
			++stats->removedItemCounter;
			stats->avgLifetime += Util::get_tracetime() - stats->itemLifetime[_oid];
        }
		mCache->addSize(-blockPool->file_size(_pc->length));
//		mCache->addSize(-_pc->length);
        if (_pc -> delayed_removal == false) {  // check this for COSS
            delete _pc;    // we free page cluser here, the caller function should not use _pc anymore
        }
        else {
            _pc->pre_destructor();  // free the resource at least
        }
    }
    else {
        // if we can not find the object, not much we can do 
        std::cout << "Unable to remove a page cluster from object table. Check data consistency." << std::endl; 
    }
	pthread_mutex_unlock(&cmut);

}

// Check the hash table for a specific Item
// return True if we have the Item in cache and return False otherwise
bool CacheManager::contains(std::string _oid) {
	if (objTable.find(_oid) != objTable.end())
		return true;
	return false;
}

// for now, we dump the object table
std::string CacheManager::toString() {
    std::string ret("");
    __gnu_cxx::hash_map<std::string, Item*>::iterator _pos;
    ret.append("Object Table starts").append(Util::CR_str);
    for (_pos = objTable.begin(); _pos != objTable.end(); ++_pos) {
        ret.append("ObjId: ").append(_pos->first);
        ret.append(" Item: ").append(_pos->second->toString());
        ret.append(Util::CR_str);
    }
    ret.append("Object Table ends").append(Util::CR_str);
    return ret;
}

// return the requested Item object from its objTable
Item* CacheManager::getItem(std::string _oid)
{
    __gnu_cxx::hash_map<std::string, Item*>::iterator _map = objTable.find(_oid);
    if (_map == objTable.end()) return NULL;   // not found
	else return _map->second;
}

// this is called by _pc, usually after received, to merge adjacent page
// clusters.  Note: We do not merge a piece larger than 16 MB. Huge page clusters
// require too many buffer pages to be swapped in
//
// Note: we do not merge any page cluster that is not in memory. Merging something
//       that is not memory is too costly 
void CacheManager::compact(PageCluster *_pc) {
    if (merge != true) return;      // if merge is not enabled, then don't compact
	pthread_mutex_lock(&cmut);
    Item* _item = getItem(_pc->objId);   // find the pcList
    if (_item == NULL) {
        std::cout << "Unable to find item object: " << _pc->toString() << std::endl;
        assert(0);
	    pthread_mutex_unlock(&cmut);
        return;
    }
    std::set<PageCluster*, DereferenceLess>::iterator _pos = _item->pcList.find(_pc);
    std::set<PageCluster*, DereferenceLess>::iterator _victim_pos = _pos;
    // postpends
    while(_victim_pos != _item->pcList.end()) {
        _victim_pos = _pos;
        ++_victim_pos;
        if (_victim_pos != _item->pcList.end() && 
            (*_victim_pos)->length + (*_pos)->length <= max_merge&&
            (*_pos)->offset + (*_pos)->length == (*_victim_pos)->offset && 
            (*_victim_pos)->isNotInMem() == false) {
            //std::cout << "Before post-merge " << (*_victim_pos)->toString() << " to " << (*_pos)->toString() << std::endl;
            bool result = (*_pos)->postpend(*_victim_pos); // try to merge
            if (result == true) {
                //std::cout << "After post-merge " << (*_victim_pos)->toString() << " to " << (*_pos)->toString() << std::endl;
                _item->pcList.erase(_victim_pos); // remove the victim page cluster
                delete (*_victim_pos);           // free the memory
            }
            else {
                //std::cout << "Fail to post-merge " << (*_victim_pos)->toString() << " to " << (*_pos)->toString() << std::endl;
                break;  // the victim page is not ready to be merged
            }
        }
        else {
            break;   // if there is a hole, stop merging
        }
    }
    // prepends
    _victim_pos = _pos;
    while(_pos != _item->pcList.begin()) {
        _victim_pos = _pos;
        --_victim_pos;
        if (_victim_pos != _pos && 
            (*_victim_pos)->length + (*_pos)->length <= max_merge&&
            (*_pos)->offset == (*_victim_pos)->offset + (*_victim_pos)->length &&
            (*_victim_pos)->isNotInMem() == false) {
            //std::cout << "Before pre-merge " << (*_victim_pos)->toString() << " to " << (*_pos)->toString() << std::endl;
            bool result = (*_pos)->prepend(*_victim_pos); // try to merge
            if (result == true) {
                //std::cout << "After pre-merge " << (*_victim_pos)->toString() << " to " << (*_pos)->toString() << std::endl;
                _item->pcList.erase(_victim_pos); // remove the victim page cluster
                delete (*_victim_pos);           // free the memory
            }
            else {
                //std::cout << "Fail to pre-merge " << (*_victim_pos)->toString() << " to " << (*_pos)->toString() << std::endl;
                break;  // the victim page is not ready to be merged
            }
        }
        else {
            break;   // if there is a hole, stop merging
        }
    }
	pthread_mutex_unlock(&cmut);
}

// Change piece sizes, this is called by traffic identifier once for every object.
// We find the subject pcList, and update all page cluster on it.
// Depending on the split policy, we may (1) drop all page clusters, or (2) copy over contents
// All dropped (old) page clusters are stored on a garbage page cluster lists. Page clusters on garbage
// list are immediately deleted once it's refCnt reaches zero. 
// Another thing we do here is update all Conn.reqQueue() with new _new_piece_size.
void CacheManager::relocate(std::string _oid, long long _old_piece_len, long long _new_piece_len) {
	pthread_mutex_lock(&cmut);
    Item* _item = getItem(_oid);   // find the pcList
    if (_item == NULL) {
        std::cout << "Unable to find item object: " << _oid << std::endl;
	    pthread_mutex_unlock(&cmut);
        return;
    }
	std::set<PageCluster*, DereferenceLess> new_pcList; // a new page cluster list
    std::set<PageCluster*, DereferenceLess>::iterator _pc;
    for (_pc = _item->pcList.begin(); _pc != _item->pcList.end(); ++_pc) {
        if (dbg_info_level & DBG_INF_CONN) {
            stats->lprintf(Stats::INFO, 
                           "CacheManager::relocate(...) pushing %s into garbage list\n", 
                           (*_pc)->toString().c_str());
        }
        garbageList.insert(*_pc);
        if (split_copyover == true) {
            std::list<PageCluster*> new_pages = (*_pc)->split(_old_piece_len, _new_piece_len);
            if (dbg_info_level & DBG_INF_CONN) {
                stats->lprintf(Stats::INFO, 
                               "CacheManager::relocate(...) split %s into %d page clusters\n", 
                               (*_pc)->toString().c_str(), new_pages.size());
            }
            int cnt = 0;
            std::list<PageCluster*>::iterator iter;
            for (iter = new_pages.begin(); iter != new_pages.end(); ++ iter) {
                assert(*iter != NULL);
                new_pcList.insert(*iter); // add page clister to new page cluster list
				mCache->addSize(blockPool->file_size((*iter)->length));	//	update the cache size
                if (dbg_info_level & DBG_INF_CONN) {
                    stats->lprintf(Stats::INFO, 
                                   "CacheManager::relocate(...) the %dth new page from spliting %s is %s\n", 
                                   ++cnt,  
                                   (*_pc)->toString().c_str(), 
                                   (*iter)->toString().c_str());
                }

//                mCache->addSize((*iter)->length);

            }    
        }
		mCache->addSize(-blockPool->file_size((*_pc)->length));	//	update the cache size
//		mCache->addSize(-(*_pc)->length);
    }
    _item->pcList = new_pcList; // switch to new pc list
    // update segments in Conn object - call update request methods
    Conn::update_all_req(_old_piece_len, _new_piece_len);    
	pthread_mutex_unlock(&cmut);
}

// check the garbage list, and free un-unsed page clusters
void CacheManager::cleanup() {
    std::set<PageCluster*, DereferenceLess>::iterator _pos;
    for (_pos = garbageList.begin(); _pos != garbageList.end(); ++_pos) {
        if ((*_pos)->refCnt <= 0) {
            delete (*_pos);
        }
    }
}

// The following function is called to get the PMF of page cluster size
// We round the pc->length to 4 kB to reduce the length of list
// Nonetheless, this function should be called infrequently due to its high
// overhead, such as once before exiting
std::map<long long, long long> CacheManager::get_pc_pmf() {
    std::map<long long, long long> ret;
    __gnu_cxx::hash_map<std::string, Item*>::iterator _map;
    for (_map = objTable.begin(); _map != objTable.end(); ++ _map) {
        Item* _item = _map->second;
        std::set<PageCluster*, DereferenceLess>::iterator _pc; 
        for (_pc = _item->pcList.begin(); _pc != _item->pcList.end(); ++_pc) {
            ret[(long long)(floor((*_pc)->length / 4096)) * 4096L] ++;
        }    
    }
    return ret;
}


