#include <ctype.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <cmath>
#include <errno.h>
#include <iostream>
#include <algorithm>
#include <linux/netfilter_ipv4.h>
#include <linux/tcp.h>
#include <poll.h>

#include "pCache.h"
#include "Conn.h"
#include "Cache.h"
#include "GnutellaIdentifier.h"
#include "BitTorrTrafficIdentifier.h"
#include "libtcpsplice.h"

// init flag, to create static mutex for Conn table (and Conn table itself)
static bool smut_init = false;
static pthread_mutex_t smut;   // Lock the CacheManager lookup table
static std::set<Conn*> conns; // all connection objs
static long long req_counter = 0;

Conn::Conn(SOCKET cs, Cache* mCache) :
    byPass(false), swap(false), ti(NULL), client_socket(INVALID_SOCKET), 
    server_socket(INVALID_SOCKET), reqLen(0), resLen(0)
{
	client_socket = cs;
	this->mCache = mCache;
    pthread_mutex_init(&mut, NULL);
    if (smut_init == false) {
        smut_init = true;
        pthread_mutex_init(&smut, NULL);
    }
}


Conn::~Conn()
{
    // clean up the left over segments if any
    std::list<Segment*>::iterator iter;
    for (iter = reqQueue.begin(); iter != reqQueue.end(); ++iter) {
        if (*iter != NULL) delete (*iter);
    }
    pthread_mutex_destroy(&mut);
    pthread_mutex_lock(&smut);
    conns.erase(this);
    pthread_mutex_unlock(&smut);
    if (NULL != ti) {
        delete ti;
        ti = NULL;
    }
}

void Conn::reg() {
    pthread_mutex_lock(&smut);
    conns.insert(this);
    pthread_mutex_unlock(&smut);
}

// This function represents the main relay loop We read something from the
// client socket and server socket, if there is one.
// We handle one current segment. Whenever a segment is finished, we pop one
// from the reqQueue. We assume that either client or server would close a
// connection when they see it's appropriate. We only terminated when one
// of our TCP connections is terminated.
void Conn::run()
{
    // the c2s and s2c are named under the assumption of swap == false
    PageCluster* pc = NULL;         // current page cluster
    bool admitted = false;          // should we copy the content into memory
    bool avail;   // whether the subsegment is in our cache
	char	c2s_buffer[BUFSIZE];    // the c2s circular buffer
	char	s2c_buffer[BUFSIZE];    // the s2c circular buffer
    char    *c2s_start = c2s_buffer, *c2s_end = c2s_buffer;   // initially, the buffer is empty
    char    *s2c_start = s2c_buffer, *s2c_end = s2c_buffer;   // initially, the buffer is empty
	int		retval;   // a temporary integer to receive send/recv return valuew
    char    *data_start = NULL;   // identifier set this to non-NULL value to indicate the begining
                                  // of the actual data
    char    *tmp_data_start = NULL;

    //recieve the first few packets from the connection initializer
	retval = recv(client_socket, recv_start_pt(c2s_buffer, c2s_start, c2s_end) , BUFSIZE - 1, 0);  // we need block-io here

	if(retval == SOCKET_ERROR) {
        stats->lprintf(Stats::INFO, "Error Recv");
		close(client_socket);
        client_socket = INVALID_SOCKET;
        return;
	}
	if(retval == 0) {
		stats->lprintf(Stats::INFO, "Client Close connection");
		close(client_socket);
        client_socket = INVALID_SOCKET;
        return;
	}
    c2s_end += retval;  // advance end pointer

    // use the static functions to identify protocols
    // passing the data received from client to traffic identifiers
	if (BitTorrTrafficIdentifier::identify(c2s_buffer, retval)) {
        ti = new BitTorrTrafficIdentifier(c2s_buffer, retval, this);
    }    
    else if (GnutellaIdentifier::identify(c2s_buffer, retval)) {
        ti = new GnutellaIdentifier(c2s_buffer, retval, this);
    }
    else {  // not a recognized initial pattern, pass the whole traffic along long 
        byPass = true;
    }

    // check the original ip address here
    struct sockaddr_in server_sin;
    socklen_t server_sinlen = sizeof(server_sin);
    if (getsockopt(client_socket, SOL_IP, SO_ORIGINAL_DST, &server_sin, &server_sinlen) != 0) {
        stats->lprintf(Stats::ERR, "Unable to recover original destination address");
    } else {
        /* success, address is in sin */
        stats->lprintf(Stats::INFO, "Connected, server ip/port = %s/%d", 
                inet_ntoa(server_sin.sin_addr), ntohs(server_sin.sin_port));
        // make a connection now
    }
    struct sockaddr client_addr;
    socklen_t server_addr_len = sizeof(client_addr);
    if (0 != getpeername(client_socket, &client_addr, &server_addr_len)) { // failed
        stats->lprintf(Stats::INFO, "Unable to recover original source address");
    }
    else {
        stats->lprintf(Stats::INFO, "Connected, client ip/port = %s/%d", 
                inet_ntoa(((struct sockaddr_in*)&client_addr)->sin_addr), 
                ntohs(((struct sockaddr_in*)&client_addr)->sin_port));
    }

    // Also, it may be a good idea to check whether the client and/or the server 
    // is in our network to determine who is the sender, and who is the receiver.
    // set the swap flag, and actually swap the buffer if nedded
    // We only get serious if one of the client/server is in our subset (domain), 
    // in all other cases, relay the data
    bool server_local = (
        (server_sin.sin_addr.s_addr & inet_addr(confMap["NETMASK"].c_str())) == 
        (inet_addr(confMap["SUBNET"].c_str()) & inet_addr(confMap["NETMASK"].c_str())));
    bool client_local = (
        (((struct sockaddr_in*)&client_addr)->sin_addr.s_addr & inet_addr(confMap["NETMASK"].c_str())) == 
        (inet_addr(confMap["SUBNET"].c_str()) & inet_addr(confMap["NETMASK"].c_str())));
    stats->lprintf(Stats::INFO, "client_local/server_local = %d/%d", client_local, server_local);  

	if (true == server_local && false == client_local) {  // need swap
        swap = true;
	}
    else if ((true == server_local && true == client_local) || 
             (false == server_local && false == client_local)) {
                 byPass = true;
	}

    // ideally, the constructor will create a connection to the actual server
    // in the case of non-transparent cache.
    // Since we only do transpartent cache, let's make a connection
	//server_socket = socket(AF_INET, SOCK_STREAM, 0); /* Open a socket */
	server_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP); /* Open a socket */
	if (server_socket < 0 ) {
		stats->lprintf(Stats::WARN, "Error: opening proxy_server socket failed.");
        // we abort the operation when calling the select()
        server_socket = INVALID_SOCKET;
	}
    // apply tproxy hack
    int on = 1;
    int rr;
    if (server_socket != INVALID_SOCKET) {
        rr = setsockopt(server_socket, SOL_IP, IP_TRANSPARENT, &on, sizeof(on)); 
        if (rr == -1) {
            stats->lperror(Stats::WARN, "");
            close(server_socket);   // need to close the socket!
            server_socket = INVALID_SOCKET;
        }
    }
    // enable address reuse, use the same variable used in tproxy
    if (server_socket != INVALID_SOCKET) {
        rr = setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on) );
        if (rr == -1) {
            stats->lperror(Stats::WARN, "");
            close(server_socket);   // need to close the socket!
            server_socket = INVALID_SOCKET;
        }
    }
    // fake our ip address/port
	if (server_socket != INVALID_SOCKET) {
        struct sockaddr_in client_sin = *((struct sockaddr_in*)&client_addr);
    
		if(bind(server_socket,(sockaddr *)&client_sin, sizeof(client_sin)) != 0) {
				stats->lprintf(Stats::WARN, "Error in binding local address.");
				close(server_socket);   // need to close the socket!
				server_socket = INVALID_SOCKET;
			}
    }
    // prepare for tcp splicing if it's enabled and byPass is true
    if (server_socket != INVALID_SOCKET && tcp_splicing == true && byPass == true) {
        /* We prepare our plice */
        tcp_splice_initfd(client_socket, server_socket);
        /* 
         * This option is critical here. We force kernel tcp stack to
         * push all pending frame while sending data. This option
         * avoid the socket queue flushing before kernel tcp splice
         * start forwarding data.
         */ 
        rr = setsockopt(server_socket, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on) );
        if (rr == -1) {
            stats->lperror(Stats::WARN, "");
            close(server_socket);   // need to close the socket!
            server_socket = INVALID_SOCKET;
        }
        rr = setsockopt(client_socket, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on) );
        if (rr == -1) {
            stats->lperror(Stats::WARN, "");
            close(client_socket);   // need to close the socket!
            client_socket = INVALID_SOCKET;
        }
    } 
    if (server_socket != INVALID_SOCKET) {
        // connect to server_sin now..
        if (connect(server_socket, (struct sockaddr*)&server_sin, sizeof(server_sin)) == SOCKET_ERROR) {
            stats->lprintf(Stats::WARN, "Error: connect to server failed.");
            // we abort the operation when calling the select()
            close(server_socket);   // need to close the socket!
            server_socket = INVALID_SOCKET;
        }
    }
    // At this point, both client and server sockets should be ready if any of
    // them is still INVALID_SOCKET, we have to terminated this thread
    if (server_socket == INVALID_SOCKET || 
        client_socket == INVALID_SOCKET) {
        stats->lprintf(Stats::WARN, "Unable to make server connection, thread exiting.");
        if (client_socket != INVALID_SOCKET) { close(client_socket); client_socket = INVALID_SOCKET; }
        if (server_socket != INVALID_SOCKET) { close(server_socket); server_socket = INVALID_SOCKET; }
        return;
    }         
    // Now, if tcp splcing is on, and we are ready, start tcp splicing and terminate this thread
    // Note that, we assume the kernel will clean up sockets whenever client or server closes its
    // connection. NEED TO VERIFY THIS!
    if (tcp_splicing == true && byPass == true) {
        // tcp splicing - splice both fds
        // I believe the last argument means how many bytes will be send to the server side before the
        // traffic being handled by kernel. Since we sends the residue data just above, we have nothing
        // more to tell the server
        //printf ("before splicing %d and %d\n", client_socket, server_socket);
        tcp_splice_splicefd(client_socket, server_socket, send_length(c2s_buffer, c2s_start, c2s_end));
        //printf ("after splicing %d and %d\n", client_socket, server_socket);
        // we first write the received data (retval bytes) to the server, then setup tcp splcing
        while (send_length(c2s_buffer, c2s_start, c2s_end) > 0) {
            // send the c2s_buffer
            int _len = send(server_socket, c2s_start, send_length(c2s_buffer, c2s_start, c2s_end), MSG_NOSIGNAL);
           // printf("sent _len: %d\n", _len);
            if(_len == SOCKET_ERROR) {
                stats->lperror(Stats::WARN, "Error send() failed (when setting up TCP splicing): ");
                // abort the whole thing now.
                if (client_socket != INVALID_SOCKET) close(client_socket);
                if (server_socket != INVALID_SOCKET) close(server_socket);
                return;
            }
            // update the buffer pointer
            c2s_start = send_update_pt(c2s_buffer, c2s_start, c2s_end, _len);
        }
        // it's interesting that Linux's tcp splicing implementation requires us to call read() 
        // although data is automatically copied to another connection. 
        // We do this in the following subrouting
        //
        do_tcp_splicing(server_socket, client_socket, c2s_buffer, BUFSIZE);  // reuse c2s_buffer
        if (client_socket != INVALID_SOCKET) {close(client_socket); client_socket = NULL;}
        if (server_socket != INVALID_SOCKET) {close(server_socket); server_socket = NULL;}
        return;  // we are done here. IMPORTANT: verify whether we have socket-leak...
    }

    //  ****************	SELECT	********************
	struct timeval selectTimeout; // timeout 
	fd_set rFds;   // reading descriptor s
	fd_set eFds;   // exception descriptors
	int fdMax;
    int fdsSelected = 0;

    // data assembler....
	CacheResponse*	mResponse = NULL;  // this holds the cache query response for the current segment
	std::list<PageCluster*>::iterator availableIter;
	std::list<PageCluster*>::iterator requestedIter;
    Segment* reqSeg = NULL;         // current segment
    Segment* subReqSeg = NULL;      // current sub-segment
    long long current = -1;         // the global offset, -1 indicates un-initialized
    bool first_time = true;         // skip the select() once and process the first few packets
    int relayLen = 0;
    bool delayed_response_header = false;

	while(TRUE) {
restart:
        // process a new request...
        if (reqSeg == NULL && !reqQueue.empty()) {
            pthread_mutex_lock(&mut);
            if (dbg_info_level & DBG_INF_CONN) {
                stats->lprintf(Stats::INFO, "request queue size: %d" , (int)reqQueue.size());
            }
            reqSeg = reqQueue.front();
            reqQueue.pop_front();
            pthread_mutex_unlock(&mut);

		//	log the client statistics
#if THIS_CODE_ARE_NOT_THREAD_SAFE_YET_NEED_REVISIT
		if (!byPass) {
			std::map<std::string, ClientStatEntry*>::iterator _itr;
			struct sockaddr_in _sockaddr;
			if (client_local)
				_sockaddr = *((struct sockaddr_in*)&client_addr);
			else
				_sockaddr = server_sin;
		
			//	we see the client for the first time
			if ((_itr = stats->clientTable.find(ti->receiver_id)) == stats->clientTable.end()) {
				stats->clientTable.insert(std::pair<std::string, ClientStatEntry*>
					(ti->receiver_id, new ClientStatEntry(reqSeg->objId.substr(0,40))));
		
			//	the client is already in the table
			} else {
				//	log the dl speed [at most every 10Secs]
				if((Util::get_timestamp() - _itr->second->lastTime) > 10000000) {
					stats->lprintf(Stats::CLIENT, "ip/port %s:%d, DL Speed: %f",
                				inet_ntoa(_sockaddr.sin_addr), ntohs(_sockaddr.sin_port),
							1000.0f*(double)(_itr->second->byte_count)/(double)(Util::get_timestamp() - _itr->second->lastTime));
					_itr->second->lastTime = Util::get_timestamp();
					_itr->second->byte_count = 0;
				}
				//	previous download is complete
				if ((_itr->second)->oid.compare(reqSeg->objId.substr(0,40)) != 0) {
					stats->lprintf(Stats::CLIENT, "ip/port %s:%d, %s DL Complete: %lld bytes in %lld",
                				inet_ntoa(_sockaddr.sin_addr), ntohs(_sockaddr.sin_port), _itr->second->total_byte, _itr->second->oid.c_str(),
							Util::get_timestamp() - _itr->second->startTime);
					_itr->second->startTime = Util::get_timestamp();
					_itr->second->oid = reqSeg->objId.substr(0,40);
					_itr->second->lastTime = _itr->second->startTime;
					_itr->second->byte_count = 0;
					_itr->second->total_byte = 0;
				}
				_itr->second->byte_count += reqSeg->rRange - reqSeg->lRange;
				_itr->second->total_byte += reqSeg->rRange - reqSeg->lRange;
			}
		}
#endif // THIS_CODE_ARE_NOT_THREAD_SAFE_YET_NEED_REVISIT

			// update the req_counter to stats module
            stats->request = (++req_counter);
            if (stats->request % req_counter_step == 0) {
                // generate a PR report
                stats->gen_PR();
            }
            
            // start to get this request...
            current = reqSeg->lRange;
            if (dbg_info_level & DBG_INF_CONN) {
                stats->lprintf(Stats::INFO, "Processing segment: <%lld, %lld>.\n", 
                               reqSeg->lRange, reqSeg->rRange);
            }
            // Sending the extracted Segment to the cache
            mResponse = mCache->query(reqSeg);
            if (dbg_info_level & DBG_INF_CONN) {
                stats->lprintf(Stats::INFO, "Query result: %s\n", 
                               mResponse->toString().c_str());
            }

            // Bear, 2008 Jan, we put a workaround here, because of the message
            // type 20 issue in BitTorrent Now, we don't request partial
            // request at all. That is, we either request for the whole segment
            // or no segment The former case happens when there is at least one
            // subsegment is missing.  This should be revisited, and traffic
            // processor should be improved to handle all types of messages.
            if (mResponse->availableSegment->size() > 0 && 
                false == complete_avail(mResponse->availableSegment, reqSeg)) {
                if (dbg_info_level & DBG_INF_CONN) {
                    stats->lprintf(Stats::INFO, 
                                   "Partial requests are required, abort this mResponse: %s\n", 
                                    mResponse->toString().c_str()); 
                }

                for (availableIter = mResponse->availableSegment->begin();
                     availableIter != mResponse->availableSegment->end(); 
                     ++availableIter) {
                    pc = (*availableIter);
                    pc->decRefCnt();
                }
                for (requestedIter = mResponse->requestedSegment->begin();
                    requestedIter != mResponse->requestedSegment->end(); 
                    ++requestedIter) {
                    pc = (*requestedIter);
                    // pc->decRefCnt(); done is abort
                    mCache->abort(pc);
                    pc = NULL;  // pc has been freed
                }
                mResponse->availableSegment->clear();
                mResponse->requestedSegment->clear();
            }
		    availableIter = mResponse->availableSegment->begin();
		    requestedIter = mResponse->requestedSegment->begin();
        }

        // find a new subsegment
        while (subReqSeg == NULL && reqSeg != NULL) {
            // the left range is always the current, but the right range is the
            // closest boundary of (*availableIter)->getLeft(), ->getRight(),
            // (*requestedIter)->getLeft(), ->getRight().  In the cases, where
            // we pick the left boundary, no in-memory buffer read/write is
            // needed.
            // If we pick (*availableIter)->getRight(); then we should pump the
            // data immediate, and find next subReSeg
            // Finally, if we pick (*requestedIter)->getRight(), we should make
            // a request, and copy data into memory
            long long next_end = LLONG_MAX;  // if we are above all pcs
            admitted = false;
            avail = false;
            if (availableIter != mResponse->availableSegment->end() &&
                (*availableIter)->getRight() <= next_end) {
                next_end = (*availableIter)->getRight();
                admitted = false;
                avail = true;           // play out from our cache
                pc = (*availableIter);
            }
            if (requestedIter != mResponse->requestedSegment->end() &&
                (*requestedIter)->getRight() <= next_end) {
                next_end = (*requestedIter)->getRight();
                admitted = true;
                avail = false;          // copy into our cache
                pc = (*requestedIter);
            }
            if (availableIter != mResponse->availableSegment->end() &&
                (*availableIter)->getLeft() > current && 
                (*availableIter)->getLeft() < next_end) {
                next_end = (*availableIter)->getLeft();
                admitted = false;
                avail = false;          // relay
                pc = NULL;
            }
            if (requestedIter != mResponse->requestedSegment->end() &&
                (*requestedIter)->getLeft() > current && 
                (*requestedIter)->getLeft() < next_end) {
                next_end = (*requestedIter)->getLeft();
                admitted = false;
                avail = false;          // relay
                pc = NULL;
            }
			next_end = std::min(next_end, reqSeg->rRange);

            if (admitted == false && avail == true) availableIter++;
            if (admitted == true && avail == false) requestedIter++;
            
            if (current == reqSeg->lRange) { 
                
                if (mResponse->availableSegment->size() > 0) {
                    memcpy(resBuf, reqBuf, reqLen);
                    bool _status;
                    _status = ti->createResponse(resBuf, &resLen, reqSeg);  
                    if (_status) {
                        if (swap == false) {
                            retval = send(client_socket, resBuf, resLen, MSG_NOSIGNAL);
                        }
                        else { // swap == true
                            retval = send(server_socket, resBuf, resLen, MSG_NOSIGNAL);
                        }
                        if(retval == SOCKET_ERROR) {
                            stats->lperror(Stats::WARN, "Error send() failed: ");
                            goto cleanup;
                        }
                    }
                    delayed_response_header = false;
                } 
                else {
                    // dealy response is set for those protocols that can not support
                    // partial subrequest. For example, BitTorrent protocol often sends
                    // message 20 before message 7 (piece message). 
                    // For these protocol, we have no choice but preventing partial
                    // subrequest. 
                    delayed_response_header = true;
                }

#if 0 // We only request complete segments for now
                if (mResponse->availableSegment->size() + 
                    mResponse->requestedSegment->size() > 1) {
                    // send a response to the client
                    // Note that, the most recent response buffer may not
                    //            be useful, because, e.g., if a request
                    //            can be locally satisfied, we would get
                    //            no response at all.
                    //            Make a copy of the latest request message
                    //            and let the protocol do the work.
                    memcpy(resBuf, reqBuf, reqLen);
                    bool _status;
                    _status = ti->createResponse(resBuf, &resLen, reqSeg);  
                    if (_status) {
                        if (swap == false) {
                            retval = send(client_socket, resBuf, resLen, MSG_NOSIGNAL);
                        }
                        else { // swap == true
                            retval = send(server_socket, resBuf, resLen, MSG_NOSIGNAL);
                        }
                        if(retval == SOCKET_ERROR) {
                            stats->lperror(Stats::WARN, "Error send() failed: ");
                            goto cleanup;
                        }
                    }
                    delayed_response_header = false;
                }
                else {
                    // dealy response is set for those protocols that can not support
                    // partial subrequest. For example, BitTorrent protocol often sends
                    // message 20 before message 7 (piece message). 
                    // For these protocol, we have no choice but preventing partial
                    // subrequest. 
                    delayed_response_header = true;
                }
#endif // We only request complete segments for now
            }

            // if we have the current part, send it to the client
            if (admitted == false && avail == true) {
                // send from current to next_end
                // no need to create subSeqSeg here, we handle the request right away
                relayLen = BUFSIZE - 1;  // at most we have this many bytes
                while (current < next_end) {
                    if (swap == false) {
                        retval = std::min(next_end - current, (long long) recv_length(s2c_buffer, s2c_start, s2c_end)); 
                        pc->read(current - pc->offset, retval, s2c_end);  // abusing retval...
                        s2c_end = recv_update_pt(s2c_buffer, s2c_start, s2c_end, retval);
                        // std::cout << "current was (3): " << current << " " << pthread_self() << std::endl;
                        current += retval;          // update here, becasue we are about to overwrite retval
                        // std::cout << "current is (3): " << current << " " << pthread_self() << std::endl;
                        // send the s2c_buffer
                        retval = send(client_socket, s2c_start, 
                                          std::min(send_length(s2c_buffer, s2c_start, s2c_end), relayLen), MSG_NOSIGNAL);
                        if(retval == SOCKET_ERROR) {
                            stats->lperror(Stats::WARN, "Error send() failed: ");
                            break;
                        }
                        // retval is fine, update the buffer pointer
                        s2c_start = send_update_pt(s2c_buffer, s2c_start, s2c_end, retval);
                    }
                    else { // swap == true
                        retval = std::min(next_end - current, (long long) recv_length(c2s_buffer, c2s_start, c2s_end)); 
                        pc->read(current - pc->offset, retval, c2s_end);  // abusing retval...
                        c2s_end = recv_update_pt(c2s_buffer, c2s_start, c2s_end, retval);
                        // std::cout << "current was (4): " << current << " " << pthread_self() << std::endl;
                        current += retval;          // update here, becasue we are about to overwrite retval
                        // std::cout << "current is (4): " << current << " " << pthread_self() << std::endl;
                        // send the c2s_buffer
                        retval = send(server_socket, c2s_start, 
                                          std::min(send_length(c2s_buffer, c2s_start, c2s_end), relayLen), MSG_NOSIGNAL);
                        if(retval == SOCKET_ERROR) {
                            stats->lperror(Stats::WARN, "Error send() failed: ");
                            break;
                        }
                        // retval is fine, update the buffer pointer
                        c2s_start = send_update_pt(c2s_buffer, c2s_start, c2s_end, retval);
                    }
                }
                if (next_end >= reqSeg->rRange) { // means we are are not in with the current segment
                    pc->decRefCnt();    // done with this page cluster
                    reqSeg = NULL;
                }
            }
            else {
                subReqSeg = new Segment(reqSeg->objId, current, next_end, 0, reqSeg->piecelen);   // 0 is the timestamp, not used
                data_start = 0;   // set a flag showing we are not in the middle of a response message
                // send out a request to the server
                bool _status;
                _status = ti->createRequest(reqBuf, &reqLen, subReqSeg);  
                if (dbg_info_level & DBG_INF_CONN) {
                    stats->lprintf(Stats::INFO, "Requesting sub segment: <%lld, %lld>.\n", 
                                                subReqSeg->lRange, subReqSeg->rRange);
                }
                if (_status) {
                    if (swap == false) {
                        retval = send(server_socket, reqBuf, reqLen, MSG_NOSIGNAL);
                    }
                    else { // swap == true
                        retval = send(client_socket, reqBuf, reqLen, MSG_NOSIGNAL);
                    }
                    if(retval == SOCKET_ERROR) {
                        stats->lperror(Stats::WARN, "Error send() failed: ");
                        goto cleanup;
                    }
                }
            }
        }

        // Bear, if the segment is done, we don't need to go through the 
        // select (which will take 10 second for nothing).
        if (reqSeg == NULL && !reqQueue.empty()) {
            if (dbg_info_level & DBG_INF_CONN) {
			    stats->lprintf(Stats::INFO, "Bear, fast restart!");
            } 
            delayed_response_header = false;
            goto restart;
        }

        // set up time out
		selectTimeout.tv_sec  = 10;
		selectTimeout.tv_usec = 0;
	 
        // reset fds
		FD_ZERO(&rFds);
		FD_ZERO(&eFds);
        if (first_time == false) {
            FD_SET(client_socket, &rFds);
            FD_SET(server_socket, &rFds);
            FD_SET(client_socket, &eFds);
            FD_SET(server_socket, &eFds);
        }

        // set up fd std::max 
		fdMax = 0;
		if(client_socket > fdMax)
			fdMax = client_socket;
		if(server_socket > fdMax)
			fdMax = server_socket;
    
        // select now...
        if (first_time == false)
		    fdsSelected = select(fdMax+1, &rFds, NULL, &eFds, &selectTimeout);

		if(fdsSelected < 0) {  // something wrong.
			stats->lprintf(Stats::WARN, "Error: select()");
			break;
		}

		if (first_time == false &&
            (FD_ISSET(client_socket, &eFds) || FD_ISSET(server_socket, &eFds))) {
			stats->lprintf(Stats::ERR, "Exception...");
            stats->lperror(Stats::WARN, "");
			break;
		}

        if (first_time == true) {
            relayLen = send_length(c2s_buffer, c2s_start, c2s_end);
            first_time = false;
        }

        // have something to read from the client_socket
        // also, the first few bytes should always go here!
		if (FD_ISSET(client_socket, &rFds)) {
            // receive until we hit the end
            if (dbg_info_level & DBG_INF_CONN) 
            	stats->lprintf(Stats::INFO, "Before recv() (1).");
            retval = recv(client_socket, recv_start_pt(c2s_buffer, c2s_start, c2s_end), recv_length(c2s_buffer, c2s_start, c2s_end), 0);
            if(retval == SOCKET_ERROR) {
                if (errno == EAGAIN) {
                    retval = 0;   // due to non-block i/o
                }
                else {
                    stats->lprintf(Stats::WARN, "Error recv() failed.");
                    goto cleanup;
                }
            }
            else if(retval == 0) {	//	Client side has shut down the connection gracefully
                stats->lprintf(Stats::INFO, "Client closed connection.");
                goto cleanup;
            }
            c2s_end = recv_update_pt(c2s_buffer, c2s_start, c2s_end, retval);
            relayLen = send_length(c2s_buffer, c2s_start, c2s_end);
        }
        if (send_length(c2s_buffer, c2s_start, c2s_end) > 0) {
            // call the filtering function if byPass is false
            bool immediate = false;  // immediately call the filter function?
            do {
                if (byPass == false) {
                    if (swap == false) {
                        relayLen = ti->c2sFilter(&c2s_start, c2s_buffer, &c2s_end);
                    }
                    else { // swap == true
                        tmp_data_start = 0;
                        relayLen = ti->s2cFilter(&c2s_start, c2s_buffer, &c2s_end, &tmp_data_start);
                        if (tmp_data_start != 0) {
                            if (data_start != 0) {
                                stats->lprintf(Stats::ERR, "Warning, data_start is not zero (1), we haven't finished the response message. Current offset: %lld", current);
                            }
                            data_start = tmp_data_start;
                            if (delayed_response_header == true) {
                                // send a response to the client
                                // Note that, the most recent response buffer may not
                                //            be useful, because, e.g., if a request
                                //            can be locally satisfied, we would get
                                //            no response at all.
                                //            Make a copy of the latest request message
                                //            and let the protocol do the work.
                                memcpy(resBuf, reqBuf, reqLen);
                                bool _status;
                                _status = ti->createResponse(resBuf, &resLen, reqSeg);  
                                if (_status) {
                                    // swap == true
                                    retval = send(server_socket, resBuf, resLen, MSG_NOSIGNAL);
                                    if(retval == SOCKET_ERROR) {
                                        stats->lperror(Stats::WARN, "Error send() failed: ");
                                        goto cleanup;
                                    }
                                }
                                delayed_response_header = false;
                            }
                        }
                    }
                    // relayLen > 0 means send out that many bytes, and
                    // immediately call the filter function again relayLen <= 0
                    // would lead to another read operation, which may block
                    // the thread.
                    if (relayLen == MAX_INT_VALUE) {
                        // don't read and forward anything, just call immediately
                        relayLen = 0;
                        immediate = true;
                    }
                    else if (relayLen > 0) {
                        immediate = true;
                    }
                    else {
                        relayLen = -relayLen;
                        immediate = false;
                    }
                }
                if (data_start && swap) {
                    if (tmp_data_start != 0) { // start copying from *data_start
                        retval = send_length(c2s_buffer, data_start, c2s_end); 
                        if (subReqSeg != NULL) retval = std::min((int)(subReqSeg->rRange - current), retval);
                        if (admitted)
                            pc->write(current - pc->offset, retval, data_start);  // abusing retval...
                    } 
                    else {  // otherwise, copy the whole buffer
                        retval = send_length(c2s_buffer, c2s_start, c2s_end);
                        if (subReqSeg != NULL) retval = std::min((int)(subReqSeg->rRange - current), retval);
                        if (admitted)
                            pc->write(current - pc->offset, retval, c2s_start);  // abusing retval...
                    }
                    // std::cout << "current was (1): " << current << " " << pthread_self() << std::endl;
                    current += retval;
                    // std::cout << "current is (1): " << current << " " << pthread_self() << std::endl;
                    if (subReqSeg != NULL && current >= subReqSeg->rRange) {
                        if (dbg_info_level & DBG_INF_CONN) {
                            stats->lprintf(Stats::INFO, 
                                   "Finished sub segment: <%lld, %lld> in segment <%lld, %lld>.\n", 
                                   subReqSeg->lRange, subReqSeg->rRange, 
                                   reqSeg->lRange, reqSeg->rRange);
                        }
                        if (subReqSeg != NULL) {
                            delete subReqSeg;
                            subReqSeg = NULL;
                        }
                        data_start = 0;
                        if (admitted) {
                            admitted = false;
                            pc->received();		//	The PageCluster is ready to store on the disk
                            pc->decRefCnt();
                            // no more store method, the pc is considered to be done 
                            // when dirty bit is 1
                            // mCache->store(pc);			//	We can add the PageCluster into the cache queue
                        }
                    }
                    if (reqSeg != NULL && current >= reqSeg->rRange) { // means we are done with the current segment
                        if (dbg_info_level & DBG_INF_CONN) {
                            stats->lprintf(Stats::INFO, 
                                   "Finished segment <%lld, %lld>.\n", 
                                   reqSeg->lRange, reqSeg->rRange);
                        }
                        if (reqSeg != NULL) {
                            delete reqSeg;
                            reqSeg = NULL;
                        }
                        if (mResponse != NULL) {
                            delete mResponse;
                            mResponse = NULL;
                        }
                    }
                }
                // send / relay the c2s_buffer
                while (relayLen > 0) {
                    if (dbg_info_level & DBG_INF_CONN) {
                        stats->lprintf(Stats::INFO, "Remaining relay length (1): %d\n", relayLen);
                    }
                    retval = send(server_socket, c2s_start, 
                                  std::min(send_length(c2s_buffer, c2s_start, c2s_end), relayLen), MSG_NOSIGNAL);
                    if(retval == SOCKET_ERROR) {
                        stats->lperror(Stats::WARN, "Error send() failed: ");
                        immediate = false;   // break the outer loop
                        break;
                    }
                    // retval is fine, update the buffer pointer
                    c2s_start = send_update_pt(c2s_buffer, c2s_start, c2s_end, retval);
                    relayLen -= retval;
                }
                if (dbg_info_level & DBG_INF_CONN) {
                    stats->lprintf(Stats::INFO, "Relay completed(1)\n");
                    stats->lprintf(Stats::INFO, "Relay immediate: %d\n", immediate);
                    stats->lprintf(Stats::INFO, "Relay send_length: %d\n", send_length(c2s_buffer, c2s_start, c2s_end));
                }
            }
            while (immediate == true && send_length(c2s_buffer, c2s_start, c2s_end) > 0); // if true, call the filter function again.
		}


		if (FD_ISSET(server_socket, &rFds)) {
            // receive until we hit the end
            if (dbg_info_level & DBG_INF_CONN) 
            	stats->lprintf(Stats::INFO, "Before recv() (2).");
            retval = recv(server_socket, recv_start_pt(s2c_buffer, s2c_start, s2c_end), recv_length(s2c_buffer, s2c_start, s2c_end), 0);
            if(retval == SOCKET_ERROR) {
                if (errno == EAGAIN) {
                    retval = 0;   // due to non-block i/o
                }
                else {
                    stats->lprintf(Stats::WARN, "Error recv() failed.");
                    goto cleanup;
                }
            }
            else if(retval == 0) {	//	Client side has shut down the connection gracefully
                stats->lprintf(Stats::INFO, "Server closed connection.");
                goto cleanup;
            }
            s2c_end = recv_update_pt(s2c_buffer, s2c_start, s2c_end, retval);
            relayLen = send_length(s2c_buffer, s2c_start, s2c_end);
        }
        if (send_length(s2c_buffer, s2c_start, s2c_end) > 0) {
            // call the filtering function if byPass is false
            bool immediate = false;  // immediately call the filter function?
            do {
                if (byPass == false) {
                    if (swap == false) {
                        tmp_data_start = 0;
                        relayLen = ti->s2cFilter(&s2c_start, s2c_buffer, &s2c_end, &tmp_data_start);
                        if (tmp_data_start != 0) {
                            if (data_start != 0) {
                                stats->lprintf(Stats::ERR, "Warning, data_start is not zero (2), we haven't finished the response message. Current offset: %lld", current);
                            }
                            data_start = tmp_data_start;
                            if (delayed_response_header == true) {
                                // send a response to the client
                                // Note that, the most recent response buffer may not
                                //            be useful, because, e.g., if a request
                                //            can be locally satisfied, we would get
                                //            no response at all.
                                //            Make a copy of the latest request message
                                //            and let the protocol do the work.
                                memcpy(resBuf, reqBuf, reqLen);
                                bool _status;
                                _status = ti->createResponse(resBuf, &resLen, reqSeg);  
                                if (_status) {
                                    // swap == false
                                    retval = send(client_socket, resBuf, resLen, MSG_NOSIGNAL);
                                    if(retval == SOCKET_ERROR) {
                                        stats->lperror(Stats::WARN, "Error send() failed: ");
                                        goto cleanup;
                                    }
                                }
                                delayed_response_header = false;
                            }
                        }
                    }
                    else {// swap == true
                        relayLen = ti->c2sFilter(&s2c_start, s2c_buffer, &s2c_end);
                    }
                    // relayLen > 0 means send out that many bytes, and
                    // immediately call the filter function again relayLen <= 0
                    // would lead to another read operation, which may block
                    // the thread.
                    if (relayLen == MAX_INT_VALUE) {
                        // don't read and forward anything, just call immediately
                        relayLen = 0;
                        immediate = true;
                    }
                    else if (relayLen > 0) {
                        immediate = true;
                    }
                    else {
                        relayLen = -relayLen;
                        immediate = false;
                    }
                }
                if (data_start && !swap) {
                    if (tmp_data_start != 0) { // start copying from *data_start
                        retval = send_length(s2c_buffer, data_start, s2c_end);
                        if (subReqSeg != NULL) retval = std::min((int)(subReqSeg->rRange - current), retval);
                        if (admitted)
                            pc->write(current - pc->offset, retval, data_start);  // abusing retval...
                    } 
                    else {  // otherwise, copy the whole buffer
                        retval = send_length(s2c_buffer, s2c_start, s2c_end);
                        if (subReqSeg != NULL) retval = std::min((int)(subReqSeg->rRange - current), retval);
                        if (admitted)
                            pc->write(current - pc->offset, retval, s2c_start);  // abusing retval...
                    }
                    // std::cout << "current was (2): " << current << " " << pthread_self() << std::endl;
                    current += retval;
                    // std::cout << "current is (2): " << current << " " << pthread_self() << std::endl;
                    if (subReqSeg != NULL && current >= subReqSeg->rRange) {
                        if (dbg_info_level & DBG_INF_CONN) {
                            stats->lprintf(Stats::INFO, 
                                   "Finished sub segment: <%lld, %lld> in segment <%lld, %lld>.\n", 
                                   subReqSeg->lRange, subReqSeg->rRange, 
                                   reqSeg->lRange, reqSeg->rRange);
                        }
                        if (subReqSeg != NULL) {
                            delete subReqSeg;
                            subReqSeg = NULL;
                        }
                        data_start = 0;
                        if (admitted) {
                            admitted = false;
                            pc->received();		//	The PageCluster is ready to store on the disk
                            pc->decRefCnt();
                            // no more store method, the pc is considered to be done 
                            // when dirty bit is 1
                            // mCache->store(pc);			//	We can add the PageCluster into the cache queue
                        }
                    }
                    if (reqSeg != NULL && current >= reqSeg->rRange) { // means we are are not in the current segment
                        if (dbg_info_level & DBG_INF_CONN) {
                            stats->lprintf(Stats::INFO, 
                                   "Finished segment <%lld, %lld>.\n", 
                                   reqSeg->lRange, reqSeg->rRange);
                        }
                        if (reqSeg != NULL) {
                            delete reqSeg;
                            reqSeg = NULL;
                        }
                        if (mResponse != NULL) {
                            delete mResponse;
                            mResponse = NULL;
                        }
                    }
 
                }
                // send / relay the s2c_buffer
                while (relayLen > 0) {
                    if (dbg_info_level & DBG_INF_CONN) {
                        stats->lprintf(Stats::INFO, "Remaining relay length (2): %d\n", relayLen);
                    }
                    retval = send(client_socket, s2c_start, 
                                  std::min(send_length(s2c_buffer, s2c_start, s2c_end), relayLen), MSG_NOSIGNAL);
                    if(retval == SOCKET_ERROR) {
                        stats->lperror(Stats::WARN, "Error send() failed: ");
                        immediate = false;   // break the outer loop
                        break;
                    }
                    // retval is fine, update the buffer pointer
                    s2c_start = send_update_pt(s2c_buffer, s2c_start, s2c_end, retval);
                    relayLen -= retval;
                }
                if (dbg_info_level & DBG_INF_CONN) {
                    stats->lprintf(Stats::INFO, "Relay completed(2)\n");
                }
            }
            while (immediate == true && send_length(s2c_buffer, s2c_start, s2c_end) > 0); // if true, call the filter function again.
		}
	} //End While

cleanup:
    // Bear, 1/14/08, shouldn't need this
    if (pc != NULL) {
        if (admitted == false && avail == true) pc->decRefCnt();
        else if (admitted == true && avail == false) {
          mCache->abort(pc);
          pc = NULL;  // pc has been freed
        }          
    }
    // clean up unfinished page clusters
    if (mResponse != NULL) {
        for (; availableIter != mResponse->availableSegment->end(); ++availableIter) {
            pc = (*availableIter);
            pc->decRefCnt();
        }
        for (; requestedIter != mResponse->requestedSegment->end(); ++requestedIter) {
            pc = (*requestedIter);
            // pc->decRefCnt(); // done in abort
            mCache->abort(pc);
            pc = NULL;  // pc has been freed
        }
        delete mResponse;
        mResponse = NULL;
    }
	if (client_socket != INVALID_SOCKET) close(client_socket);
	if (server_socket != INVALID_SOCKET) close(server_socket);
	/********** select/poll finished **********/
}


void Conn::cacheRequest(char* from, int len) {
    if (len >= BUFSIZE) len = BUFSIZE - 1;
    memcpy(reqBuf, from, len);
    reqLen = len;
    reqBuf[len] = 0;
}

void Conn::cacheResponse(char* from, int len) {
    if (len >= BUFSIZE) len = BUFSIZE - 1;
    memcpy(resBuf, from, len);
    resLen = len;
    resBuf[len] = 0;
}

void Conn::cacheRequest(char *circbuff_ptr1, char *circbuff_cptr, char *circbuff_ptr3)
//caches data in the circular buffer from address circbuff_ptr1 to circbuff_ptr3
{
	bool two_part = circbuff_ptr3 < circbuff_ptr1;		//whether data is splitted over the circular buffer or not
														//(*circbuff_ptr3==*circbuff_ptr1 means an empty buffer)
	int len1 = two_part ? (circbuff_cptr + BUFSIZE - circbuff_ptr1) :
							(circbuff_ptr3 - circbuff_ptr1);		//length of the first part of data
	int len2 = two_part ? (circbuff_ptr3 - circbuff_cptr) : 0;		//length of the second part of the data (if exists)

	memcpy(reqBuf, circbuff_ptr1, len1);
	memcpy(reqBuf + len1, circbuff_cptr, len2);
    reqLen = len1 + len2;
    reqBuf[len1 + len2] = 0;
}

void Conn::cacheResponse(char *circbuff_ptr1, char *circbuff_cptr, char *circbuff_ptr3)
//caches data in the circular buffer from address circbuff_ptr1 to circbuff_ptr3
{
	bool two_part = circbuff_ptr3 < circbuff_ptr1;		//whether data is splitted over the circular buffer or not
														//(*circbuff_ptr3==*circbuff_ptr1 means an empty buffer)
	int len1 = two_part ? (circbuff_cptr + BUFSIZE - circbuff_ptr1) :
							(circbuff_ptr3 - circbuff_ptr1);		//length of the first part of data
	int len2 = two_part ? (circbuff_ptr3 - circbuff_cptr) : 0;		//length of the second part of the data (if exists)

	memcpy(resBuf, circbuff_ptr1, len1);
	memcpy(resBuf + len1, circbuff_cptr, len2);
    resLen = len1 + len2;
    resBuf[len1 + len2] = 0;
}


void Conn::push_req(Segment* seg){
    if (dbg_info_level & DBG_INF_CONN) 
        stats->lprintf(Stats::INFO, "Pushing request: %s\n", seg->toString().c_str());
    pthread_mutex_lock(&mut);
    reqQueue.push_back(seg);
    pthread_mutex_unlock(&mut);
}

void Conn::update_req(long long old_piece_len, long long new_piece_len) {
    pthread_mutex_lock(&mut);
    std::list<Segment*>::iterator iter;
    for (iter = reqQueue.begin(); iter != reqQueue.end(); ++iter) {
        // update lRange, rRange, and piecelen
        assert((*iter)->piecelen == old_piece_len);
        long long _old_piece_num = (long long) floor((float)(*iter)->lRange / (float) old_piece_len);
        long long _old_residue = (*iter)->lRange % (long long)old_piece_len;
        long long _old_seg_len = (*iter)->rRange - (*iter)->lRange;
        // new semgnet:
        // [_old_piece_num * new_piece_length + _old_residue, _old_piece_num * new_piece_length + old_piece_len)
        (*iter)->lRange = _old_piece_num * new_piece_len + _old_residue;
        (*iter)->rRange = (*iter)->lRange + _old_seg_len;
        (*iter)->piecelen = new_piece_len;
    }
    pthread_mutex_unlock(&mut);
}

// The following function calls all active Conn objects to update segments in their queue
void Conn::update_all_req(long long old_piece_len, long long new_piece_len){
    pthread_mutex_lock(&smut);
    std::set<Conn*>::iterator _pos; 
    for (_pos = conns.begin(); _pos != conns.end(); ++_pos) {
        (*_pos)->update_req(old_piece_len, new_piece_len);
    }
    pthread_mutex_unlock(&smut);
}

// This function do the polling thing for tcp splicing connection
void Conn::do_tcp_splicing(int fd1, int fd2, char* buf, int len) {
    int n;
    struct pollfd pollinfo[2]; 
    pollinfo[0].fd = fd1;
    pollinfo[0].events = POLLIN | POLLERR | POLLHUP;
    pollinfo[0].revents = 0;
    pollinfo[1].fd = fd2;
    pollinfo[1].events = POLLIN | POLLERR | POLLHUP;
    pollinfo[1].revents = 0;

    // this is cheap!
    bool fd1_done = false;
    bool fd2_done = false;
    bool more_to_read = false;

   // printf("Start polling fd1: %d, fd2: %d\n", fd1, fd2);
    while (fd1_done == false && fd2_done == false) {
        // We can potentially use a very long wait
        // This is because the actual work has been done in kernel
        // We are only waiting for a notify saying it's done
        int i = poll(pollinfo, 2, 30000);  
        if (i > 0) {
            if ((pollinfo[0].revents & POLLIN)) {
                more_to_read = true;   
                while (more_to_read == true) {
                    more_to_read = false;
                    n = read(fd1, buf, len);
                    if (n > 0) {
                        //printf("read %d bytes from sock1 and write to sock2\n", n);
                        // OK, it's not clear how can we read something here, didn't
                        // we always take a shortcut in iptables?
                        // Note that, we have to write whatever we get, otherwise
                        // TCP connections get frozen
                        //send(fd2, buf, n, MSG_NOSIGNAL);
                        more_to_read = true;
                    }
                    else if (n == 0) {
                        //perror("read 0 bytes from sock1");
                        fd1_done = true;
                    }
                    else {
                        //perror("read < 0 bytes from sock1");
                        fd1_done = true;
                    }
                }
                pollinfo[0].events = 0;
                pollinfo[0].revents = 0;
            }
            else if (pollinfo[0].revents & POLLERR || 
                pollinfo[0].revents & POLLHUP) {
                perror("sock 1 ERR or HUP");
                fd1_done = true;
                pollinfo[0].events = 0;
                pollinfo[0].revents = 0;
                close(fd1);
            }

            if ((pollinfo[1].revents & POLLIN)) {
                more_to_read = true;   
                while (more_to_read == true ) {
                    more_to_read = false;
                    n = read(fd2, buf, len);
                    if (n > 0) {
                       // printf("read %d bytes from sock2 and write to sock1\n", n);
                        //send(fd1, buf, n, MSG_NOSIGNAL);
                        more_to_read = true;
                    } else if (n == 0) {
                        //perror("read 0 bytes from sock2");
                        fd2_done = true;
                    } else {
                        //perror("read < 0  bytes from sock2");
                        fd2_done = true;
                    }
                }
                pollinfo[1].events = 0;
                pollinfo[1].revents = 0;
            }
            else if (pollinfo[1].revents & POLLERR || 
                pollinfo[1].revents & POLLHUP) {
                perror("sock 2 ERR or HUP");
                fd2_done = true;
                pollinfo[1].events = 0;
                pollinfo[1].revents = 0;
                close(fd2);
            }
        }
    }
    //printf("Stop polling fd1: %d, fd2: %d\n", fd1, fd2);
}

// This function returns whether any requests (even partial) are required
bool Conn::complete_avail(std::list<PageCluster*> *avail_list, Segment *seg) {
    long long available = 0; 
	std::list<PageCluster*>::iterator iter = avail_list->begin();
    for (;iter != avail_list->end(); ++iter) {
        available += std::min((*iter)->getRight(), seg->rRange) - std::max((*iter)->getLeft(), seg->lRange);
    }
    return (available >= (seg->rRange - seg->lRange));
}

