#ifndef __LRUCACHE_H
#define __LRUCACHE_H

#include <deque>
#include <list>
#include <string>
#include <algorithm>

#include "Cache.h"
#include "Segment.h"
#include "Item.h"

class LRUQueue {
public:
    LRUQueue() :
        first(NULL), last(NULL), size(0)
    {

    }

    PageCluster* first; // points to first page cluster, NULL for empty list
    PageCluster* last;  // points to last page cluster, NULL for empty list
    int size;   

    inline void insertAfter(PageCluster *_old_pc, PageCluster *_new_pc) {
        _new_pc->aux_prev = _old_pc;
        _new_pc->aux_next = _old_pc->aux_next;
        if (_old_pc->aux_next == NULL) {
            _old_pc->aux_next = _new_pc;
            last = _new_pc;
        }
        else {
            _old_pc->aux_next->aux_prev = _new_pc;
        }
        _old_pc->aux_next = _new_pc;
        ++size;
    }

    inline void insertBefore(PageCluster *_old_pc, PageCluster *_new_pc) {
        _new_pc->aux_prev = _old_pc->aux_prev;
        _new_pc->aux_next = _old_pc;
        if (_old_pc->aux_prev == NULL) {
            _old_pc->aux_prev = _new_pc;
            first = _new_pc;
        }
        else {
            _old_pc->aux_prev->aux_next = _new_pc;
        }
        _old_pc->aux_prev = _new_pc;
        ++size;
    }

    inline void insertHead(PageCluster *_new_pc) {
        if (first == NULL) {
            first = _new_pc;
            last = _new_pc;
            _new_pc->aux_prev = NULL;
            _new_pc->aux_next = NULL;
            ++size;
        }
        else {
            insertBefore(first, _new_pc);
        }
    }

    inline void insertTail(PageCluster *_new_pc) {
        if (last == NULL) {
            insertHead(_new_pc);
        }
        else {
            insertAfter(last, _new_pc);
        }
    }

    // note that, remove a page cluster from LRUQueue does not free the memory
    inline void remove(PageCluster *_old_pc) {
        if (_old_pc->aux_prev == NULL) {
            first = _old_pc->aux_next;
        }
        else {
            _old_pc->aux_prev->aux_next = _old_pc->aux_next;
        }
        if (_old_pc->aux_next == NULL) {
            last = _old_pc->aux_prev;       
        }
        else {
            _old_pc->aux_next->aux_prev = _old_pc->aux_prev;
        }
        --size;
    }

    // push is an alias of insertTail
    inline void push(PageCluster *_pc) { insertTail(_pc); }
    // pop gets the head item (but not remove it from the list)
    inline PageCluster* pop() { return first; }
    // refresh moves a pc to the tail
    inline void refresh(PageCluster *_pc) {
        remove(_pc);
        insertTail(_pc);
    }
    // size returns the number of page clusters
    inline int get_size() { return size; }
    // get_next helps on traversals
    inline PageCluster* get_next(PageCluster *current) { return current->aux_next;}
};

class LRUCache : public Cache
{
public:	
	LRUCache();								//	default constructor
	void free_init();						//	initialize the candidate iterators
	PageCluster* getFreeCandidate();		//	find a candidate PageCluster to be removed
	void freeCurrent();						//	remove the current PageCluster which candidate
											//	iterators point to, from the cache

	void removeItem(PageCluster *_pc);
	void insertItem(PageCluster *_pc, long long serverSize, long long pcServedSize);
	void abortPC(PageCluster* _pc);
	void _free();
    inline bool item_based() {return false; };
    inline int policy() { return LRU; };

private:
    // 07/12/30 Cheng, we use page clusters (instead of items) as the basic unit of replacement policy.
    LRUQueue q;
    //	iterator to the page cluster candidate to be removed from the cache in order to
    //	find enough continuous disk space.  This should be checked by BlockPool
    //	whenever it wants to find a continuous disk space.
    PageCluster *removePcCandidate;
	bool initChk;
	void printQueueSize(); //	check the queue size
};

#endif	//	__LRUCACHE_H
