#include "P2PCache.h"
#include "CacheManager.h"
#include "pCache.h"
#include <iostream>
#include <cmath>
#include <limits>

//	Calling the parent default constructor
P2PCache::P2PCache() : Cache()
{
	initial_byte = atol(confMap[std::string("P2P_INITIAL_BYTE")].c_str());
	expo_factor = atol(confMap[std::string("P2P_EXPO_FACTOR")].c_str());
}

long long P2PCache::admit_limit(Item* _item) {
    if (_item == NULL) return initial_byte; 
    else return initial_byte * (long long)pow((double)expo_factor, (double)_item->get_num_hit() - 1);
}

long long P2PCache::admit(Segment* _seg){
    // fast warm-up policy.
	if (size + (_seg->rRange - _seg->lRange) <= min_filled_size) return _seg->rRange - _seg->lRange;

    Item *_item = cacheManager->getItem(_seg->objId);
    long long residue = admit_limit(_item);
    if (_item != NULL) {
        residue -= _item->size;
    }
    return std::max(0LL, std::min(_seg->rRange - _seg->lRange, residue));
}


void P2PCache::removeItem(PageCluster *_pc)
{
    assert(_pc->item);
	if (!q.empty())
		//	find the place of Item
		for (std::multiset<Item*>::iterator _itemItr = q.begin(); _itemItr != q.end(); ++_itemItr)
            if ((*_itemItr) == _pc->item) {
				q.erase(_itemItr);
				break;
			}
}

void P2PCache::abortPC(PageCluster* _pc)
{
    assert(_pc->item);
	if (!q.empty())
		//	find the place of Item
		for (std::multiset<Item*>::iterator _itemItr = q.begin(); _itemItr != q.end(); ++_itemItr)
            if ((*_itemItr) == _pc->item) {

                // update hit record
	            (*_itemItr)->total_hit -= (_pc->length * _pc->num_hit);
				//	in addition the PageCluster should be removed from the CacheManager's Hash Table
				cacheManager->remove(_pc->objId, _pc);
                // update item frequency
	            (*_itemItr)->freq = (double)(*_itemItr)->total_hit / (double)(*_itemItr)->size;

				if ((*_itemItr)->pcList.empty()) {
					delete (*_itemItr);
					q.erase(_itemItr);
				}
				break;
			}
}

//	we don't need to use the servedSize for P2P
void P2PCache::insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize)
{
    assert(_pc->item);
    assert(_pc->item->pcList.empty() == false);
    assert(servedSize >= 0);
    assert(pcServedSize >= 0);
    Item *_item = _pc->item;
	if (_item->size > 0) _item->freq = (double)_item->total_hit / (double)_item->size; // compute the gamma value
	q.insert(_item);
}

void P2PCache::updateFreq(PageCluster *_pc, long long servedSize, long long pcServedSize)
{
    _pc->update_freq(pcServedSize);         // update per pc freq
	_pc->item->total_hit += pcServedSize;  // update per item freq
}

//	Remove a PageCluster from the candidate Item in the queue
void P2PCache::_free()
{
	std::multiset<Item*>::iterator			_itemItr;
	std::multiset<Item*>::iterator			tempItr;
	Item*			_item;
	PageCluster*	_pc;
	std::set<PageCluster*, DereferenceLess>::iterator	_pcItr;
	std::set<PageCluster*, DereferenceLess>::iterator	_pcVic;

	//	For P2P policy select the Item at the bottom of the queue
	_itemItr = q.begin();
	while (size > min_filled_size && _itemItr != q.end()) {
        // we find the page cluster with the smallest gamma value that is not
        // in-used
        _item = *_itemItr;
        double min_gamma = std::numeric_limits<double>::max();
        for (_pcItr = _item->pcList.begin(); _pcItr != _item->pcList.end(); ++_pcItr) {
            if (victim_pc(*_pcItr) == true && (*_pcItr)->gamma < min_gamma) {
                _pcVic = _pcItr;
                min_gamma = (*_pcVic)->gamma;
            }
        }
        // if we find a good victim
        if (min_gamma < std::numeric_limits<double>::max()) {  
            _pc = *_pcVic;

            // update item hit record
//printf("before total_hit: %lld, pc->length: %lld, num_hit: %d\n", _item->total_hit, _pc->length, _pc->num_hit);
	        _item->total_hit -= (_pc->length * _pc->num_hit);
//printf("after total_hit: %lld, pc->length: %lld, num_hit: %d\n", _item->total_hit, _pc->length, _pc->num_hit);

	        //stats->lprintf(Stats::INFO, "P2PCache: Remove PageCluster with ObjID: %s", _pc->objId.c_str());
            //	in addition the PageCluster should be removed from the CacheManager's Hash Table
            cacheManager->remove(_pc->objId, _pc);

//printf("before item->freq: %10.3f, item->size: %lld\n", _item->freq, _item->size);
            // compute freq after cacheManager->remove() updates the item size
	        if (_item->size > 0) _item->freq = (double)_item->total_hit / (double)_item->size;
//printf("after item->freq: %10.3f, item->size: %lld\n", _item->freq, _item->size);

            //	if there is no PageCluster remains for the Item
            //	we need to remove the whole Item from the queue
            if (_item->pcList.empty()) {
				tempItr = _itemItr;
				_itemItr++;
                q.erase(tempItr);
                delete(_item);
            }
        }
		else	//	no more eligible page cluster
			_itemItr++;
	}
}

void P2PCache::free_init()
{
    pthread_mutex_lock(&qmut);
	std::set<PageCluster*, DereferenceLess>::iterator	_pcItr;
    Item *_item;
	if (q.empty()) {
		std::cout << "WoW: queue is emtpy and yet you want to free the queue!" << std::endl;
		std::cout << "Cache size: " << size << std::endl;
        pthread_mutex_unlock(&qmut);
		return;
	}

	removeItemCandidate	= q.begin();   
	removePcCandidate	= (*removeItemCandidate)->pcList.begin(); // default one

    // could we find a better victim?
    double min_gamma = std::numeric_limits<double>::max();
    for (_pcItr = _item->pcList.begin(); _pcItr != _item->pcList.end(); ++_pcItr) {
        if (victim_pc(*_pcItr) == true && (*_pcItr)->gamma < min_gamma) {
            removePcCandidate = _pcItr;
            min_gamma = (*_pcItr)->gamma;
        }
    }
	initChk = true;
    pthread_mutex_unlock(&qmut);
}

PageCluster* P2PCache::getFreeCandidate()
{
    pthread_mutex_lock(&qmut);
	if (!initChk)
		removePcCandidate++;
	else
		initChk = false;

	while (removeItemCandidate != q.end() && !q.empty()) {
		while (removePcCandidate != (*removeItemCandidate)->pcList.end()) {
			if ((*removePcCandidate)->refCnt <= 0) {	//	we have a candidate to be removed
                pthread_mutex_unlock(&qmut);
				return *removePcCandidate;
			}
			removePcCandidate++;
		}
		removeItemCandidate++;

		//	we assume that we never have an Item with empty PC list
		if (removeItemCandidate != q.end()) removePcCandidate = (*removeItemCandidate)->pcList.begin();
	}
    pthread_mutex_unlock(&qmut);
	return NULL;
}

void P2PCache::freeCurrent()
{
    pthread_mutex_lock(&qmut);
	Item*	_item = *removeItemCandidate;
	std::multiset<Item*>::iterator tempItr;

	PageCluster *_pc = *removePcCandidate;
	removePcCandidate++;
	initChk = true;

    // update total hit byte
    _item->total_hit -= (_pc->length * _pc->num_hit);
	//	in addition the PageCluster should be removed from the CacheManager's Hash Table
    cacheManager->remove(_pc->objId, _pc);
    // update item frequency
    if (_item->size > 0) _item->freq = (double)_item->total_hit / (double)_item->size;

    //	if there is no PageCluster remains for the Item
    //	we need to remove the Item from the queue
    if (_item->pcList.empty()) {
		tempItr = removeItemCandidate;
		removeItemCandidate++;
		if (removeItemCandidate != q.end()) {
			removePcCandidate = (*removeItemCandidate)->pcList.begin();
		}
        q.erase(tempItr);
        delete(_item);
    }
    pthread_mutex_unlock(&qmut);
}

void P2PCache::printQueueSize()
{
	//long long _size = 0;
    int i = 0;

    printf("Dumping queue: \n");
	for (std::multiset<Item*, DereferenceLess>::iterator _itr = q.begin(); _itr!= q.end(); ++_itr) {
		//_size += (*_itr)->size;
        printf("item %d is %s\n", ++i, (*_itr)->toString().c_str());
	}	
    printf("Finished queue: \n");

	//std::cout << "Queue size: " << _size << "   Cache size: " << size << std::endl;
}
