#include <cmath>
#include <cassert>
#include <iostream>
#include <malloc.h>
#include <algorithm>
#include "PageCluster.h"
#include "pCache.h"
#include "Util.h"

// The disk buffer is shared by all page clusters keeps track of current
// disk_buf size. disk_buf is reallocated whenever apce is not enough
long long PageCluster::disk_buf_length = 1048576;   

// disk write buffer array, we aggreate all data from pages first then commit
// them to the disk in one shot. This buffer is shared by all page cluster, and
// should be protected by cache manager's semaphore
char *PageCluster::disk_buf = (char*)memalign(4096, PageCluster::disk_buf_length);

// page cluster is created when a requested segment is admitted into our system
PageCluster::PageCluster(std::string _objId, long long _left, long long _right) :
    objId(_objId), offset(_left), length(_right - _left), refCnt(1), blocks(NULL), 
    item(NULL), swapping_out(false), sec_chanced(false), delayed_removal(false), 
    state(NEW), gamma(0), num_hit(0), aux_prev(NULL), aux_next(NULL)
{
	assert(length >= 0);

    // allocate memory pages, we only allocate page cluster when some segment
    // is admitted into the system. To accomodate the data, memory buffer is
    // always allocated
    if (1 == allocate_pages(true)) {  // we can force releasing pages because we are
                                      // still in constructor, and no one would have
                                      // our reference
        // Uah, we see an exception, but the caller (LRUCache) will
        // handle this...add exceptions later on
        decRefCnt();
        std::cout << "Requested length: " << length << std::endl;
    }
}

// page cluster is removed when it is evicted by the disk cache algorithm
PageCluster::~PageCluster() {
    pre_destructor();
}

void PageCluster::pre_destructor() {
    // free blocks cluster here
    // remove a file (if there is any) and return the blocks interval
    if (blocks != NULL) {
        blockPool->accessor->blk_remove(objId, offset, blocks->start_num, blocks->no_blocks);
        stats->num_disk_bytes -= length;
        stats->disk_byte_utilization = (float)stats->num_disk_bytes / 
                                       (float)((long long)blockPool->numBlocks * (long long)blockPool->blockSize);;
        stats->num_disk_blocks -= blocks->no_blocks;
        stats->disk_block_utilization = (float)stats->num_disk_blocks / (float)blockPool->numBlocks;
        stats->num_file--;
        blockPool->deallocate(blocks);  
        blocks = NULL;
    }
    // if we have more than one page, then return them to PagePool
    if (!pages.empty()) {
        deallocate_pages(false);
    }
}

// requests for memory pages 
// This function returns 0 for successful allocation and 1 for out of memory
int PageCluster::allocate_pages(bool force) {
    // loop counters
    int i;
    long long _offset = 0;  // offset counter, to compute the map's key
    // first compute the number of required pages
    int _lNumPage = (int) floor((float)length / (float)pagePool->lPageSize);
    if (_lNumPage > pagePool->lPageNum) _lNumPage = pagePool->lPageNum;
    int _mNumPage = (int) floor((float)(length - _lNumPage * pagePool->lPageSize) / 
                          (float) pagePool->mPageSize);
    if (_mNumPage > pagePool->mPageNum) _mNumPage = pagePool->mPageNum;
    int _sNumPage = (int) ceil((float)(length - _lNumPage * pagePool->lPageSize - 
                                 _mNumPage * pagePool->mPageSize) / 
                          (float) pagePool->sPageSize);
    // the following line may cause problem, although unlikely.
    if (_sNumPage > pagePool->sPageNum) _sNumPage = pagePool->sPageNum;

//    std::cout << "allocating L/M/S pages = " << _lNumPage << "/" << 
//		           _mNumPage << "/" << _sNumPage << std::endl;

    // allocate pages from PagePool
    for (i = 0; i < _lNumPage; i++) {
        Page *_page = pagePool->allocate(this, PagePool::LPAGE);
        if (_page != NULL) {  // if allocate is ok
            pages.insert(std::pair<long long, Page*>(_offset, _page));
            _offset += pagePool->lPageSize;
        } else { // otherwise
            deallocate_pages(force);   // may have different size of pages
			//	return 1 means unsuccessful page allocation
			return 1;
        }
    }
    for (i = 0; i < _mNumPage; i++) {
        Page *_page = pagePool->allocate(this, PagePool::MPAGE);
        if (_page != NULL) {  // if allocate is ok
            pages.insert(std::pair<long long, Page*>(_offset, _page));
            _offset += pagePool->mPageSize;
        } else { // otherwise
            deallocate_pages(force);   // may have different size of pages
			//	return 1 means unsuccessful page allocation
			return 1;
        }
    }
    for (i = 0; i < _sNumPage; i++) {
        Page *_page = pagePool->allocate(this, PagePool::SPAGE);
        if (_page != NULL) {  // if allocate is ok
            pages.insert(std::pair<long long, Page*>(_offset, _page));
            _offset += pagePool->sPageSize;
        } else { // otherwise
            deallocate_pages(force);   // may have different size of pages
			//	return 1 means unsuccessful page allocation
			return 1;
        }
    }
	return 0;
}

// returns all pages
void PageCluster::deallocate_pages(bool force) {
    // return pages to pagePool. Notice that the ref cnt should be zero by now.
    if (force == false) {
        assert(refCnt <= 0);
    }
    else {
        refCnt = 0;  // force clean up -- abort
    }
    // return the pages..
    std::map<long long, Page*>::iterator _pos;   // to traverse through the page list
    for (_pos = pages.begin(); _pos != pages.end(); ++ _pos) {
        if ((_pos->second)->size == pagePool->lPageSize) {
            pagePool->deallocate(_pos->second, PagePool::LPAGE);
        } 
        else if ((_pos->second)->size == pagePool->mPageSize) {
            pagePool->deallocate(_pos->second, PagePool::MPAGE);
        }
        else {// if ((*(_pos->second))->size == pagePool->sPageSize) {
            pagePool->deallocate(_pos->second, PagePool::SPAGE);
        }
    }
    // clear the list
    pages.clear();
}

// following two access methods can be further optimizaed as shared buffer (to avoid memcpy)
void PageCluster::write(long long _in_pc_offset, long long _length, char* _from) {
    if (_in_pc_offset + _length > length) {
        std::cout << "Bug #1, _in_pc_offset: " << _in_pc_offset << " _length: " << _length;
        std::cout << " PC: " << toString() << std::endl;
        assert(_in_pc_offset + _length <= length);
    }
    //	std::cout << _in_pc_offset << " " << _length << std::endl;
    std::map<long long, Page*>::iterator _iter;       // current page
    std::map<long long, Page*>::iterator _iter_next;  // next page
    // find the first page 
    _iter = pages.upper_bound(_in_pc_offset);   // _iter points to the next item!
    --_iter; // _iter points to the first page
    long long _external_offset = 0; // where to copy (in external buffer)
    // found the first page, copy the data
    for (; _iter != pages.end() && _length > 0; ++_iter) {
        long long _effective_page_length; // length of a page
        long long _copy_length;     // how many bytes do we want to copy
        long long _in_page_offset;  // where to copy (in the page)
        _iter_next = _iter; ++_iter_next;  // prepare the next item
        if (_iter_next == pages.end()) { // the copy _length is limited by the pc length
            _effective_page_length = length -_iter->first;
        }
        else {  // max copy length is defined by the diff of next and current item
                // _iter_next->first - _iter->first means the actual data length store on the page...
            _effective_page_length = _iter_next->first - _iter->first;
        }
        // shall we start in the middle of the page?
        _in_page_offset = std::max(0LL, _in_pc_offset - _iter->first);
        // copy length is limited by the effective page length and remaining external buffer length
        _copy_length = std::min(_effective_page_length - _in_page_offset, _length);
        // move the data
#ifndef	NO_DATA
        memcpy(_iter->second->data + _in_page_offset, _from + _external_offset, _copy_length);
#endif
        // move the pointers
        _length -= _copy_length;
        _external_offset += _copy_length;
    }
}

void PageCluster::read (long long _in_pc_offset, long long _length, char* _to) {
    assert(_in_pc_offset + _length <= length);
    //	std::cout << _in_pc_offset << " " << _length << std::endl;
    std::map<long long, Page*>::iterator _iter;       // current page
    std::map<long long, Page*>::iterator _iter_next;  // next page
    // find the first page 
    _iter = pages.upper_bound(_in_pc_offset);   // _iter points to the next item!
    --_iter; // _iter points to the first page
    long long _external_offset = 0; // where to copy (in external buffer)
    // found the first page, copy the data
    for (; _iter != pages.end() && _length > 0; ++_iter) {
        long long _effective_page_length; // length of a page
        long long _copy_length;     // how many bytes do we want to copy
        long long _in_page_offset;  // where to copy (in the page)
        _iter_next = _iter; ++_iter_next;  // prepare the next item
        if (_iter_next == pages.end()) { // the copy _length is limited by the pc length
            _effective_page_length = length -_iter->first;
        }
        else {  // max copy length is defined by the diff of next and current item
                // _iter_next->first - _iter->first means the actual data length store on the page...
            _effective_page_length = _iter_next->first - _iter->first;
        }
        // shall we start in the middle of the page?
        _in_page_offset = std::max(0LL, _in_pc_offset - _iter->first);
        // copy length is limited by the effective page length and remaining external buffer length
        _copy_length = std::min(_effective_page_length - _in_page_offset, _length);
        // move the data
#ifndef	NO_DATA
        memcpy(_to + _external_offset, _iter->second->data + _in_page_offset, _copy_length);
#endif
		// move the pointers
        _length -= _copy_length;
        _external_offset += _copy_length;
    }
}

// for the following two swapping functions, we now assume no two page clusters would
// share the same blocks. So, blocks do not contain any meta data right now.
// write contents to the disk
void PageCluster::swapout() {
    // valid states include NOT_IN_MEM, IN_MEM_CLEAN, IN_MEM_DIRTY
    assert(FILLED == state || IN_MEM_CLEAN == state || IN_MEM_DIRTY == state);
    assert(refCnt == 0 && !pages.empty());
    // before locking this page, first memorize whether we are dirty
    bool dirtyBit = isDirty(); 
    // We get into the swapping state (locked mode)
    swapping_out = true; // lock this page

    // 1. find a starting blocks number (from BlockPool)
    // 2. write the blocks.
    // 3. return pages 
    if (dirtyBit == true || blockPool->accesorType == 6) { // these pages have been written...
        // 1 and 2 are the same as coss_write(...), so we use the function 
        coss_write();  
    } 
    // return memory pages
    if (!pages.empty()) {
        deallocate_pages(false);
    }
    swapping_out = false; // unlock this page
    // We get out of this function with not-in-mem state
    state = NOT_IN_MEM;
    sec_chanced = false; // we erase the second chance flag here
}


// read contents from the disk
// returns status code
bool PageCluster::swapin() {
    // make sure we are not in memory
    assert(NOT_IN_MEM == state);
    assert(pages.empty() && blocks != NULL);

    // 1. allocate pages
    // 2. read and parse blocks.
    assert(blocks != NULL);

    // before calling allocate_pages, we mark our delayed_removal bit, this is to prevent
    // COSS from freeing us
    delayed_removal = true;

    if (pages.empty()) {
        int result = allocate_pages(true);  // we need to check result code, if use forced mode
        if (result == 1) {
            assert(0);
            return true;    // not much we can do, the caller will handle this
                            // note, we failed, but the caller will check our
                            // pages to see whether a recovery is required
        }
    }
    // this is for COSS, in some cases, during swapin, the content on disk is overwritten
    // we have to abort and delete the page cluster
    if (blocks == NULL) {
        // this means (because of the delayed_removal is on), all resources are freed
        // page cluster is removed from the object table, however, pc itself is not
        // freed. Need to do that in the caller, when pc is not on the stack anymore.
        return false;
    } 
    // unset the lock
    delayed_removal = false;
    // check the read buffer size to see whether we need to enlarge it
    if (length > disk_buf_length) {
        disk_buf_length = length;
        free(disk_buf);
        disk_buf = (char*)memalign(4096, PageCluster::disk_buf_length);
    }
    // read from the disk
    assert(blocks != NULL);
    blockPool->accessor->blk_read(objId, offset, blocks->start_num, blocks->no_blocks, disk_buf);
    stats->ap_read_len_in_bytes += length;
    stats->ap_read++;
    // copy data to memory pages
    // unlike swapout, swapin never faces partially-filled buffer pages, we
    // don't need to cacluate the page length
    std::map<long long, Page*>::iterator _pos;
    long long _offset = 0;  // keep track how far we have written in swap out buf
    for (_pos = pages.begin(); _pos != pages.end(); ++_pos) {
#ifndef	NO_DATA
        memcpy((_pos->second)->data, disk_buf + _offset, std::min(length - _offset, (_pos->second)->size));
#endif
        _offset += std::min(length - _offset, (_pos->second)->size);
    }
    // we get into the memory
    if (pages.empty() == false) { 
        state = IN_MEM_CLEAN;
        sec_chanced = false;  // reset the second chance flag
    }

    // in the case of COSS we always need to write to the tail to emulate LRU queue
    // Do a coss write if necessary
    if (blockPool->accesorType == 6) { 
        coss_write();
    }
    // break the chained page cluster 
    blocks->first_pc = NULL;
    blocks->next_pc = NULL;

    return true;  // OK
}

void PageCluster::coss_write() {
    if (blocks != NULL) {
        // for simplicity, we always return the blocks, 
        // this is because page cluster may grow because of merges
        blockPool->accessor->blk_remove(objId, offset, blocks->start_num, blocks->no_blocks);
        stats->num_disk_bytes -= length;
        stats->disk_byte_utilization = (float)stats->num_disk_bytes / 
                                       (float)((long long)blockPool->numBlocks * (long long)blockPool->blockSize);;
        stats->num_disk_blocks -= blocks->no_blocks;
        stats->disk_block_utilization = (float)stats->num_disk_blocks / (float)blockPool->numBlocks;
        stats->num_file--;
        blockPool->deallocate(blocks);  
        blocks = NULL;
    }
    // IMPORTANT: The following allocation may lead to eviction of page
    //            clusters that include us as a target, we use the locking 
    //            state SWAPPING_OUT to indicate that the current page cluster 
    //            should not be removed from the system
    assert(blocks == NULL);
    blocks = blockPool->allocate(length, objId, this); // no available hint
    assert(blocks != NULL);

    // check the write buffer size to see whether we need to enlarge it
    if (length > disk_buf_length) {
        disk_buf_length = length;
        free(disk_buf);
        disk_buf = (char*)memalign(4096, PageCluster::disk_buf_length);
    }
    // copy data to swap out buffer
    // unlike swapin, swapout may face partially-filled buffer pages, we
    // need to cacluate the page length
    std::map<long long, Page*>::iterator _pos;
    std::map<long long, Page*>::iterator _pos_next;
    long long _offset = 0;  // keep track how far we have written in swap out buf
    long long _effective_page_length; // length of a page
    for (_pos = pages.begin(); _pos != pages.end(); ++_pos) {
        _pos_next = _pos; ++_pos_next;  // prepare the next item
        if (_pos_next == pages.end()) { // the copy _length is limited by the pc length
            _effective_page_length = length -_pos->first;
        }
        else {  // max copy length is defined by the diff of next and current item
            _effective_page_length = _pos_next->first - _pos->first;
        }
        long long _copy_length = std::min(_pos->second->size, _effective_page_length);
#ifndef	NO_DATA
        memcpy(disk_buf + _offset, (_pos->second)->data, std::min(length - _offset, _copy_length));
#endif
        _offset += std::min(length - _offset, _copy_length);
    }
    // commit the swap out buf to the disk
    blockPool->accessor->blk_write(objId, offset, blocks->start_num, blocks->no_blocks, disk_buf);
    stats->ap_write_len_in_bytes += length;
    stats->ap_write++;
    stats->num_disk_bytes += length;
    stats->disk_byte_utilization = (float)stats->num_disk_bytes / 
                                   (float)((long long)blockPool->numBlocks * (long long)blockPool->blockSize);;
    stats->num_disk_blocks += blocks->no_blocks;
    stats->disk_block_utilization = (float)stats->num_disk_blocks / (float)blockPool->numBlocks;
    stats->num_file++;
}

// dump info for debugging
std::string PageCluster::toString() {
    std::string ret("");
    ret.append("ObjId: ").append(objId);
    ret.append(" Offset: ").append(Util::stringify(offset));
    ret.append(" Length: ").append(Util::stringify(length));
    ret.append(" RefCnt: ").append(Util::stringify(refCnt));
    ret.append(" State: ").append(Util::stringify(state));
    // then is the blocks and the pages
    // Don't print blocks, leads to forever loops...
    ret.append(" Block: ");
    if (blocks == NULL) ret.append(" null"); else ret.append(blocks->toString(true)); // true for skipping page cluster
    ret.append(" Gamma: ").append(Util::stringify(gamma));
    ret.append(" Num_hit: ").append(Util::stringify(num_hit));
    std::map<long long, Page*>::iterator _pos;
    ret.append(" Page: | ");
    for(_pos = pages.begin(); _pos != pages.end(); ++_pos) {
        ret.append("<");
        ret.append(Util::stringify(_pos->first)).append(",");
        ret.append((_pos->second)->toString()).append("> | ");
    }
    ret.append(Util::CR_str);
    return ret;
}
void PageCluster::refresh() {
    std::map<long long, Page*>::iterator _pos;   // to traverse through the page list
    for (_pos = pages.begin(); _pos != pages.end(); ++ _pos) {
        pagePool->refresh(_pos->second);
    }
    sec_chanced = false; // we erase the second chance flag if this 
                         // page cluster is used again before being swapped out
    // Do a coss write if necessary
    if (blockPool->accesorType == 6) { 
        coss_write();
    }
}
// we only define less operator for sorted set
// we use left_offset to sort them
bool PageCluster::operator< (const PageCluster &other) const{
    return offset < other.offset;
}
// We define more for testing..
bool PageCluster::operator> (const PageCluster &other) const{
    return offset > other.offset;
}

// this function is called when a page cluster is received, we advance from state NEW to FILLED
void PageCluster::received() { 
    stats->lprintf(Stats::INFO, "PageCluster::received(...) is called: %s\n", toString().c_str());
    assert(NEW == state);
    assert(refCnt >= 1);
    state = FILLED;
    assert(refCnt >= 1);
}

// once data assembler finishes its job, this should be called (in the descructor?)
void PageCluster::decRefCnt() {
    --refCnt; 
    if (refCnt <0) refCnt = 0;
    // check whether merge is possible
    cacheManager->compact(this);
}

// Note: Callers of the following three functions need to acquire CacheManager's
// mutex before calling this.
//
// The following three functions are used to merge and un-merge page clusters
// This is required by piece-based protocols such as BitTorrent postpand() and
// prepand() are called in received() function (before changing state to
// FILLED), where a page cluster is just filled out by a Conn object. (Note
// that, when a page cluster is not filled out, only one Conn object has its
// reference: Other receivers will have this segment relayed because we haven't
// got the data)
// It is important that victim page clusters must have refCnt == 0 and 
// state == FILLED or NOT_IN_MEM or IN_MEM_CLEAN or IN_MEM_DIRTY. In the case
// of NOT_IN_MEM, a swap in is required first. If this is not satified, we can
// not merge page clusters since changing victim pages will affect other threads
// who are using these pages at the same moment. It is fine that we do not merge
// some pages because of these limitation, because merging is only an optimization
// and is not mandatory.
bool PageCluster::postpend(PageCluster *_victim) {
assert(0);
    assert(offset + length == _victim->offset);
    assert(objId == _victim->objId);
    assert(refCnt <= 1);
    if (_victim->refCnt > 0 || 
        (_victim->state != FILLED &&
         _victim->state != NOT_IN_MEM &&
         _victim->state != IN_MEM_CLEAN &&
         _victim->state != IN_MEM_DIRTY)) {
        // if the _victim page is not a good candidate, then abort
        //std::cout << "Victim page is not ready to be merged (postpend): " << _victim->toString() << std::endl;
        return false;
    }
    if (_victim->state == NOT_IN_MEM) {
        assert(0);      // don't support NO_IN_MEM anymore
        _victim->chained_swapin();
    }
    // move the pages from _victim to us
    std::map<long long, Page*>::iterator _pos; 
    while ((_pos = _victim->pages.begin()) != _victim->pages.end()) {
        // move this page to us, update the offset
        pages.insert(std::pair<long long, Page*>(_pos->first + length, _pos->second));
        // update its pc pointer
        _pos->second->pc = this;
        // remove it from _victim
        _victim->pages.erase(_pos);
    }
    // set up meta data
    length += _victim->length;
    // We don't deal with block clusters, the _victim's destructor will
    // deallocate them for us. However, we free our own blocks as they are not valid anymore
    if (blocks != NULL) {
        blockPool->accessor->blk_remove(objId, offset, blocks->start_num, blocks->no_blocks);
        stats->num_disk_bytes -= length;
        stats->disk_byte_utilization = (float)stats->num_disk_bytes / 
                                       (float)((long long)blockPool->numBlocks * (long long)blockPool->blockSize);;
        stats->num_disk_blocks -= blocks->no_blocks;
        stats->disk_block_utilization = (float)stats->num_disk_blocks / (float)blockPool->numBlocks;
        stats->num_file--;
        blockPool->deallocate(blocks);  
        blocks = NULL;
    }
    state = IN_MEM_DIRTY;  // we will need to rewrite this page cluster to the disk
    // next, we let the caller to free the victim page from the pcList and memory
    // Caller will remove page cluster from pcList and delete _victim
    //std::cout << "Merge is done (postpend): " << _victim->toString() << std::endl;
    return true;
}
// Prepend is more dangerous. We need to make sure that page clusters on the same Item pcList are
// access sequentially. Otherwise, we will free our previous pagecluster, and result in some
// wild pointers. We check the _victim->refCnt, if it's 0, then we are fine
bool PageCluster::prepend(PageCluster *_victim) {
assert(0);
    assert(offset == _victim->offset + _victim->length);
    assert(objId == _victim->objId);
    assert(refCnt <= 1);
    if (_victim->refCnt > 0 || 
        (_victim->state != FILLED &&
         _victim->state != NOT_IN_MEM &&
         _victim->state != IN_MEM_CLEAN &&
         _victim->state != IN_MEM_DIRTY)) {
        // if the _victim page is not a good candidate, then abort
        std::cout << "Victim page is not ready to be merged (prepend): " << _victim->toString() << std::endl;
        return false;
    }
    if (_victim->state == NOT_IN_MEM) {
        assert(0);      // don't support NO_IN_MEM anymore
        _victim->chained_swapin();
    }
    // reinsert our own pages with new offset, this need to be done before
    // inserting pages from _victim (o.w. overlapping may happend, and we will
    // have floating page cluster pointers).
    std::map<long long, Page*>::iterator _pos; 
    std::map<long long, Page*> tmp_pages; 
    while ((_pos = pages.begin()) != pages.end()) {
        // add the page to new pages map, which will be copied over to the main pages map
        tmp_pages.insert(std::pair<long long, Page*>(_pos->first + _victim->length, _pos->second));
        // remove it from _victim
        pages.erase(_pos);
    }
    pages = tmp_pages;   // copy over
    // move the pages from _victim to us
    while ((_pos = _victim->pages.begin()) != _victim->pages.end()) {
        // move this page to us, update the offset
        pages.insert(std::pair<long long, Page*>(_pos->first, _pos->second));
        // update its pc pointer
        _pos->second->pc = this;
        // remove it from _victim
        _victim->pages.erase(_pos);
    }
    // set up meta data
    offset = _victim->offset;
    length += _victim->length;
    // We don't deal with block clusters, the _victim's destructor will
    // deallocate them for us. However, we free our own blocks as they are not valid anymore
    if (blocks != NULL) {
        blockPool->accessor->blk_remove(objId, offset, blocks->start_num, blocks->no_blocks);
        stats->num_disk_bytes -= length;
        stats->disk_byte_utilization = (float)stats->num_disk_bytes / 
                                       (float)((long long)blockPool->numBlocks * (long long)blockPool->blockSize);;
        stats->num_disk_blocks -= blocks->no_blocks;
        stats->disk_block_utilization = (float)stats->num_disk_blocks / (float)blockPool->numBlocks;
        stats->num_file--;
        blockPool->deallocate(blocks);  
        blocks = NULL;
    }
    state = IN_MEM_DIRTY;  // we will need to rewrite this page cluster to the disk
    // next, we let the caller to free the victim page from the pcList and memory
    // Caller will remove page cluster from pcList and delete _victim
    //std::cout << "Merge is done (prepend): " << _victim->toString() << std::endl;
    return true;
}

// The function allocates a list of new page cluster with new piece length The
// caller will then insert these page cluster into pcList. The old page cluster
// (us) will be added into a garbage list. Page clusters on the garbage list
// will never be returned for future lookups. They are freed once their refCnt
// reaches zero.
// We always do copy over here, because for no-copy-over case, cache manager just
// has to move old page cluster to garbage list.
std::list<PageCluster*> PageCluster::split(long long _old_piece_len, 
                              long long _new_piece_len) {
    std::list<PageCluster*> ret;
    PageCluster * _pc;
    // The page cluster that subject to be split should never be NEW. If it's
    // NEW, then we have to drop it since no Conn object will try to fill it
    // for us.
    // if it's not in memory, then we have to swap it in first. 
     if (state != FILLED &&
         state != NOT_IN_MEM &&
         state != IN_MEM_CLEAN &&
         state != IN_MEM_DIRTY) {
        // if the page is not a good candidate, then abort
        std::cout << "Page is not ready to be split: " << toString() << std::endl;
        return ret;
    }
    if (state == NOT_IN_MEM) {
        chained_swapin();
        // in case swapin failed, we drop the page
        if (state == NOT_IN_MEM) {
            return ret;
        }
    }
    // allocate buffer
    char* _buf = new char[_old_piece_len];
    // New page clusters will have zero refernce count.
    long long _old_piece_num = (long long) floor((float)offset / (float) _old_piece_len);
    long long _old_residue = offset % (long long)_old_piece_len;
    // We now remap old pieces to new pages
    // for piece 0: [offset, offset + _old_piece_len - _old_residue) to
    // [_old_piece_num * new_piece_length + _old_residue, _old_piece_num * new_piece_length + _old_piece_len)
    long long _len_residue = length;
    long long _copy_len = std::min(_old_piece_len - _old_residue, _len_residue);
    _pc = new PageCluster(objId, (long long)_old_piece_num * (long long)_new_piece_len + 
                                                            (long long)_old_residue, 
                                 (long long)_old_piece_num * (long long)_new_piece_len + 
                                                            (long long)_old_residue + (long long)_copy_len);
    _len_residue -= _copy_len;
    if (NULL != _pc) {
        _pc->refCnt = 0;   // reset reference count
        _pc->state = state; // copy over state, should be either FILLED or IN_MEM_CLEAN
        // copy over data
        read (0, _copy_len, _buf);
        _pc->write(0, _copy_len, _buf);
        _pc->item = this->item;  // copy over the item pointer
        _pc->num_hit = this->num_hit;
        _pc->gamma = this->gamma;
        ret.push_back(_pc); 
    }
    // for piece i = 1, 2, .....
    // map [(_old_piece_num + i) * _old_piece_len, (_old_piece_num + i + 1) * _old_piece_len) to 
    //     [(_old_piece_num + i) * _new_piece_len, (_old_piece_num + i ) * _new_piece_len + old_piece_length)
    int i = 0; 
    while(_len_residue > 0) {
        _copy_len = std::min(_old_piece_len, _len_residue);
        _pc = new PageCluster(objId, (long long)(_old_piece_num + 1) * (long long)_new_piece_len ,  
                                     (long long)(_old_piece_num + 1) * (long long)_new_piece_len + (long long)_copy_len);
        _len_residue -= _copy_len;
        if (NULL != _pc) {
            _pc->refCnt = 0;   // reset reference count
            _pc->state = state; // copy over state, should be either FILLED or IN_MEM_CLEAN
            // copy over data
            read (0, _copy_len, _buf);
            _pc->write(0, _copy_len, _buf);
            _pc->item = this->item;  // copy over the item pointer
            _pc->num_hit = this->num_hit;
            _pc->gamma = this->gamma;
            ret.push_back(_pc); 
        }
        i++;
    }
    // free the buffer
    delete [] _buf;
    return ret;
}

bool PageCluster::chained_swapin() {
    
    if (blocks->first_pc != NULL) {
//int i = 0;
        PageCluster *_iter = blocks->first_pc;
        PageCluster *_tmp_iter;
        while(_iter != NULL) {
            assert(_iter->blocks != NULL);
            _tmp_iter = _iter->blocks->next_pc;
            if (blockPool->accesorType == 6) { 
                // We return whether this page cluster requires delayed free.  Note
                // that, we ignore the chained case, because this additional check is
                // only for COSS file system
                return _iter->swapin();
            }
            _iter->swapin();
//++i;
            assert(_iter->blocks != NULL);
            _iter = _tmp_iter;  // move to the next
        }
//printf("swapping in %d page clusters together\n", i);
    }
    else {  // this is a standalone item, so swap in only one pc
        // We return whether this page cluster requires delayed free.  Note
        // that, we ignore the chained case, because this additional check is
        // only for COSS file system
        return swapin();
    }
    return true;
}


