#ifndef __PAGECLUSTER_H
#define __PAGECLUSTER_H

#include <map>
#include <string>
#include <cassert>
#include "BlockPool.h"

class Page; // cyclic includes.
class Item;

// a page of pre-allocated memory
class PageCluster {
    private:
        static char *disk_buf;  // disk read/write buffer array, for write, we aggreate all data from pages first
                                // then commit them to the disk in one shot. Reverse procedure is used for read. 
                                // This buffer is shared by all page cluster, and should be protected by cache 
                                // manager's semaphore

        static long long	disk_buf_length;	// keeps track of current disk_buf size. disk_buf is reallocated
										// whenever it's size is not sufficient for swap operations.
        int			allocate_pages(bool force);	// requests for memory pages 
        void		deallocate_pages(bool force);	// returns memory pages, force == true if refCnt should
                                                    // be ignored 

    public:
        std::string objId;		// object id, which is fixed for each page cluster
        long long   offset;		// the offset (w.r.t. the whole shared object) of this page cluster
        long long   length;		// total allocated length. Note this could be smaller than the
								// sum of all the assigned page.size(), which is the capacity and
								// has nothing to do with data length.
        int         refCnt;		// reference count, how many data assemblers are using this page cluster 
        std::map<long long, Page*> pages; // keys point to the (in pc) offsets, while values point to the pages
        BlockCluster*    blocks;// points to the block cluster

        PageCluster(std::string _objId, long long _left, 
                    long long _right); // page cluster is created when a requested segment 
                                       // is admitted into our system
        ~PageCluster();     // page cluster is removed when it is evicted by the disk cache algorithm
                            // remember to free bc here
                            // pointer to the current item, that we belong to,
                            // NULL if we are not assigned any item
        void pre_destructor();   // This function does the actual destructor job
        Item *item;                            
        bool swapping_out; // an intermediate state, page clusters in this state are considered
                           // to be locked, and will not be freed using freeCurrent() method.
        bool sec_chanced;  // this bit is true, if the page cluster is marked to be swapped out
        bool delayed_removal;   // used by COSS, when a pagecluster is removed by the disk FIFO,
                                // we may not delete the page cluster immediately, because it is
                                // still on the stack. Wait until the swapin(...) returns false
                                // We delete that page cluster then
                                         // Each page cluster is in one of the following five states
        const static int NEW = 1;        // a page cluster is just defined (intentionally skip 0)
        const static int FILLED = 2;     // content of this page cluster has been received
        const static int NOT_IN_MEM = 4; // page cluster has been swapped out to the disk and not in memory
        const static int IN_MEM_CLEAN = 5; // page cluster is in memory and on disk
        const static int IN_MEM_DIRTY = 6; // page cluster is in memory only
        int state;

        double gamma;         // per page gamma value.
        int num_hit;          // how many time this page cluster has been requested, this is used
                              // for some admission policy, such as P2P
        inline void update_freq(long long hit_length) {
//printf("(1) before update gamme: %10.7f, hit_length: %lld, length: %lld, num_hit: %d\n", gamma, hit_length, length, num_hit);
            gamma += (double)hit_length / (double)length;
            num_hit++;
//printf("(1) after update gamme: %10.7f, hit_length: %lld, length: %lld, num_hit: %d\n", gamma, hit_length, length, num_hit);
        };
        
        // The following two auxiliary pointers are used in "real" linked-list
        // implementation.  This is required when we want to have a huge LRU
        // queue. Using STL leads to linear searches and kills the performance.
        // We (PageCluster class) don't do much except setting these two
        // pointers to NULL at init time.  Right now, only LRUCache uses these
        // pointers
        PageCluster *aux_prev; PageCluster *aux_next;

        // following two access methods can be further optimizaed as shared buffer (to avoid memcpy)
        void write(long long _in_pc_offset, long long _length, char* _from);
        void read (long long _in_pc_offset, long long _length, char* _to);
        void swapout();  
        bool swapin();      // returns status code
        void coss_write();  // COSS filesystem uses a huge file as FIFO, so, whenever there is a hit (swapin)
                            // or in memory hit, the content should be written to the head of the disk 
                            // We do this by calling coss_write, which deallocate the blocks (if any), allocate new blocks
                            // and write to the disk. This is only called when COSS filesystem is used
        bool chained_swapin();  // do the prefectch of all segments belong to one object, return false if pc needs to be freed 
        void received();
        void refresh();
        bool postpend(PageCluster *_victim); // merage a page cluster to us
        bool prepend(PageCluster *_victim); // merage a page cluster to us
        std::list<PageCluster*> split(long long _old_piece_len, 
                                      long long _new_piece_len);
        // for any cache hit, this should be called. 
        inline void incRefCnt() { ++refCnt; }
        // once data assembler finishes its job, this should be called (in the descructor?)
        void decRefCnt();
        // return left offset w.r.t. to the whole object (inclusive)
        inline long long getLeft() { return offset; }
        // return right offset w.r.t. to the whole object (exclusive)
        inline long long getRight() { return offset + length; }
        // this function returns true if a page cluster is still new, and is being updated
        inline bool isNew() { return (NEW == state); } 
        // this function returns true if a page cluster is locked (in the middle of swapping out)
        // a locked page cluster should not be removed from the cache
        inline bool isSwappingOut() { return (swapping_out == true); }
        // this function returns true, if the page cluster is considered to be dirty
        // a clean page cluster, when needs to be swapped out, can be dropped from the memory
        inline bool isDirty() { return (FILLED == state || IN_MEM_DIRTY == state); }
        // this function determines whether a swap-in is needed
        inline bool isNotInMem() { return (NOT_IN_MEM == state); }
        // dump info of this page cluster
        std::string toString();
        // we define less operator for sorted set
        bool operator< (const PageCluster &other) const;
        // we define more operator for testing
        bool operator> (const PageCluster &other) const;
};

#endif // __PAGECLUSTER_H
