#include <iostream>
#include <sstream>
#include <string>
#include <algorithm>
#include <functional>

#include "pCache.h"  // for confMap
#include "PagePool.h"
#include "PageCluster.h"
#include "Util.h"

// allocate memory here.
PagePool::PagePool() {
    // counter
    int i;

    // for ascii to integer conversion
    std::stringstream lPageSizeString(confMap[std::string("LARGE_PAGE_SIZE")]);
    lPageSizeString >> lPageSize;
    std::stringstream mPageSizeString(confMap[std::string("MEDIUM_PAGE_SIZE")]);
    mPageSizeString >> mPageSize;
    std::stringstream sPageSizeString(confMap[std::string("SMALL_PAGE_SIZE")]);
    sPageSizeString >> sPageSize;
    std::stringstream lPageNumString(confMap[std::string("LARGE_PAGE_NUM")]);
    lPageNumString >> lPageNum;
    std::stringstream mPageNumString(confMap[std::string("MEDIUM_PAGE_NUM")]);
    mPageNumString >> mPageNum;
    std::stringstream sPageNumString(confMap[std::string("SMALL_PAGE_NUM")]);
    sPageNumString >> sPageNum;
    std::stringstream secChanceRatioString(confMap[std::string("SEC_CHANCE_RATIO")]);
    secChanceRatioString >> sec_chance_ratio;

    // only do second chance swapout for our file-systems
    if (blockPool->accesorType != 4 && 
        blockPool->accesorType != 5) {
        sec_chance_ratio = 0;
    }

    // update the stats module
    stats->small_mem_page_size    = sPageSize;
    stats->small_mem_page_number  = sPageNum;
    stats->medium_mem_page_size   = mPageSize;
    stats->medium_mem_page_number = mPageNum;
    stats->large_mem_page_size    = lPageSize;
    stats->large_mem_page_number  = lPageNum;

    //std::cout << "Page size, L/M/S = " << lPageSize << "/" << mPageSize << "/" << sPageSize << std::endl;
    //std::cout << "Page number, L/M/S = " << lPageNum << "/" << mPageNum << "/" << sPageNum << std::endl;

    // allocate pages, and put all of them on free list
    for (i = 0; i < lPageNum; i++) {
        lFreePages.push_back(new Page(lPageSize));
    }
    for (i = 0; i < mPageNum; i++) {
        mFreePages.push_back(new Page(mPageSize));
    }
    for (i = 0; i < sPageNum; i++) {
        sFreePages.push_back(new Page(sPageSize));
    }
}

// The distructor is not properly implemented. 
// Right now, we assume pages are never freed
// Also, we do not have a exit hook to write these pages now.
PagePool::~PagePool() {

}

// This function is used to find the first busy page that is
// not inused (with zero ref. cnt.)
bool lru_page (Page* _page) {
    return (_page->pc->refCnt <= 0 && _page->pc->sec_chanced == false);
}

// this function find a free or least recently used page
Page* PagePool::allocate(PageCluster* _pc, int _type) {
    // what to return
    Page *ret = NULL; 
    // common iterator
    std::multiset<Page*>::iterator victim; 

    switch (_type) {
        case LPAGE:
            // We start marking page clusters to be swapped out, when the free page number
            // is too small. We do the swapping out only when the pages are used up eventually.
            if ((float)lFreePages.size() / (float) lPageNum < sec_chance_ratio || lFreePages.size() == 0) {
                 // if free pages is empty, then swap out some pages
                victim = find_if(lBusyPages.begin(), lBusyPages.end(), lru_page); 
                if (victim != lBusyPages.end()) {  // if we do find something
                    // mark the page cluster of this page to be swapped out
                    (*victim)->pc->sec_chanced = true;
                    sec_chance_list.insert((*victim)->pc->item);
                }
				else {
					//	no victim page is found
                    std::cout << "Warning no page is found for second chance (1): " << _pc->toString() << std::endl;
                    std::cout << toString() << std::endl;
				}
            }
            // if we really run out of pages, start a sec_chance_swapout()
            if (lFreePages.empty()) {
                sec_chance_swapout();
            }
            // if we get here, there is something is free page list...
            if (lFreePages.empty()) { 
                std::cout << "Warning no victim page is found (1): " << _pc->toString() << std::endl;
                std::cout << toString() << std::endl;
                return NULL;
	        } 
            ret = lFreePages.front();   // hold the reference to this page
            lFreePages.pop_front();      // remove this new page
            // change states in the page
            ret->pc = _pc;
            // update weight
            if (mCache->policy() != Cache::P2P) {
                ret->weight = (double) Util::get_timestamp();
            }
            // push it to busy list
            lBusyPages.insert(ret);
            break;

        case MPAGE:
            // We start marking page clusters to be swapped out, when the free page number
            // is too small. We do the swapping out only when the pages are used up eventually.
            if ((float)mFreePages.size() / (float) mPageNum < sec_chance_ratio || mFreePages.size() == 0) {
                 // if free pages is empty, then swap out some pages
                victim = find_if(mBusyPages.begin(), mBusyPages.end(), lru_page); 
                if (victim != mBusyPages.end()) {  // if we do find something
                    // mark the page cluster of this page to be swapped out
                    (*victim)->pc->sec_chanced = true;
                    sec_chance_list.insert((*victim)->pc->item);
                }
				else {
					//	no victim page is found
                    std::cout << "Warning no page is found for second chance (2): " << _pc->toString() << std::endl;
                    std::cout << toString() << std::endl;
				}
            }
            // if we really run out of pages, start a sec_chance_swapout()
            if (mFreePages.empty()) {
                sec_chance_swapout();
            }
            // if we get here, there is something is free page list...
            if (mFreePages.empty()) { 
                std::cout << "Warning no victim page is found (2): " << _pc->toString() << std::endl;
                std::cout << toString() << std::endl;
                return NULL;
	        } 
            ret = mFreePages.front();   // hold the reference to this page
            mFreePages.pop_front();      // remove this new page
            // change states in the page
            ret->pc = _pc;
            // update weight
            if (mCache->policy() != Cache::P2P) {
                ret->weight = (double) Util::get_timestamp();
            }
            // push it to busy list
            mBusyPages.insert(ret);
            break;

        case SPAGE:
            // We start marking page clusters to be swapped out, when the free page number
            // is too small. We do the swapping out only when the pages are used up eventually.
            if ((float)sFreePages.size() / (float) sPageNum < sec_chance_ratio || sFreePages.size() == 0) {
                 // if free pages is empty, then swap out some pages
                victim = find_if(sBusyPages.begin(), sBusyPages.end(), lru_page); 
                if (victim != sBusyPages.end()) {  // if we do find something
                    // mark the page cluster of this page to be swapped out
                    (*victim)->pc->sec_chanced = true;
                    sec_chance_list.insert((*victim)->pc->item);
                }
				else {
					//	no victim page is found
                    std::cout << "Warning no page is found for second chance (3): " << _pc->toString() << std::endl;
                    std::cout << toString() << std::endl;
				}
            }
            // if we really run out of pages, start a sec_chance_swapout()
            if (sFreePages.empty()) {
                sec_chance_swapout();
            }
            // if we get here, there is something is free page list...
            if (sFreePages.empty()) { 
                std::cout << "Warning no victim page is found (3): " << _pc->toString() << std::endl;
                std::cout << toString() << std::endl;
                return NULL;
	        } 
            ret = sFreePages.front();   // hold the reference to this page
            sFreePages.pop_front();      // remove this new page
            // change states in the page
            ret->pc = _pc;
            // update weight
            if (mCache->policy() != Cache::P2P) {
                ret->weight = (double) Util::get_timestamp();
            }
            // push it to busy list
            sBusyPages.insert(ret);
            break;
    }

    return ret;
}

// this function moves a page from busy to free list
void PagePool::deallocate(Page* _page, int _type) {
    // common iterator
    std::multiset<Page*>::iterator target;

    switch (_type) {
        case LPAGE:
            target = find(lBusyPages.begin(), lBusyPages.end(), _page);
            if (target != lBusyPages.end()) {  // found it
                lBusyPages.erase(target);
				_page->pc = NULL;
                lFreePages.push_back(_page);
            }
            break;

        case MPAGE:
            target = find(mBusyPages.begin(), mBusyPages.end(), _page);
            if (target != mBusyPages.end()) {  // found it
                mBusyPages.erase(target);
				_page->pc = NULL;
                mFreePages.push_back(_page);
            }
            break;

        case SPAGE:
            target = find(sBusyPages.begin(), sBusyPages.end(), _page);
            if (target != sBusyPages.end()) {  // found it
                sBusyPages.erase(target);
				_page->pc = NULL;
                sFreePages.push_back(_page);
            }
            break;
    }
}

// this function moves a page to the tail of its list
void PagePool::refresh(Page* _page){

    // common iterator
    std::multiset<Page*>::iterator target;

    if (_page->size == lPageSize) {
        target = find(lBusyPages.begin(), lBusyPages.end(), _page);
        if (target != lBusyPages.end()) {  // found it
            lBusyPages.erase(target);
            // update weight
            if (mCache->policy() != Cache::P2P) {
                _page->weight = (double) Util::get_timestamp();
            }
            lBusyPages.insert(_page);
        }
    }
    else if (_page->size == mPageSize) {
        target = find(mBusyPages.begin(), mBusyPages.end(), _page);
        if (target != mBusyPages.end()) {  // found it
            mBusyPages.erase(target);
            // update weight
            if (mCache->policy() != Cache::P2P) {
                _page->weight = (double) Util::get_timestamp();
            }
            mBusyPages.insert(_page);
        }
    }
    else { // _page->size == sPageSize
            target = find(sBusyPages.begin(), sBusyPages.end(), _page);
        if (target != sBusyPages.end()) {  // found it
            sBusyPages.erase(target);
            // update weight
            if (mCache->policy() != Cache::P2P) {
                _page->weight = (double) Util::get_timestamp();
            }
            sBusyPages.insert(_page);
        }
    }
}

void PagePool::sec_chance_swapout() {
    for (std::set<Item*>::iterator _item_iter = sec_chance_list.begin(); 
        _item_iter != sec_chance_list.end(); 
        ++_item_iter) {
        PageCluster *_first_pc = NULL;  // the first page cluster of the same item
        PageCluster *_prev_pc = NULL;   // the next page cluster of the same item
                                        // note that we traverse in the reverse direction
//int i = 0;
        // go through page clusters belong to this item
        // if a page cluster's dirty bit is not set, we can ignore it, let's swap it out immediately
        for (std::set<PageCluster*, DereferenceLess>::iterator _pc_iter = (*_item_iter)->pcList.begin();
             _pc_iter != (*_item_iter)->pcList.end();
             ++_pc_iter) {
            if ((*_pc_iter)->sec_chanced == true && (*_pc_iter)->refCnt <= 0) {
                if ((*_pc_iter)->isDirty() == true) {
                    if (_first_pc == NULL) _first_pc = (*_pc_iter); 
                    // ask victim->bc to serialize itself to the disk, 
                    // which will likely call deallocate() function, so that the next
                    // while loop can find some free pages.
                    (*_pc_iter)->swapout();
assert((*_pc_iter)->isNotInMem());
assert((*_pc_iter)->blocks != NULL);
                    assert((*_pc_iter)->blocks != NULL);
                    (*_pc_iter)->blocks->first_pc = _first_pc;  // set the first pc...
                    (*_pc_iter)->blocks->next_pc = NULL;  // set the default for the current one
                    if (_prev_pc != NULL) {
                        assert(_prev_pc->blocks != NULL);
                        _prev_pc->blocks->next_pc = (*_pc_iter); // chain them now...
                    }
                    _prev_pc = (*_pc_iter);
//++i;
                }
                else {   // if it's clean, then just swap it out
                    (*_pc_iter)->swapout();
                }
            }
            else if ((*_pc_iter)->sec_chanced == true) {
                // ref cnt > 0, unset the sec_chance
                (*_pc_iter)->sec_chanced = false;
            }
        }
//printf("swapping out %d page clusters for one item together\n", i);
    }
    sec_chance_list.clear();
}

// We only print busy pages
std::string PagePool::toString() {
    std::string ret("");
    std::multiset<Page*>::iterator _pos;
    ret.append("Large_page_num: ").append(Util::stringify(lPageNum));
    ret.append(" Large_page_size: ").append(Util::stringify(lPageSize));    
    ret.append(" Large_pages: | ");
    for(_pos = lBusyPages.begin(); _pos != lBusyPages.end(); ++_pos) {
        ret.append((*_pos)->toString()).append(" | ");
    } 
    /*
    ret.append(Util::CR_str);
    ret.append("Medium_page_num: ").append(Util::stringify(mPageNum));
    ret.append(" Medium_page_size: ").append(Util::stringify(mPageSize));    
    ret.append(" Medium_pages: | ");
    for(_pos = mBusyPages.begin(); _pos != mBusyPages.end(); ++_pos) {
        ret.append((*_pos)->toString()).append(" | ");
    } 
    ret.append(Util::CR_str);
    ret.append("Small_page_num: ").append(Util::stringify(sPageNum));
    ret.append(" Small_page_size: ").append(Util::stringify(sPageSize));    
    ret.append(" Small_pages: | ");
    for(_pos = sBusyPages.begin(); _pos != sBusyPages.end(); ++_pos) {
        ret.append((*_pos)->toString()).append(" | ");
    } */
    ret.append(Util::CR_str);
    return ret;
}

