#include "PieceLenAnalyzer.h"

#ifdef SIMULATION
	long long virt_time;
	long long get_time()		{ return virt_time;	}
	void set_time(long long t)	{ virt_time = t;	}
#else
	long long get_time()
	{
		struct timeval tv;
		int temp[2];
		if( gettimeofday(&tv, (struct timezone *) &temp) )
		{
			perror("error retrieving time: ");
			return 0;
		}

		return tv.tv_sec * (long long) 1000 + tv.tv_usec / 1000L;
	}
#endif

PieceLenAnalyzer::PieceLenAnalyzer()
{
	pthread_mutex_init(&pla_mutex1, NULL);
	pthread_mutex_init(&pla_mutex2, NULL);

#ifdef SIMULATION
	max_size_piece_sta = 0;
	temp_fp.open("piecelens.txt", fstream::out);
	if(!temp_fp)
		perror("Unable to open piecelens.txt: ");
#endif

#ifdef LOG_REQUESTES
	pthread_mutex_init(&thndlr_mutex, NULL);
	tester_handler._listen();
#endif
}

PieceLenAnalyzer::~PieceLenAnalyzer()
{
	go_down();
	pthread_mutex_destroy(&pla_mutex1);
	pthread_mutex_destroy(&pla_mutex2);
}

void PieceLenAnalyzer::go_down()
{
	sleep(1);	//to let possible still-working threads die (hopefully!)

	for(map<string, FileStatistics *>::iterator iter = piece_sta.begin(); iter != piece_sta.end(); )
	{
		map<string, FileStatistics *>::iterator iter_bak = iter++;
		delete_file(iter_bak->first);
	}

	cerr << "\nPieceLenAnalyzer::~PieceLenAnalyzer(): deleted all in-memory state info" << endl << endl;

	#ifdef SIMULATION
		temp_fp.close();
	#endif
	#ifdef LOG_REQUESTS
		tester_handler.disconnect();
		pthread_mutex_destroy(&thndlr_mutex);
	#endif
}

void PieceLenAnalyzer::delete_file(const string &hash_id)
//delete TorrentPieceStatistic objects pointed in this FileStatistics map, and also delete the map itself
{
	pthread_mutex_lock(&pla_mutex1);	//PieceLenAnalyzer mutex 1: protecting piece_sta and last_access maps

	FileStatistics *file = piece_sta[hash_id];		//it is a map<int, PieceTracker *>
	for(map<int, PieceTracker *>::iterator iter = file->begin(); iter != file->end(); iter++)
		delete iter->second;		//delete the PieceTracker object

	delete file;
	piece_sta.erase(hash_id);
	last_access.erase(hash_id);

	if(dbg_info_level & DBG_INF_MEMFREE)
		cerr << "\nPieceLenAnalyzer::delete_file(): removed in-memory state info of file\n\t" << hash_id << endl;

	pthread_mutex_unlock(&pla_mutex1);
}

int PieceLenAnalyzer::analyze_pieces(const string &hash_id)
//checks whether an almost reliable piece length can be guessed for the file "hash_id" (returns 0 if not),
//every function calling this function must make sure that the piece length is not already guessed for hash_id
//the only thing this functions does after guessing a piece length is reporting it; putting in the map and
//	deleting no-more-needed memories must be handled by the function calling this function,
//note: this function is not thread-safe, so its caller MUST be
{
	static map<string, long long> last_analyzed;

	long long curr_time = get_time();

	map<string, long long>::iterator iter = last_analyzed.find(hash_id);
//	do not analyze if the file "hash_id" is analyzed a few moments ago
	if(iter != last_analyzed.end())
	{
		if(curr_time - iter->second < PLEN_ANALYSIS_INTERVAL)	//e.g. one time analysis per 0.500 seconds
			return 0;
		else
			iter->second = curr_time;
	}
	else		//first time for this hash_id
		last_analyzed[hash_id] = curr_time;

//TODO: optimize the use of mutexes

//	I look for all piece statistics gathered for the file "hash_id" up to now. th maximum guess, as well as
//	the number of pieces supporting that, are remembered, and after all the pieces were analyzed,
//	I look to see if a sufficient number of same guesses is made. that number is [22 - lg(guessed-piecelen)]
//	because (intuitively and obviously!) a piece length of 32KB requires more evidence than a 2MB one.
//	I also make sure there have been no request message for a segment beyond the guess

	static int guess, n_same_guesses, max_req_off;
	n_same_guesses = guess = max_req_off = 0;

//	#ifdef SIMULATION
		string piece_segmap = "--------------------";
		int max_reqed_piece = -1;
//	#endif

	if(dbg_info_level & DBG_INF_PLEN_ANALYSIS0)
		cerr << "\t******************************** analyze_pieces() for " << hash_id << endl;

	FileStatistics *file = piece_sta[hash_id];		//it is a map<int, PieceTracker *>
	for(map<int, PieceTracker *>::iterator iter = file->begin(); iter != file->end(); iter++)
	{
		PieceTracker *piece_tracker = iter->second;
		int dtime = int (curr_time - piece_tracker->last_time);

/*		if(dbg_info_level & DBG_INF_PLEN_ANALYSIS0)
			cerr << "\t****\tpiece #" << iter->first << ": curr - last_access: " << dtime <<
					", piece_tracker->max_delta_time: " << piece_tracker->max_delta_time <<
					"\ttemp info: " << piece_sta.size() << ", " << file->size() << endl;
*/

//		cerr << "" << piece_tracker->get_segmap() << " (" << hash_id << ", " << iter->first << ")" << endl;

		if(piece_tracker->max_req_off > max_req_off)
		{
			max_req_off = piece_tracker->max_req_off;
//			#ifdef SIMULATION
				piece_segmap = piece_tracker->get_segmap();		//a string like XXX--XXXX------X for debugging
				max_reqed_piece = iter->first;
//			#endif
		}
//		check if the piece tracker has some guess, and if it (now) is sufficiently passed from
//		the piece tracker's last added request (this is checked through piece tracker's max_delta_time)
//		it must be passed 2 times the max_delta_time plus 1 second

		if(piece_tracker->guessed_piecelen > 0 &&		//check whether piece_tracker has a guess or not
			piece_tracker->max_delta_time >= 0 && dtime > 1000 + 2 * piece_tracker->max_delta_time)
		{
			if(dbg_info_level & DBG_INF_PLEN_ANALYSIS2)
				cerr << "\nPieceLenAnalyzer::analze_pieces(): piece length " <<
						piece_tracker->guessed_piecelen << " is a (partial) valid guess for the file\n\t"
						<< hash_id << "\n\t\tthat is obtained from its piece #" << iter->first << endl;

			if(piece_tracker->guessed_piecelen > guess)
			{
				guess = piece_tracker->guessed_piecelen;
				n_same_guesses = 1;
			}
			else if(piece_tracker->guessed_piecelen == guess)
				n_same_guesses++;
		}
	}//	end for

//	#ifdef SIMULATION
				long long n_bytes = 0;
				for(map<int, PieceTracker *>::iterator iter = file->begin(); iter != file->end(); iter++)
				{
					PieceTracker *piece_tracker = iter->second;
					n_bytes += piece_tracker->get_n_bytes();
				}
		cerr << "\t" << piece_segmap << "\n\t\thash_id: " << hash_id << ", n_bytes: " << n_bytes << endl << endl;
//	#endif

	if(n_same_guesses && guess >= max_req_off)
	{
		int lg_plen = 0;		//logarithm of guessed piece length
		while(guess >> lg_plen)
			++lg_plen;
		--lg_plen;

		int required = MAX(22 - lg_plen, 1);
		if(n_same_guesses >= required)		//yeahhhh!! successsfully guessed (hopefully a correct guess!)
		{
			#ifdef SIMULATION
				temp_fp << guess << "(piecelen) obtained for the file: " << hash_id << endl;
				int n_requests = 0;
				long long n_bytes = 0;
				for(map<int, PieceTracker *>::iterator iter = file->begin(); iter != file->end(); iter++)
				{
					PieceTracker *piece_tracker = iter->second;
					n_requests += piece_tracker->n_requests;
					n_bytes += piece_tracker->get_n_bytes();
				}
				temp_fp << "\tafter observing " << n_same_guesses << " completed and " <<
						file->size() << " uncompleted pieces: " << n_requests <<
						" requests in total (" << n_bytes << " bytes)" << endl << endl;

				cerr << guess << "\n\tpiecelen" << guess << " obtained for the file " << hash_id << endl;
				cerr	<< "\t\tafter observing " << n_same_guesses << " completed and " <<
						file->size() << " uncompleted pieces: " << n_requests <<
						" requests in total (" << n_bytes << " bytes)" << endl << endl;
			#else
				if(dbg_info_level & DBG_INF_PLEN_ANALYSIS0)
					cerr << "\t****\tpiecelen: " << guess << " obtained from " <<
							n_same_guesses << " pieces ,  for the file " << hash_id << endl;
			#endif

			#ifdef LOG_REQUESTS
				tester_handler.send_piecelen(hash_id, guess, 0 /*n_bytes*/);
			#endif

//			pthread_mutex_unlock(&ap_mutex);
			return guess;		//this is my final guess
		}
		else if(dbg_info_level & DBG_INF_PLEN_ANALYSIS0)
			cerr << "\n\t\tinsufficient number of guesses (" << n_same_guesses << ") supporting \"" <<
						guess << "\" as piece length (" << required << " guesses required)" <<
						endl << "\t\t\tfor the file " << hash_id << endl << endl;
	}
	else if(n_same_guesses && dbg_info_level & DBG_INF_PLEN_ANALYSIS0)
		cerr << "\n\t\tthe guess " << guess << " is less than max_req_off (" << max_req_off << ")" << endl;

	return 0;
}

void PieceLenAnalyzer::revisit_garbages()
//this function must be called infrequently, to delete data kept for not-recently-referenced files
{
	static long long last_time = get_time();
	long long curr_time = get_time();

	if(curr_time - last_time > TIMEOUT)
	{
		if(dbg_info_level & DBG_INF_PLEN_ANALYSIS1)
			cerr << "\nPieceLenAnalyzer::revisit_garbages():looking for " <<
					"not-recently-accessed files to remove their state info" << endl;

		last_time = curr_time;

		pthread_mutex_lock(&pla_mutex1);	//PieceLenAnalyzer mutex 1: protecting piece_sta and last_access maps
			map<string, long long>::iterator begin = last_access.begin();
			map<string, long long>::iterator end = last_access.end();
		pthread_mutex_unlock(&pla_mutex1);

		map<string, long long>::iterator iter_bak;
		for ( map<string, long long>::iterator iter = begin; iter != end; )
		{
			if(curr_time - iter->second > TIMEOUT)		//iter->second is the last acess time
			{
				#ifdef SIMULATION
//					gathering evaluative info (the following line must be commented out later)
					temp_fp << "\n\tremoving in-memory info of the not-recently-accessed file: " <<
																			iter->first << "\n\n";
				#endif
				cerr << "\n\tremoving in-memory info of the not-recently-accessed file: " << iter->first << "\n\n";

				iter_bak = iter++;
				delete_file(iter_bak->first);		//remove info of the not-recently-accessed file
			}
			else
				iter++;
		}//	end for
	}
}


/*int PieceLenAnalyzer::get_piecelen(const string &hash_id)
//returns -1 if piece length does not exist in "piecelen" map
{
	pthread_mutex_lock(&??_mutex);

	if(piecelen.find(hash_id) != piecelen.end())
		return piecelen[hash_id];

//	int plen = analyze_pieces(hash_id);		//analyzes piece states of the file "hash_id"
//	if(plen > 0)
//	{
//		pthread_mutex_unlock(&??_mutex);
//		return plen;
//	}

	pthread_mutex_unlock(&??_mutex);
	return -1;
}*/

int PieceLenAnalyzer::add_request(const string &hash_id, int piece_index, int off_in_piece, int req_len)
//returns 0, unless the just-added request leads to a guess for piece length (the guessed value is returned)
{
#ifdef LOG_REQUESTS
	pthread_mutex_lock(&thndlr_mutex);

	tester_handler.log_request(hash_id, piece_index, off_in_piece, req_len);

	if(tester_handler.get_relisten_request() > 0)
		tester_handler.relisten();

	pthread_mutex_unlock(&thndlr_mutex);
#endif

	pthread_mutex_lock(&pla_mutex2);	//PieceLenAnalyzer mutex 2: protecting the "piecelen" map
		map<string, int>::iterator iter = piecelen.find(hash_id);
		if(iter != piecelen.end())
		{
			pthread_mutex_unlock(&pla_mutex2);
			return iter->second;
		}
	pthread_mutex_unlock(&pla_mutex2);

	pthread_mutex_lock(&pla_mutex1);	//PieceLenAnalyzer mutex 1: protecting piece_sta and last_access maps
//	if(dbg_info_level & DBG_INF_PLEN_ANALYSIS3)
//		cerr << "\nPieceLenAnalyzer::add_request(): adding a request: {" << piece_index <<
//				", " << off_in_piece << ", " << req_len << "}, from file \n\t" << hash_id << endl;

	FileStatistics *file;			//it is a map<int, PieceTracker *>
	map<string, FileStatistics *>::iterator iter1 = piece_sta.find(hash_id);
	if(iter1 == piece_sta.end())	//a new file
	{
		file = piece_sta[hash_id] = new map<int, PieceTracker *>();		//new FileStatistics

		#ifdef SIMULATION
			cerr << "\ta new file: " << hash_id << "\n";
			if(piece_sta.size() > (unsigned) max_size_piece_sta)
			{
				max_size_piece_sta = piece_sta.size();

	//			gathering evaluative info (the following line must be commented out later)
				temp_fp << "\n\tnew max_size_peice_sta= " << max_size_piece_sta << "\t\tfile= " << hash_id << "\n\n";
				cerr	<< "\n\tnew max_size_peice_sta= " << max_size_piece_sta << "\t\tfile= " << hash_id << "\n\n";
			}
		#endif
	}
	else
		file = iter1->second;		//this somewhat-dirty code is to minimize the number of lookups in the map

	PieceTracker *piece_tracker;
	map<int, PieceTracker *>::iterator iter2 = file->find(piece_index);
	if(iter2 == file->end())		//a new piece
	{
		if(file->size() <= MAX_LOGGED_PIECES)
			piece_tracker = (*file)[piece_index] = new PieceTracker();
		else
		{
			cerr << "for file " << hash_id << " more than " << file->size() <<
					" pieces are kept track of (limit is " << MAX_LOGGED_PIECES <<
					" pieces)\n\tno more request will be logged for that" << endl;
			piece_tracker = NULL;
//			pthread_mutex_unlock(&pla_mutex1);
//			return 0;		no! do not return 0 here, let the analysis be done (12 lines later!)
		}
	}
	else
		piece_tracker = iter2->second;	//this somewhat-dirty code os to minimize the number of lookups in the map

	if(piece_tracker != NULL)
	{
		piece_tracker->add_segment(off_in_piece, req_len);	 //add the seg to tracker's statistics
		last_access[hash_id] = get_time();		//used for removing non-active files
	}

	int plen = analyze_pieces(hash_id);		//analyzes piece states of the file "hash_id"
//		this call must be done here, do not migrate it to get_piecelen: the identifier wants to
//		add a request and see the result of analysis, it is more efficient to be done in this
//		way rather than calling this func and then get_piecelen (a useless lookup in map is bypassed)

	pthread_mutex_unlock(&pla_mutex1);

	if(plen > 0)
	{
		pthread_mutex_lock(&pla_mutex2);	//PieceLenAnalyzer mutex 2: protecting the "piecelen" map
			piecelen[hash_id] = plen;
		pthread_mutex_unlock(&pla_mutex2);

		delete_file(hash_id);		//delete info stored for this file in memory
		revisit_garbages();			//this func (revisit_...) must not be called so frequently, so I put it here

		return plen;	//tell the BitTorrTrafficIdentifier that a piece length is guessed
	}

	return 0;
}

/*void PieceLenAnalyzer::set_piecelen(const string &hash_id, int plen)
{
	pthread_mutex_lock(&sp_mutex);

	map<string, int>::iterator iter = piecelen.find(hash_id);
	if(iter == piecelen.end())
//	as expected, no previous piece length is set for the file "hash_id"
	{
		piecelen[hash_id] = plen;
		delete_file(hash_id);		//delete info stored for this file in memory
		revisit_garbages();
//		if(dbg_info_level & DBG_INF_PLEN_ANALYSIS2)
//			cerr << "\nPieceLenAnalyzer::set_piecelen(): piece length " << plen <<
//					" was set for file \n\t" << hash_id <<
//					"\n\n\t\tand associated in-memory state info were successfully deleted" << endl;
	}
	else		//a piece length is already set for this hash_id
	{
		if(iter->second != plen)	//mismatch
		{
			cerr << "\nPieceLenAnalyzer:: error !!!! : setting a piecelen for file " <<
						hash_id << " more than once (different from previous one)" << endl;
			iter->second = plen;
		}
		else
			cerr << "\nPieceLenAnalyzer:: warning !!!! : setting the same piecelen for file " <<
																hash_id << " more than once" << endl;
	}

	pthread_mutex_unlock(&sp_mutex);
}*/

////////////////////////////////////////////////////////////////////////////////
//                              PieceTracker                                  //
////////////////////////////////////////////////////////////////////////////////
PieceTracker::PieceTracker()
{
	segments = new map<int, int>();
	guessed_piecelen = max_req_off = 0;
	last_time = max_delta_time = -1;

	#ifdef SIMULATION
		n_requests = 0;
	#endif
}

PieceTracker::~PieceTracker()
{
	delete segments;
}

void PieceTracker::add_segment(int off_in_piece, int req_len)
{

//	first, try to merge the new request range with the just-after one in the "segments" map, if not
//	successful, try the just-before-one, if again not successful, insert it in the map as a new request range

	bool update_time = false;	//if the new request range is covered before, last access time must not be updated
	guessed_piecelen = 0;	//drop the previously guessed pieceln (if any), it will be
							//set if its value remains valid at the end of this function

	map<int, int>::iterator iter = segments->upper_bound(off_in_piece);		//the first key greater than off

	while(iter != segments->end() && off_in_piece + req_len >= iter->first + iter->second)		//this case
//	happens very very rarely (IF happens!), only when a new requested range is a superset of a previous one
	{
		map<int, int>::iterator old_iter = iter++;
//		if(dbg_info_level & DBG_INF_PTRACKER)
//			cerr << "\n\t\tPiecetracker::add_segment(): request range (" << old_iter->first << ", " <<
//					old_iter->second << ") is erased from map as it is a subset of the new request (" <<
//					off_in_piece << ", " << req_len << ")" << endl;
		segments->erase(old_iter);	//remove the range that is a subset (the super set will be added later)
		update_time = true;
	}

	if(iter != segments->end() && iter->first <= off_in_piece + req_len)	//if request ranges are mergable
	{
//		if(dbg_info_level & DBG_INF_PTRACKER)
//			cerr << "\n\t\tPiecetracker::add_segment(): merging the new request(" <<
//					off_in_piece << ", " << req_len << ") with a previous one (" <<
//					iter->first << ", " <<	iter->second << ")" << endl;
		int new_len = iter->first + iter->second - off_in_piece;
		segments->erase(iter);
		(*segments)[off_in_piece] = new_len;
		update_time = true;
	}
	else	//check if the request range can be merged with something "before" it ("after" has just been checked)
	{
		if(segments->empty())
			iter = segments->end();
		else
		{
//			retrieve the last key less than or equal to off_in_piece
			iter = segments->upper_bound(off_in_piece);
			if(iter != segments->begin())
				iter--;
			else	//there is no key before off_in_piece
				iter = segments->end();
		}

		if(iter != segments->end() && off_in_piece + req_len <= iter->first + iter->second)
//		in this rare case, the new requested range is a subset of a previous one
		{
			;	//do nothing, the request range is already considered
//			if(dbg_info_level & DBG_INF_PTRACKER)
//				cerr << "\n\t\tPiecetracker::add_segment(): the new request(" <<
//						off_in_piece << ", " << req_len << ") is a subset of a previous one (" <<
//						iter->first << ", " <<	iter->second << ")" << endl;
		}

		else if(iter != segments->end() && iter->first + iter->second >= off_in_piece)	//merge
		{
//			if(dbg_info_level & DBG_INF_PTRACKER)
//				cerr << "\n\t\tPiecetracker::add_segment(): merging the new request(" <<
//						off_in_piece << ", " << req_len << ") with the already existing one (" <<
//						iter->first << ", " <<	iter->second << ")" << endl;
			iter->second = off_in_piece + req_len - iter->first;
			update_time = true;
		}

		else	//the new request range and the just-before one are not overlapping
		{
			(*segments)[off_in_piece] = req_len;
			update_time = true;
//			if(dbg_info_level & DBG_INF_PTRACKER)
//				cerr << "\n\t\tPiecetracker::add_segment(): the new request(" << off_in_piece << ", " <<
//							req_len << ") does not overlap any previous one, put it separately in the map" << endl;
		}
	}

	if(off_in_piece + req_len > max_req_off)
		max_req_off = off_in_piece + req_len;

//	set info holding time stuff
	if(update_time)
	{
		long long new_time = get_time();
		if(last_time < 0)		//so this is the first request
			last_time = new_time;
		else
		{
			int delta_time = int (new_time - last_time);
			last_time = new_time;
			if(delta_time > max_delta_time)
				max_delta_time = delta_time;
		}

		if(dbg_info_level & DBG_INF_PTRACKER_TIME)
			cerr << "\n\t\tPiecetracker::add_segment(): last_time was updated to " <<
					new_time << ", max_delta_time: " << max_delta_time << endl << endl;
	}
	else
		if(dbg_info_level & DBG_INF_PTRACKER_TIME)
			cerr << "\n\t\tPiecetracker::add_segment(): update_time was false" << endl << endl;

	if(segments->size() == 1 && segments->begin()->first == 0)
//	we have a continuous range: if it is a power of 2, set it as this PieceTracker's guess (for piece len),
//	validity of this guess (e.g. ignoring a guess based on only 1 request) is done in PieceLenAnalyzer
	{
		int temp_guess = segments->begin()->second;
		while( !(temp_guess & 1) )
			temp_guess >>= 1;

		if( !(temp_guess >> 1) )	//it is (was!) a power of two
		{
//			if(dbg_info_level & DBG_INF_PTRACKER)
//				cerr << "\n\t\tPiecetracker::add_segment(): a potential guess for the piece length: " <<
//																		segments->begin()->second << endl;
			guessed_piecelen = segments->begin()->second;
		}

		#ifdef SIMULATION
			n_requests++;
		#endif
	}
}

#ifdef SIMULATION
#endif
	long long PieceTracker::get_n_bytes()	//used by PieceLenAnalyzer::analyze_pieces() for evaluations
	{
		long long res = 0;
		for(map<int, int>::iterator iter = segments->begin(); iter != segments->end(); iter++)
			res += iter->second;

		return res;
	}
	string PieceTracker::get_segmap()
	{
		if(!max_req_off)
			return "---------- 0";

		string res;
		for(int i = 0; i < 64; i++)	res += '-';
		for(map<int, int>::iterator iter = segments->begin(); iter != segments->end(); iter++)
		{
			int i0 = iter->first * 64 / max_req_off;
			int i1 = (iter->first + iter->second) * 64 / max_req_off;
			for(int j = i0; j <= i1; j++)
				res[j] = 'X';
		}

		char temp[20];
		sprintf(temp, "%d", max_req_off);
		res = res + "  max_req_off: " + temp + ",  max_delta_time: ";

		sprintf(temp, "%d", max_delta_time);
		return res + temp;
	}
#ifdef SIMULATION

struct LoggedRequest
{
	long long t;
	string hash_id;
	int piece_idx, off, len;
};

int main(int argc, char *argv[])
{
	if(argc != 2)
	{
		cerr << "usage: " << argv[0] << " reqlog-file" << endl;
		return 0;
	}

	fstream rlog;
	rlog.open(argv[1], fstream::in);
	if(!rlog)
		perror("Unable to open argv[1]: ");

	vector<LoggedRequest> reqvect;
	while(1)
	{
		LoggedRequest temp;
		rlog >> temp.t;
		rlog >> temp.hash_id;
		rlog >> temp.piece_idx;
		rlog >> temp.off;
		rlog >> temp.len;
		if(rlog.eof())
			break;

		reqvect.push_back(temp);
	}

	rlog.close();

	PieceLenAnalyzer *pla = PieceLenAnalyzer::get_instance();

	for(int i = 0; i < reqvect.size(); i++)
	{
		set_time(reqvect[i].t);
		int plen = pla->add_request(reqvect[i].hash_id, reqvect[i].piece_idx, reqvect[i].off, reqvect[i].len);
		if(plen > 0)
			break;
	}

	cerr << "end!" << endl;

	delete pla;
	return 0;
}
#endif
