#ifndef __STATS_H
#define __STATS_H

#include <string>
#include <map>
#include <set>
#include <pthread.h>
#include <netinet/in.h>

// We flush log queue every 500 events
//#define LOG_QUEUE_LEN 500
#define LOG_QUEUE_LEN 10
#define LOG_FILENAME "log.txt"
#define CLIENT_LOG_FILENAME "clog.txt"

// Note: We embed timestamp and level into description when creating
//       log entries. This is to avoid additional manipulation when
//       sending them over the network. That is, timestamp and level
//       here are used to filter logs, but not for presentation.
class LogEntry {
public:
    long long timestamp;
    int level;
    std::string description;
    LogEntry(long long _timestamp, int _level, std::string _description);
    bool operator< (const LogEntry &other) const {
        return timestamp < other.timestamp;
    }
};

class ClientAddr {
public:
	struct sockaddr_in client_addr;

	ClientAddr(struct sockaddr_in);
	ClientAddr(struct sockaddr);
    // we define less operator for sorted set
    bool operator< (const ClientAddr &other) const;
    // we define is equale operator for sorted set
//    bool operator== (const ClientAddr &other) const;
};

// abc
class ClientStatEntry {
public:
	std::string oid;
    long long startTime;
    long long lastTime;
    long long byte_count;
    long long total_byte;
	std::map<struct sockaddr_in, long long> sender;

	ClientStatEntry(std::string oid);
};

class Stats{
private:
	pthread_mutex_t mut; // for log queue

public:
	Stats();
    ~Stats();

    // printf-like functions are used to log detailed events, we support
    // three level of events: INFO, WARN, and ERR
	// and also one level for logging the client statistics: CLIENT
    const static int ERR = 1;
    const static int WARN = 2;
    const static int INFO = 3;
    const static int CLIENT = 4;
    void lprintf(int level, const char *lFormat, ...);
    // perror's counter part
    void lperror(int level, const char *desc);
    // events sent to lprintf, will be pushed into a set of LogEntry (sorted by
    // timestamp)
    std::set<LogEntry> logQueue;
    // events sent to lprintf in CLIENT level, will be pushed into a set of LogEntry
	// (sorted by timestamp)
    std::set<LogEntry> clientLogQueue;
    // This function checks whether we reach LOG_QUEUE_LEN, if yes, write all
    // logs into a file, and remove them from the queue
    void flushLogs(bool forced);
    // This function is called by our Java daemon, to ask for all event after
    // a specific timestamp. We return a huge string here.
    std::string getLogs(int level, long long timestamp_since);
    // This function is periodically called by either pCache or Conn to generate
    // a periodical report (PR). PR is prefixed with the keyword PR to allow scripts to
    // grep them.
    void gen_PR(); 
    // This function is called once to generate a final report (FR). FR is prefixed
    // with the keyword to allow scripts to grep them
    void gen_FR();

    // Following counters are updated by each modules, and will be presented
    // through our web interaface in a table
    // General Info
    long long request;                  // number of requests (in no. segments)
    long long total_thread;             // total number of thread since the connection manager start
    long active_thread;                 // number of active thread
    
    // Table contains client stats
	std::map<std::string, ClientStatEntry*> clientTable;

	// Cache module
    float high_watermark;
    float low_watermark;
	long long	totalTraff;
	long long	availTraff;
	long long	memTraff;
	std::map<std::string, long long> itemLifetime;
	long long	removedItemCounter;
	long long	avgLifetime;

    // In-memory Buffer
    long long small_mem_page_size;
    long long small_mem_page_number;
    long long medium_mem_page_size;
    long long medium_mem_page_number;
    long long large_mem_page_size;
    long long large_mem_page_number;
    bool merge;   
    bool split_copyover;   // copyover during piece resizing
    // Block device
    long long disk_block_size;
    long long disk_block_number;
    std::string access_type;
    std::string fit_method;
    std::string root_dir;
    std::string dev_name;
    long long num_file;                      // number of files (as well as page clusters)
    long long num_disk_bytes;           // how many bytes are in used
    long long num_disk_blocks;          // how many blocks are in used
    float disk_byte_utilization;        // cache_size_in_bytes / total_disk_capacity_in_bytes
    float disk_block_utilization;       // used disk block / total block number
    long long ap_read;                  // accumulated number of high-level read operations
    long long ap_write;                 // accumulated number of high-level write operations
    long long ap_read_len_in_bytes;     // accumulated write bytes
    long long ap_write_len_in_bytes;    // accumulated read bytes

    // The following function construct the HTML page by filling out all the
    // counters
    void cons_table(std::string &ret);
};

#endif  // __STATS_H

