#ifndef __TRAFFICIDENTIFIER_H
#define __TRAFFICIDENTIFIER_H

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#include <string>
#include <iostream>

#include "Segment.h"
#include "Conn.h"

class TrafficIdentifier
{
public:
	std::string receiver_id, sender_id;

    virtual ~TrafficIdentifier() {
    }

    // The Conn class calls this function to determine whether
    // the incoming data stream speaks our protocol.
    // Returns true if found a match, returns false o.w.
    // Note that the buf will not be discard, and will be
    // used to create an actual TrafficIdentifier if there
    // is a match.
	static bool identify(char* buf, int buf_len);

    // Conn class creates a traffic identifier class when
    // seeing a match. The constructor does the following things:
    //     (1) Look into the first buffer, and try to find out
    //         actual server location in it.
    //     (2) If there is such a thing, call the Conn.connectToServer()
    //         method to setup server socket. Notice that, this is
    //         not mandatory. In the transparent proxy case, Conn class
    //         determines the server location and set up a connection
    //         automatically.
    //     (3) Send a reply to client. This is only required for
    //         non-transparent mode. 
    // Similar to identify, the processed buffer will not be
    // discard, and will be filtered and delivered to the server.
    // See below for traffic filtering methods.
    //
    // Uah, since we don't want to keep any states in this abstrat
    // class, no constructor is needed.
	TrafficIdentifier(char* buf, int buf_len, Conn* conn) {}

    // After the connection object is set up. A thread is spwan to
    // read from both client and server. In general, the received
    // buffer should be relayed to the other party. However, there
    // are special cases where messages need to be augmented.
    //
    // The following two functions are used to filter out these
    // special messages, such as HANDSAHKE (provides oid), REQUEST 
    // (provides required segment), and RESPONSE.
    // Traffic identifier may modify the byte array, remove portion
    // of it, or insert more data into it. The insertions shall never
    // overflow the byte array.
    //
    // When seeing a REQUEST message, c2sFilteres push a new item into
    // the request queue, Conn::reqQueue. c2sFilters also call the 
    // function Conn.cacheRequest(char*, int) to cache the request
    // message in the format of byte array. The usage of this cache
    // will be discussed later.
    //
    // A RESPONSE message provides traffic identifier hints on how to 
    // speed us filtering operations. The first byte of the payload 
    // (actual data) should be returned in start_offset. Start_offset == 0
    // indicates there is no new reply in the buffer. Similar to 
    // REQUEST messages, s2cFilters call the function Conn.cacheResponse
    // (char*, int) to cache the response message in the format of 
    // byte array.
    // 
    // The return value tells Conn class how many bytes should be 
    // relayed to the other party. These two functions return a
    // positive integer X to indicate that the first X bytes should
    // be sent out. The left-over bytes, if any, will be shifted 
    // toward left. This buffer movement (shift) implies that current
    // buffer contains control messages. Conn class immediately calls
    // the same filter function (without further receiving). Filter
    // functions then process and remove the control messages from
    // the buffer.
    //
    // In the case where control message is not completed 
    // (cut in the middle), these two functions return a negative 
    // integer -X to indicate that the first X bytes should be sent 
    // out. The rest content should be moved toward the left, and
    // more data should be received. After more data are received, 
    // Conn class will call the filter function again.
    // Return value 0 is valid means read more data without sending out
    // any portion of the current buffer.
    //
    // Arguments: buf and buf_len are both INPUT and OUTPUT. 
    //            max_len is INPUT. start_offset is output.
    //
    virtual int c2sFilter(char** _start, char* _begin, char** _end) = 0;
    virtual int s2cFilter(char** _start, char* _begin, char** _end, char** start_offset) = 0;

    // when Conn class needs to send a partial request or a fake response,
    // traffic identifier shall prepare the control messages (as they are
    // highly protocol dependent). 
    // The following two functions are responsible to construct these messages.
    // Conn class helps the message construction by providing the most
    // recent cached control messages. The following function augment
    // the data in the buffer and return true if the message should be sent. 
    // It returns false if the message is not necessary or there is an internal
    // error.
    //
    // Import notice: Once createRequest() returns a true to Conn class, the
    //                traffic identifier class will keep track of that "fake"
    //                request, and should filter out the redudant response
    //                sent by the server. This is done without any assistance
    //                from the Conn class.
    // 
    // Arguments: buf and buf_len are both INPUT and OUTPUT. 
    //            max_len is INPUT.
    virtual bool createRequest(char *buf, int* buf_len, Segment* seg) = 0; 
    virtual bool createResponse(char *buf, int* buf_len, Segment* seg) = 0;
};

#endif	//	ifndef __TRAFICIDENTIFIER_H
