/*
 * Soft:        Userspace TCP Splicing library.
 *
 * Author:      Alexandre Cassen, <acassen@linux-l7sw.org>
 *
 * License:
 *              This program is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *              See the GNU General Public License for more details.
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Copyright (C) 2006,2007 Alexandre Cassen, <acassen@linux-l7sw.org>
 */

// [Cheng 12/16/07] Port tcp splicing library to pCache. 
//                  Added pthread mutex protection.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <pthread.h>

#include "tcp_splice.h"
#include "libtcpsplice.h"

/*
 * Global Vars.
 */
static int sockfd = -1;
tcp_splice_getinfo_t tcp_splice_info;
static tcp_splice_user_t *krule = NULL;
static pthread_mutex_t mut;   // lock global resource such as krule
                              // Note that we only lock in-used functions  

/* Kernel channel initialization */
int
tcp_splice_init(void)
{
	socklen_t len;

	if (sockfd > 0)
		return -1;

	len = sizeof(tcp_splice_info);
	if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
		return -1;

	if (getsockopt(sockfd, IPPROTO_IP, TCP_SPLICE_SO_GET_INFO,
		       (char *) &tcp_splice_info, &len)) {
		close(sockfd);
		sockfd = -1;
		return -1;
	}

    pthread_mutex_init(&mut, NULL);

	return 0;
}

/* Kernel channel close */
static void
tcp_splice_close(void)
{
	close(sockfd);
	sockfd = -1;

    // Do we really want to do this? What if there are some threads still out there?
    // pthread_mutex_destroy(&mut);
}

/* Get global informations */
int
tcp_splice_getinfo(void)
{
	socklen_t len;

	len = sizeof(tcp_splice_info);
	return getsockopt(sockfd, IPPROTO_IP, TCP_SPLICE_SO_GET_INFO,
			  (char *) &tcp_splice_info, &len);
}

/* Get kernel module version */
int
vod_version(void)
{
	return tcp_splice_info.version;
}

/* Handle SPLICE_INIT command */
static int
tcp_splice_cmd_splice_init(tcp_splice_user_t * user)
{
	int ret = setsockopt(sockfd, IPPROTO_IP, TCP_SPLICE_SO_SET_SPLICE_INIT,
			             (char *) user, sizeof (*user));
    return ret;
}

/* Handle SPLICE command */
static int
tcp_splice_cmd_splice(tcp_splice_user_t * user)
{
	int ret = setsockopt(sockfd, IPPROTO_IP, TCP_SPLICE_SO_SET_SPLICE,
			             (char *) user, sizeof (*user));
    return ret;
}

#if 0
/* Service table lookup */
tcp_splice_get_services_t *tcp_splice_cmd_get_services(void)
{
	tcp_splice_get_services_t *get;
	socklen_t len;

	len = sizeof(*get) +
	      sizeof(tcp_splice_service_entry_t) * tcp_splice_info.num_services;
	if (!(get = malloc(len)))
		return NULL;

	get->num_services = tcp_splice_info.num_services;
	if (getsockopt(sockfd, IPPROTO_IP,
		       TCP_SPLICE_SO_GET_SERVICES, get, &len) < 0) {
		free(get);
		return NULL;
	}
	return get;
}
#endif

/* Talk dispatcher */
static void
tcp_splice_talk(int cmd)
{
	int result = -1;

	switch (cmd) {
	case TCP_SPLICE_SO_SET_SPLICE_INIT:
		result = tcp_splice_cmd_splice_init(krule);
		break;
	case TCP_SPLICE_SO_SET_SPLICE:
		result = tcp_splice_cmd_splice(krule);
		break;
	}
}

/* Kernel module probe */
int
modprobe_tcp_splice(void)
{
	char *argv[] = { "/sbin/modprobe", "-s", "-k", "--", "tcp_splice", NULL };
	int child;
	int status;
	int rc;

	if (!(child = fork())) {
		execv(argv[0], argv);
		exit(1);
	}

	rc = waitpid(child, &status, 0);

	if (!WIFEXITED(status) || WEXITSTATUS(status))
		return 1;

	return 0;
}

/* Initialize kernel communication channel */
int
tcp_splice_start(void)
{
	if (tcp_splice_init()) {
        // No longer do modprobe, we move the kernel code into kernel
		//if (modprobe_tcp_splice() || tcp_splice_init()) {
			return -1;
		//}
	}

	/* Allocate global user rules */
	krule = (tcp_splice_user_t *) malloc(sizeof(tcp_splice_user_t));

	return 0;
}

/* Close kernel communication channel */
void
tcp_splice_stop(void)
{
	tcp_splice_close();
	free(krule);
}

/* INIT command */
int
tcp_splice_initfd(int sd1, int sd2)
{
	memset(krule, 0, sizeof(tcp_splice_user_t));
	krule->s1 = sd1;
	krule->s2 = sd2;

    pthread_mutex_lock(&mut);
	tcp_splice_talk(TCP_SPLICE_SO_SET_SPLICE_INIT);
    pthread_mutex_unlock(&mut);
	return 0;
}

/* SPLICE command */
int
tcp_splice_splicefd(int sd1, int sd2, int len)
{
    pthread_mutex_lock(&mut);
	memset(krule, 0, sizeof(tcp_splice_user_t));
	krule->s1 = sd1;
	krule->s2 = sd2;
	krule->n = len;

	tcp_splice_talk(TCP_SPLICE_SO_SET_SPLICE);
    pthread_mutex_unlock(&mut);
	return 0;
}



