/*
 * Soft:        Kernelspace TCP Splicing forwarding engine. Based on work done
 *              by Wensong Zhang, <wensong@linux-vs.org>.
 *
 * Author:      Alexandre Cassen, <acassen@linux-l7sw.org>
 *
 * License:
 *              This program is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *              See the GNU General Public License for more details.
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Copyright (C) 2006,2007 Alexandre Cassen, <acassen@linux-l7sw.org>
 */

#ifndef _TCP_SPLICE_H
#define _TCP_SPLICE_H

#define CONFIG_TCP_SPLICE_DEBUG

#define TCP_SPLICE_VERSION_CODE			0x000102
#define TCP_SPLICE_DATE_CODE			0x0f0107
#define NVERSION(version)                       \
	(version >> 16) & 0xFF,                 \
	(version >> 8) & 0xFF,                  \
	version & 0xFF
#define NVERSION_STRING "v%d.%d.%d (%.2d/%.2d, 20%.2d)\n", \
	NVERSION(TCP_SPLICE_VERSION_CODE), NVERSION(TCP_SPLICE_DATE_CODE)

/*
 *      TCP SPLICE socket options
 */
#define TCP_SPLICE_BASE_CTL		(64+2048+64)		/* base */

#define TCP_SPLICE_SO_SET_NONE		TCP_SPLICE_BASE_CTL	/* just peek */
#define TCP_SPLICE_SO_SET_SPLICE_INIT	(TCP_SPLICE_BASE_CTL+1)
#define TCP_SPLICE_SO_SET_SPLICE	(TCP_SPLICE_BASE_CTL+2)
#define TCP_SPLICE_SO_SET_MAX		TCP_SPLICE_SO_SET_SPLICE

#define TCP_SPLICE_SO_GET_VERSION	TCP_SPLICE_BASE_CTL
#define TCP_SPLICE_SO_GET_INFO		(TCP_SPLICE_BASE_CTL+1)
#define TCP_SPLICE_SO_GET_MAX		TCP_SPLICE_SO_GET_INFO

/*
 *	Declaration structure exported to userland application
 */
typedef struct _tcp_splice_user {
	int s1;		/* Socket file descriptor 1 */
	int s2;		/* Socket file descriptor 2 */
	int n;		/* Number of bytes written to s2 */
} tcp_splice_user_t;

typedef struct _tcp_splice_getinfo {
        unsigned int version;
        unsigned int size;
} tcp_splice_getinfo_t;


/*
 *      TCP SPLICE Connection Flags
 */
#define TCP_SPLICE_CONN_F_HASHED	0x0001		/* hashed entry */
#define TCP_SPLICE_CONN_F_TPROXY	0x0002		/* hashed entry */


#ifdef __KERNEL__

#include <linux/autoconf.h>
#include <linux/list.h>                 /* for struct list_head */
#include <linux/spinlock.h>             /* for struct rwlock_t */
#include <linux/skbuff.h>               /* for struct sk_buff */
#include <linux/ip.h>                   /* for struct iphdr */
#include <asm/atomic.h>                 /* for struct atomic_t */
#include <net/dst.h>			/* for struct dst_entry */
#include <linux/compiler.h>

#define TCP_SPLICE_BUG() BUG()
#define TCP_SPLICE_ERR(msg...) printk(KERN_ERR "tcp_splice: " msg)
#define TCP_SPLICE_INFO(msg...) printk(KERN_INFO "tcp_splice: " msg)
#define TCP_SPLICE_WARNING(msg...) printk(KERN_WARNING "tcp_splice: " msg)
#define TCP_SPLICE_ERR_RL(msg...)				\
    do {							\
	    if (net_ratelimit())				\
		    printk(KERN_ERR "tcp_splice: " msg);	\
    } while (0)

#ifdef CONFIG_TCP_SPLICE_DEBUG
extern int tcp_splice_get_debug_level(void);
#define TCP_SPLICE_DBG(level, msg...)			\
    do {						\
	    if (level <= tcp_splice_get_debug_level())	\
		    printk(KERN_DEBUG "tcp_splice: " msg);	\
    } while (0)
#define TCP_SPLICE_DBG_RL(msg...)				\
    do {						\
	    if (net_ratelimit())			\
		    printk(KERN_DEBUG "tcp_splice: " msg);	\
    } while (0)
#else	/* NO DEBUGGING at ALL */
#define TCP_SPLICE_DBG(level, msg...)  do {} while (0)
#define TCP_SPLICE_DBG_RL(msg...)  do {} while (0)
#endif

#ifdef CONFIG_TCP_SPLICE_DEBUG
#define EnterFunction(level)						\
    do {								\
	    if (level <= tcp_splice_get_debug_level())			\
		    printk(KERN_DEBUG "Enter: %s, %s line %i\n",	\
			   __FUNCTION__, __FILE__, __LINE__);		\
    } while (0)
#define LeaveFunction(level)						\
    do {								\
	    if (level <= tcp_splice_get_debug_level())			\
			printk(KERN_DEBUG "Leave: %s, %s line %i\n",	\
			       __FUNCTION__, __FILE__, __LINE__);	\
    } while (0)
#else
#define EnterFunction(level)   do {} while (0)
#define LeaveFunction(level)   do {} while (0)
#endif


/*
 *      TCP SPLICE State Values
 */
enum {
	TCP_SPLICE_S_NONE = 0,
	TCP_SPLICE_S_ESTABLISHED,
	TCP_SPLICE_S_SYN_SENT,
	TCP_SPLICE_S_SYN_RECV,
	TCP_SPLICE_S_FIN_WAIT,
	TCP_SPLICE_S_TIME_WAIT,
	TCP_SPLICE_S_CLOSE,
	TCP_SPLICE_S_CLOSE_WAIT,
	TCP_SPLICE_S_LAST_ACK,
	TCP_SPLICE_S_LISTEN,
	TCP_SPLICE_S_SYNACK,
	TCP_SPLICE_S_LAST
};


struct tcp_splice_timeout_table {
	atomic_t refcnt;
	int scale;
	int timeout[TCP_SPLICE_S_LAST+1];
};


struct tcp_splice_ctx {
	__u32                   laddr;		/* locate address */
	__u32                   raddr;          /* remote address */
	__u16                   lport;
	__u16                   rport;

	/* sequence numbers at the spliced point */
	__u32			splice_iss;	/* initial send sequence */
	__u32			splice_irs;	/* initial receive sequence */

	/* tcp timestamps at the spliced point*/
	__u32			splice_tsv;	/* timestamp value */
	__u32			splice_tse;	/* timestamp echo reply */
	char			timestamp_ok;

	/* routings destination cache */
	spinlock_t		dst_lock;	/* lock of dst_cache */
	struct dst_entry	*dst_cache;	/* destination cache entry */
	u32			dst_rtos;	/* RT_TOS(tos) for dst */
};


/* direction definition */
#define FROM_FIRST_CONN		0
#define FROM_SECOND_CONN	1

/*
 *	TCP splicing connection
 */
struct tcp_splice_conn {
	struct list_head        f_list;		/* first hash table */
	struct list_head        s_list;		/* second hash table */

	atomic_t		refcnt;		/* reference count */

	/* two sockets/connections */
	struct socket		*socket[2];
	atomic_t		need_rst;	/* need resetting socks */
	struct tcp_splice_ctx	ctx[2];		/* two spliced connections context */

	/* Flags and state transition */
	spinlock_t		lock;		/* lock for state transition */
	volatile __u16		flags;		/* status flags */
	volatile __u16          state;		/* state info */

	/* timeout */
	volatile unsigned long  timeout;
	struct tcp_splice_timeout_table *timeout_table;
	struct timer_list       timer;		/* expiration timer */

	/* packet transmitter */
	int (*packet_xmit)(struct sk_buff *skb);
};


/*
 *      TCP SPLICE core functions
 *      (from tcp_splice_core.c)
 */
extern const char *tcp_splice_proto_name(unsigned proto);


/*
 *     TCP SPLICE connection handling functions
 *     (from tcpsp_conn.c)
 */
extern struct tcp_splice_timeout_table tcp_splice_timeout_tbl;

/* TCPSP connection entry hash table */
#ifndef CONFIG_TCP_SPLICE_TAB_BITS
#define CONFIG_TCP_SPLICE_TAB_BITS   16
#endif
#define TCP_SPLICE_CONN_TAB_BITS     CONFIG_TCP_SPLICE_TAB_BITS
#define TCP_SPLICE_CONN_TAB_SIZE     (1 << TCP_SPLICE_CONN_TAB_BITS)
#define TCP_SPLICE_CONN_TAB_MASK     (TCP_SPLICE_CONN_TAB_SIZE - 1)


/*
 *	TCP SPLICE connection handling
 *	(from tcp_splice_conn.c)
 */
extern struct tcp_splice_conn *
tcp_splice_conn_get(__u32 saddr, __u16 sport, __u32 daddr, __u16 dport, int *dir);

/* put back the conn without restarting its timer */
static inline void __tcp_splice_conn_put(struct tcp_splice_conn *cp)
{
	atomic_dec(&cp->refcnt);
}
extern void tcp_splice_conn_put(struct tcp_splice_conn *cp);

extern struct tcp_splice_conn *
tcp_splice_conn_new(struct socket *sock1, struct socket *sock2, int n);
#define tcp_splice_conn_expire_now(cp)  cp->timer.function((unsigned long)cp)
extern int tcp_splice_socket_init(struct socket *sock1, struct socket *sock2);

extern const char * tcp_splice_state_name(int state);
extern int tcp_splice_set_state(struct tcp_splice_conn *cp, int direction,
				struct iphdr *iph, struct tcphdr *th);
extern int tcp_splice_conn_init(void);
extern void tcp_splice_conn_cleanup(void);

/*
 *      TCP SPLICE control data and functions
 *      (from tcp_splice_ctl.c)
 */
extern int tcp_splice_control_init(void);
extern void tcp_splice_control_cleanup(void);

/*
 *	TCP SPLICE xmit routine
 *	(from tcp_splice_xmit.c)
 */
extern void __tcp_splice_reset_sock(struct sock *sk);
extern void tcp_splice_dst_reset(struct tcp_splice_ctx *ctx);
extern int tcp_splice_forward_skb(struct tcp_splice_conn *cp,
				  struct sk_buff *skb, int direction);


#endif /* __KERNEL__ */

#endif	/* _TCP_SPLICE_H */
